package App::Env;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

App::Env - Application environment

=head1 SYNOPSIS

 use App::Env;

=head1 DESCRIPTION

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Pod::Usage;
use Cwd qw(abs_path);
use FindBin;
use File::Basename;
use Getopt::Long;
use Storable qw(dclone);

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Config::Validator;
use Config::Parser;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant FLAG_GENERATE_CFG_FILE	=> 'DEFAULT';

use constant KW_PATH_BASE => 'path_base';
use constant KW_PATH_CONF => 'path_conf';
use constant KW_PATH_LIB  => 'path_lib';
use constant KW_PATH_LOG  => 'path_log';
use constant KW_PATH_TMP  => 'path_tmp';

use constant KW_CONFIGS           => 'configs';
use constant KW_CFG_FILE_NAME     => 'cfg_file_name';
use constant KW_CFG_PARSER_PARAMS => 'cfg_parser_params';

use constant KW_ARGUMENT    => 'argument';
use constant KW_CONFIG      => 'config';
use constant KW_OPTION      => 'option';
use constant KW_TYPE        => 'type';
use constant KW_STACK       => 'stack';
use constant KW_DEFAULT     => 'default';
use constant KW_REQUIRED    => 'required';
use constant KW_DESCRIPTION => 'description';


#-- Static public class variables (our) ---------------------------------------#

# Locations of application components
our $APP_PATH_BASE;
our $APP_PATH_BIN;
our $APP_PATH_CONF;
our $APP_PATH_LIB;
our $APP_PATH_LOG;
our $APP_PATH_TMP;

# Name of the current program
our $APP_NAME;

#-- Static protected class variables (my) -------------------------------------#

# All configurations loaded from the external config file
my $CONFIGS;

# Storage for optional values
my $OPTIONS;

# Autogenerated name of the config file
my $DEFAULT_CONFIG_FILE;

# Used config file
my $CONFIG_FILE;

# Flags to display help and man page
my $FLAG_HELP;
my $FLAG_MAN;
my $FLAG_VERSION;
my $FLAG_CFGVIEW;
my $FLAG_CFGVALID;
my $FLAG_GENHLP_C;
my $FLAG_GENHLP_O;

# List of all valid datatypes
my @DATA_TYPES;

# Translation table from parser data types to Getopt::Long data types
my %DATA_TYPES;

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1.3';
    $DEVEL   = 0;

    $APP_PATH_BIN  = (abs_path("$FindBin::Bin") =~ m{^(/.*)$})[0];
    $APP_PATH_BASE = $APP_PATH_BIN;
	$APP_PATH_CONF = $APP_PATH_BIN;
	$APP_PATH_LIB  = $APP_PATH_BIN;
	$APP_PATH_LOG  = $APP_PATH_BIN;
	$APP_PATH_TMP  = $APP_PATH_BIN;

	$APP_NAME = basename($0);

	$CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID} = {};

	$OPTIONS = {};

	$DEFAULT_CONFIG_FILE = "$APP_PATH_CONF/$APP_NAME.conf";

	$FLAG_HELP     = 0;
	$FLAG_MAN      = 0;
	$FLAG_VERSION  = 0;
	$FLAG_CFGVIEW  = 0;
	$FLAG_CFGVALID = 0;
	$FLAG_GENHLP_C = 0;
	$FLAG_GENHLP_O = 0;

	@DATA_TYPES = Config::Validator->valid_datatypes();

	%DATA_TYPES = (
		Config::Validator::DT_BOOL()     => '!',
		Config::Validator::DT_FLAG()     => '+',
		Config::Validator::DT_INT()      => '=i',
		Config::Validator::DT_FLOAT()    => '=f',
		Config::Validator::DT_STRING()   => '=s',
		Config::Validator::DT_DATE()     => '=s',
		Config::Validator::DT_TIME()     => '=s',
		Config::Validator::DT_DATETIME() => '=s',
		Config::Validator::DT_IPV4()     => '=s',
		Config::Validator::DT_IPV6()     => '=s',
		Config::Validator::DT_IP()       => '=s',
		Config::Validator::DT_HOSTNAME() => '=s',
		Config::Validator::DT_HOST()     => '=s',
		Config::Validator::DT_EMAIL()    => '=s',
		Config::Validator::DT_FILE()     => '=s',
		Config::Validator::DT_ENUM()     => '=s',
		Config::Validator::DT_LIST()     => '=s@',
		Config::Validator::DT_HASH()     => '=s%',
	);

}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item init($) [PUBLIC,STATIC]

  Usage     : my $error = App::Env->init($ENV_CONFIG);
  Purpose   : Initialize the application environment
  Arguments :
  Returns   :
  Throws    :

=cut

sub init($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($params,) 	= @_;

	my $error;

	while(1) {
		### Checking validity and sanity of configuration specifications
		$error = $class->_validate_cs($params);
		last if $error;

		### Updating the application locations
		$error = $class->_init_paths($params->{KW_PATH_BASE()},
									 $params->{KW_PATH_CONF()},
									 $params->{KW_PATH_LIB()},
									 $params->{KW_PATH_LOG()},
									 $params->{KW_PATH_TMP()});
		last if $error;

		### Loading the application options from command line
		$error = $class->_parse_options($params->{KW_CONFIGS()});
		last if $error;

		### Setting the default configuration values
		$error = $class->_set_defaults($params->{KW_CONFIGS()});
		last if $error;

		### Loading and setting the configurations from config file
		$error = $class->_set_configs($params->{KW_CFG_FILE_NAME()}, $params->{KW_CONFIGS()}, $params->{KW_CFG_PARSER_PARAMS()});
		last if $error;

		### Seting the application options
		$error = $class->_set_options($params->{KW_CONFIGS()});
		last if $error;

		### Loading and setting the application arguments
		$error = $class->_set_arguments($params->{KW_CONFIGS()});
		last if $error;

		### Validate the presence of necessary configurations
		$error = $class->_check_config($params->{KW_CONFIGS()});
		last if $error;

		last;
	}

	### Reacting to the rest of the general action flags
	$class->_view_versions()                            if $FLAG_VERSION;
	$class->_view_config($params->{KW_CONFIGS()})       if $FLAG_CFGVIEW;
	$class->_validate_config($error)                    if $FLAG_CFGVALID;

	return $error if $error;
	return 0;
}

=item absolute_path($) [PUBLIC,STATIC]

  Usage     :
  Purpose   : Generate the absolute path to the given relative file path from application base path
  Arguments :
  Returns   :
  Throws    :

=cut

sub absolute_path($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($file_name,) = @_;

	# If the given file_name is absolute, return it back
	return $file_name if ($file_name =~ m{^/.*$});

	# Otherwise generate the absolute path to the file
	return abs_path("$APP_PATH_BASE/$file_name");
}

=item configs(;$$) [PUBLIC,STATIC]

  Usage     : my $configs = App::Env->configs();
              my $configs = App::Env->configs('section');
              my $configs = App::Env->configs('section','id');
  Purpose   : Retrieve the configuration values, optionaly from given section and ID
  Arguments :
  Returns   :
  Throws    :

=cut

sub configs(;$$)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($section, $id) = @_;

	if ($section) {
		if ($id) { return dclone($CONFIGS->{$section}->{$id}) if $CONFIGS->{$section}->{$id}; }
		else     { return dclone($CONFIGS->{$section}) if $CONFIGS->{$section}; }
	}
	else {
		return dclone($CONFIGS);
	}
	return undef;
}

=item configs_general() [PUBLIC,STATIC]

  Usage     : my $configs = App::Env->configs_general();
  Purpose   : Retrieve all configuration values from general section
  Arguments :
  Returns   :
  Throws    :

=cut

sub configs_general()
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	return $class->configs(Config::Parser::DFLT_SECTION_NAME, Config::Parser::DFLT_SECTION_ID);
}

=item config($;$) [PUBLIC,STATIC]

  Usage     : my $configs = App::Env->config('key','default value');
  Purpose   : Retrieve given configuration value from general section
  Arguments :
  Returns   :
  Throws    :

=cut

sub config($;$)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($key, $default) = @_;
	croak ((caller(0))[3] . ": config key must be given") unless defined($key);

	# Look into the loaded configuration values
	if (exists($CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID}->{$key})) {
		return $CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID}->{$key};
	}
	return $default;
}

=item AUTOLOAD [PUBLIC,STATIC]

  Usage     : my $pid_file = App::Env->pid_file;
              my $log_file = App::Env->log_file;
  Purpose   : Direct access to the configurations in general section via the autogenerated access methods
  Arguments :
  Returns   :
  Throws    :

=cut

sub AUTOLOAD {
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my $attr = our $AUTOLOAD;

	# Perform some sanity checks first
	$attr =~ s/.*:://;
	return if $attr =~ /[A-Z]+/;
	return unless $attr =~ /[_a-zA-Z][_a-zA-Z0-9]*/;

	# We do not want any missspelings here, so croak on error
	croak ((caller(0))[3] . ": requested invalid configuration '$attr'") unless exists($CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID}->{$attr});

	# Try to retrieve and return the configuration value
	return $class->config($attr);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _validate_cs($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Validate configuration specifications
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if not invoked on class

sub _validate_cs($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($specification,) = @_;
	return "Environment specifications must be stored in a hash reference" unless ref $specification eq 'HASH';

	my $configs = $specification->{KW_CONFIGS()};
	return "Configuration specifications must be stored in a hash reference" unless ref $configs eq 'HASH';
	map {
		return "Specification of configuration '$_' must be stored in a hash reference" unless ref $configs->{$_} eq 'HASH';
		return "Configuration '$_' does not have specified data type" unless $configs->{$_}->{KW_TYPE()};
		return "Configuration '$_' does have invalid data type '".$configs->{$_}->{KW_TYPE()}."', possible values are: @DATA_TYPES" unless Config::Validator->is_valid_datatype($configs->{$_}->{KW_TYPE()});
		return "Configuration '$_' can be given both by argument and option, choose only one" if $configs->{$_}->{KW_OPTION()} and $configs->{$_}->{KW_ARGUMENT()};
		return "Configuration '$_' does not have any description" unless $configs->{$_}->{KW_DESCRIPTION()};
	} keys(%{$configs});
	return 0;
}

# _merge_configs($$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Merge configurations in given hashes
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  : Performs deep copy of HASHES from source to destination

sub _merge_configs($$)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($configs, $destination) = @_;

	# Unless explicitly given the destination, assume the root of config storage
	$destination = $CONFIGS unless $destination;

	# Walk through the given configurations
	foreach my $cfg (keys %$configs) {
		# If configuration contains HASH, perform deep copy
		if (ref $configs->{$cfg} eq 'HASH') {
			$destination->{$cfg} = {} unless exists $destination->{$cfg};
			$class->_merge_configs($configs->{$cfg},$destination->{$cfg});
		}
		# Otherwise copy the configuration value to destination storage
		else {
			$destination->{$cfg} = $configs->{$cfg};
		}
	}
}

#-------------------------------------------------------------------------------

# _init_paths($$$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Initialize appplication paths
#  Arguments :
#  Returns   :
#  Throws    :

sub _init_paths($$$$$)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($base_path, $conf_path, $lib_path, $log_path, $tmp_path) = @_;

	if ($base_path) {
		return "BASE path '$base_path' must point to an existing directory" unless -d $base_path;
		$APP_PATH_BASE	= $base_path;
	}
	if ($conf_path) {
		return "CONF path '$conf_path' must point to an existing directory" unless -d $conf_path;
		$APP_PATH_CONF	= $conf_path;
	}
	if ($lib_path) {
		return "LIB path '$lib_path' must point to an existing directory" unless -d $lib_path;
		$APP_PATH_LIB	= $lib_path;
	}
	if ($log_path) {
		return "LOG path '$log_path' must point to an existing directory" unless -d $log_path;
		$APP_PATH_LOG	= $log_path;
	}
	if ($tmp_path) {
		return "TMP path '$tmp_path' must point to an existing directory" unless -d $tmp_path;
		$APP_PATH_TMP	= $tmp_path;
	}
	return 0;
}

# _parse_options($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Parse options from command line
#  Arguments :
#  Returns   :
#  Throws    :

sub _parse_options($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($config_specs,) = @_;

	# Proceed only if configuration specifications were given
	my %params = ();
	my $dt;

	# First some general options and action flags
	$params{"help"}            = \$FLAG_HELP;
	$params{"man"}             = \$FLAG_MAN;
	$params{"version"}         = \$FLAG_VERSION;
	$params{"config-file=s"}   = \$CONFIG_FILE;
	$params{"config-view"}     = \$FLAG_CFGVIEW;
	$params{"config-validate"} = \$FLAG_CFGVALID;
	$params{"dev-ghc"}         = \$FLAG_GENHLP_C;
	$params{"dev-gho"}         = \$FLAG_GENHLP_O;

	# Now add user defined options to the list
	foreach my $config (keys(%$config_specs)) {
		if (my $name = $config_specs->{$config}->{KW_OPTION()}) {
			$dt = Config::Validator->is_valid_datatype($config_specs->{$config}->{KW_TYPE()});
			unless ($config_specs->{$config}->{KW_STACK()}) {
				$params{"$name" . $DATA_TYPES{$dt}} = \$OPTIONS->{$config};
			}
			else {
				$params{"$name" . $DATA_TYPES{Config::Validator::DT_LIST()}} = \$OPTIONS->{$config};
			}
		}
	}

	### Getopt params: %params

	# Try to parse the options
	my $result = GetOptions(%params) or pod2usage(2);
	return "Error parsing application options" unless $result;

	### Getopt options: $OPTIONS

	# Immediatelly react to some general action flags
	pod2usage(-verbose => 1, -exitval => 0) if $FLAG_HELP;
	pod2usage(-verbose => 2, -exitval => 0) if $FLAG_MAN;

	$class->_generate_help_cfg($config_specs) if $FLAG_GENHLP_C;
	$class->_generate_help_opt($config_specs) if $FLAG_GENHLP_O;

	return 0;
}

# _set_defaults($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Initialize configurations to given default values
#  Arguments :
#  Returns   :
#  Throws    :

sub _set_defaults($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($config_specs,) = @_;

	my ($error, $type, $value);
	foreach my $config (keys(%$config_specs)) {
		$value = $config_specs->{$config}->{KW_DEFAULT()};
		if (defined $value) {
			$CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID}->{$config} = $value;
		}
		else {
			$CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID}->{$config} = undef;
		}
	}
	return 0;
}

# _set_configs($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Load configuration from external config file
#  Arguments :
#  Returns   :
#  Throws    :

sub _set_configs($$$)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($cfg_file, $config_specs, $parser_settings) = @_;

	# If the config file name is given, we are supposed to load configurations from external file
	if ($cfg_file) {
		# Different configuration file may be given via options
		$cfg_file = $CONFIG_FILE if $CONFIG_FILE;
		# If the 'cfg_file' name equals FLAG_GENERATE_CFG_FILE, use the generated name
		$cfg_file = $DEFAULT_CONFIG_FILE if (uc($cfg_file) eq FLAG_GENERATE_CFG_FILE);
		return "Config file '$cfg_file' must be an existing ordinary file" unless -T $cfg_file;

		# Generate the parser settings
		$parser_settings = {} unless ref $parser_settings eq 'HASH';;
		if ($config_specs) {
			foreach my $config (keys(%$config_specs)) {
				$parser_settings->{'structure'}->{Config::Parser::DFLT_SECTION_NAME}->{$config} = $config_specs->{$config}->{KW_TYPE()};
				if ($config_specs->{$config}->{KW_STACK()}) {
					$parser_settings->{'stack_values'} = ($parser_settings->{'stack_values'})?join(' ', $config, $parser_settings->{'stack_values'}):$config;
				}
			}
		}

		# Attempt to load the configurations
		my $parser = Config::Parser->new($parser_settings);
		my $cfgs = $parser->config($cfg_file);
		return "Configuration file error: " . join("\n", @{$parser->get_errors()}) unless $cfgs;

		# Merge the obtained configurations with current default values
		$class->_merge_configs($cfgs);
	}
	return 0;
}

# _set_options() [PROTECTED]
#
#  Usage     :
#  Purpose   : Set the option values loaded from command line to the configurations
#  Arguments :
#  Returns   :
#  Throws    :

sub _set_options()
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($config_specs, ) = @_;

	my ($error, $type, $value);
	foreach my $config (keys(%$OPTIONS)) {
		$value = $OPTIONS->{$config};
		if (defined $value) {
			$CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID}->{$config} = $value;
		}
	}
	return 0;
}


# _set_arguments($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Apply the options loaded from command line to the configurations
#  Arguments :
#  Returns   :
#  Throws    :

sub _set_arguments($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($config_specs,) = @_;

	# Proceed only if configuration specifications were given
	if ($config_specs,) {
		# First make HASH in form position => value
		my %args = ();
		for (my $i = 0; $i <= $#ARGV; $i++) {
			$args{++$i} = $ARGV[$i];
		}

		# Now assign the the argument values loaded from command line to the configurations
		my ($error, $type, $value);
		foreach my $config (keys(%$config_specs)) {
			if (my $index = $config_specs->{$config}->{KW_ARGUMENT()}) {
				$value = $args{$index};
				if (defined $value) {
					$CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID}->{$config} = $value;
				}
				else {
					return "Configuration '$config' is missing argument on position '$index'";
				}
			}
		}
	}
	return 0;
}

# _check_config($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Validate the environment configurations and EXIT
#  Arguments :
#  Returns   :
#  Throws    :

sub _check_config($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($specification,) = @_;

	my $configs = $CONFIGS->{Config::Parser::DFLT_SECTION_NAME}->{Config::Parser::DFLT_SECTION_ID};
	my ($error, $type, $value);
	foreach my $config (keys(%$specification)) {
		# Mandatory/optional test
		if ($specification->{$config}->{KW_REQUIRED()} and not defined $configs->{$config}) {
			return "Configuration '$config' is missing value, but marked to be required";
		}
		# Datatype conversion and test
		if (defined $configs->{$config}) {
			$type = $specification->{$config}->{KW_TYPE()};
			$type = "list:$type" if ($specification->{$config}->{KW_STACK()});
			($error, $value) = Config::Validator->validate($type, $configs->{$config});

			return "Invalid value for configuration '$config': $error" if $error;
			$configs->{$config} = $value;
		}
	}
	return 0;
}

# _view_versions() [PROTECTED]
#
#  Usage     :
#  Purpose   : View the versions of all used modules and EXIT
#  Arguments :
#  Returns   :
#  Throws    :

sub _view_versions()
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);

	no strict;

	print STDERR uc($APP_NAME) . " - USED MODULE NAME AND VERSION LIST:\n";
	foreach my $module (sort keys %INC) {
		my $module_name = $module;
		$module_name =~ s|/|::|g;
		$module_name =~ s|\.pm$||;
		$module_name =~ s|\.pl$||;
		my $version_var = $module_name . "::VERSION";
		print STDERR " $module_name: " . ((${$version_var}) ? ${$version_var} : 'unknown' ) . "\n";
	}

	exit(0);
}

# _view_config($) [PROTECTED]
#
#  Usage     :
#  Purpose   : View the environment configurations and EXIT
#  Arguments :
#  Returns   :
#  Throws    :

sub _view_config($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($specification,) = @_;

	print STDERR "Application name:      $APP_NAME\n";
	print STDERR "Application path:      $ENV{PATH}\n";
	print STDERR "Application BASE path: $APP_PATH_BASE\n";
	print STDERR "Application BIN path:  $APP_PATH_BIN\n";
	print STDERR "Application CONF path: $APP_PATH_CONF\n";
	print STDERR "Application LIB path:  $APP_PATH_LIB\n";
	print STDERR "Application LOG path:  $APP_PATH_LOG\n";
	print STDERR "Application TMP path:  $APP_PATH_TMP\n";
	print STDERR "Application libraries: ".join(', ',@INC)."\n";
	print STDERR "Application configs:\n\n";
	$class->_dump_config($CONFIGS, $specification, 0);
	print STDERR "\n";
	exit(0);
}

# _dump_config($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   : View the environment configurations and EXIT
#  Arguments :
#  Returns   :
#  Throws    :

sub _dump_config($$$)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($config, $specs, $depth) = @_;

	++$depth;
	foreach my $key (sort keys(%$config)) {
		print STDERR ("    " x $depth);
		if    (ref $config->{$key} eq 'HASH')  { print STDERR $key . ":"; }
		elsif (ref $config->{$key} eq 'ARRAY') { print STDERR $key . ": " . join(',', @{$config->{$key}}); }
		else                                   { print STDERR $key . ": " . $config->{$key}; }

		if (defined $specs->{$key})	{
			print STDERR " [" . $specs->{$key}->{KW_DEFAULT()} . "]" if defined $specs->{$key}->{KW_DEFAULT()};
			print STDERR " (" . $specs->{$key}->{KW_DESCRIPTION()} . "|" . $specs->{$key}->{KW_TYPE()} . ")";
		}
		print STDERR "\n";

		if (ref $config->{$key} eq 'HASH')     { $class->_dump_config($config->{$key}, $specs, $depth); }
	}
}

# _validate_config($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Validate the environment configurations and EXIT
#  Arguments :
#  Returns   :
#  Throws    :

sub _validate_config($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($error,) = @_;

	if ($error)	{
		print STDERR "CONFIGURATION ERROR: $error\n";
		exit(1);
	}
	print STDERR "Configuration OK\n";
	exit(0);
}

# _generate_help_cfg($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Validate the environment configurations and EXIT
#  Arguments :
#  Returns   :
#  Throws    :

sub _generate_help_cfg($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($configs,) = @_;

	foreach my $cfg_key (sort keys(%$configs)) {
		my $cfg = $configs->{$cfg_key};
		if ($cfg->{KW_CONFIG()}) {
			my $dt = Config::Validator->is_valid_datatype($cfg->{KW_TYPE()});
			print STDERR '# '.$cfg->{KW_DESCRIPTION()}."\n";
			print STDERR "# Type      : $dt" . (($cfg->{KW_STACK()})?'(MULTI)':'')."\n";
			print STDERR "# Default   : ".((defined $cfg->{KW_DEFAULT()})?$cfg->{KW_DEFAULT()}:'UNDEFINED')."\n";
			print STDERR "# Occurency : ".(($cfg->{KW_REQUIRED()})?'MANDATORY':'OPTIONAL')."\n";
			print STDERR "\n#$cfg_key = ".((defined $cfg->{KW_DEFAULT()})?$cfg->{KW_DEFAULT()}:'UNDEFINED')."\n\n";
		}
	}

	exit(0);
}

# _generate_help_opt($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Validate the environment configurations and EXIT
#  Arguments :
#  Returns   :
#  Throws    :

sub _generate_help_opt($)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($configs,) = @_;

	foreach my $cfg_key (sort keys(%$configs)) {
		my $cfg = $configs->{$cfg_key};
		if ($cfg->{KW_OPTION()}) {
			my $dt = Config::Validator->is_valid_datatype($cfg->{KW_TYPE()});
			my @spec = ();
			if ($dt eq Config::Validator::DT_BOOL) {
				print STDERR '=item B<--'.$cfg->{KW_OPTION()}.">\n";
				print STDERR '=item B<--no'.$cfg->{KW_OPTION()}.">";
				push(@spec, (($cfg->{KW_REQUIRED()})?'MANDATORY':'OPTIONAL'));
				push(@spec, "DEFAULT: ".$cfg->{KW_DEFAULT()});
			}
			elsif ($dt eq Config::Validator::DT_FLAG) {
				print STDERR '=item B<--'.$cfg->{KW_OPTION()}.">";
				push(@spec, (($cfg->{KW_REQUIRED()})?'MANDATORY':'OPTIONAL'));
				push(@spec, "DEFAULT: ".$cfg->{KW_DEFAULT()});
				push(@spec, "MULTI");
			}
			else {
				print STDERR '=item B<--'.$cfg->{KW_OPTION()}."=".uc($dt).">";
				push(@spec, (($cfg->{KW_REQUIRED()})?'MANDATORY':'OPTIONAL'));
				push(@spec, "DEFAULT: ".((defined $cfg->{KW_DEFAULT()})?$cfg->{KW_DEFAULT()}:'UNDEFINED'));
				push(@spec, "MULTI") if defined $cfg->{KW_STACK()};
			}
			print STDERR ' I<['.join(',',@spec)."]>\n";
			print STDERR "\n".$cfg->{KW_DESCRIPTION()}."\n\n";
		}
	}

	exit(0);
}

=pod

=back

=cut

1;
