package App::Feature;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

App::Feature - Short module description

=head1 SYNOPSIS

use App::Feature;

=head1 DESCRIPTION

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Initiated flag - every feature can be initiated only once
my $INITIATED = 0;

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = App::Feature->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : App::Feature reference on success, croaks on failure
  Throws    : Croaks, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
	### Creating App.Feature instance...
	my $self = bless ({}, $class);
	return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Croaks, if invoked on class

=cut

sub DESTROY()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	### Destroying App.Feature instance...
}

=item initiate(;@) [PUBLIC,HYBRID]

  Usage     : App::Feature->initiate(@args);
              $instance->initiate(@args);
  Purpose   : Configure and initiate the application feature
  Arguments : Unknown, see the documentation of appropriate subclass
  Returns   : Nothing
  Throws    : Croaks in case of any error

=cut

sub initiate(;@)
{
	my $invocant = shift;
	croak ((caller(0))[3] . ": method needs to be implemented in subclass");
}

=item initiated(;$) [PUBLIC,HYBRID]

  Usage     : my $flag = App::Feature->initiated(1);
              my $flag = $instance->initiated(1);
              unless (App::Feature->initiated()) { ... }
              unless ($instance->initiated()) { ... }
  Purpose   : Set/get the initiated flag
  Arguments : BOOL $flag - pass TRUE to set the flag
  Returns   : BOOL $flag
  Throws    : Nothing
  Comment   : Once the flag is set, it can`t be unset

=cut

sub initiated(;$)
{
	my $invocant = shift;
	my ($flag,) = @_;
	$INITIATED = 1 if $flag;
	return ($INITIATED)?$INITIATED:0;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new App::Feature instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : App::Feature $self reference
#  Throws    : Croaks, if not invoked on instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
	my $self = shift;
	croak ((caller(0))[3] . ": method needs to be implemented in subclass");
}

=pod

=back

=cut

1;
