package App::Feature::Daemon;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

App::Feature::Daemon - Daemonize the application

=head1 SYNOPSIS

 use App::Feature::Daemon;

 # Initiate the daemonization feature
 App::Feature::Daemon->initiate('/var/run/app.pid','user','group');

=head1 DESCRIPTION

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Cwd 'abs_path';
use File::Basename;
use IO::File;
use POSIX qw(setsid);

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use App::Feature;
use Log::Loger;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA $DEVEL);
    $VERSION 	= '0.1';
    $DEVEL		= 1;
    @ISA 		= ('App::Feature');
}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant PID_PATH  	=> '/var/run/';
use constant LOG_PATH 	=> '/var/log/';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#
my $PID_FILE = undef;

#*******************************************************************************
#
# MODULE (DE)INITIALIZATIONS
#
#*******************************************************************************

#-- Module clean-up code (global destructor) ----------------------------------#
END {
	# If the PID file was created, we have to remove it
	if ($PID_FILE) {
		App::Feature::Daemon->regain_privileges();
		die "Insufficient privileges to remove PID file $PID_FILE (C)\n" unless -w $PID_FILE;
		die "Can`t remove PID file $PID_FILE (C)\n" unless unlink $PID_FILE;
		Log::Loger->info(__PACKAGE__, "[AFD] PID file '$PID_FILE' succesfully removed (C)");
	}
}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

sub daemonize {
	my $invocant = shift;
	return $invocant->initiate(@_);
}

=item initiate(;$$$) [PUBLIC,HYBRID]

  Usage     : App::Feature::Daemon->initiate($pid_path, $user, $group);
              $instance->initiate($pid_path, $user, $group);
  Purpose   : Configure and initiate the application`s damonization feature
  Arguments : STRING $pid_path - path to the PID file
              STRING $user - name of the user account to which to drop privileges
              STRING $group - name of the group account to which to drop privileges
  Returns   : Nothing
  Throws    : Croaks in case of any error

=cut

sub initiate(;$$$$)
{
	my $invocant = shift;
	my ($pid_path, $user, $group, $force) = @_;

	# Feature can be initiated only once, or it must be forced (if you know, what
	# are you doing...do you?)
	if (not($invocant->initiated) or $force) {
		Log::Loger->info(__PACKAGE__, "[AFD] Daemonization sequence has begun...");

		# First init the logging subsystem to catch 'warn' and 'die' messages, we
		# do not want any message to be lost
		$invocant->_setup_loging();

		# Now detach from the terminal
		$invocant->_detach();

		# Create PID file in the desired directory
		$invocant->_create_pid_file($pid_path);

		# Normalize the application environment
		$invocant->_normalize();

		# Drop privileges to another account (if desirable)
		$invocant->drop_privileges($user, $group) if $user;

		$invocant->initiated(1);
		Log::Loger->info(__PACKAGE__, "[AFD] Daemonization sequence completed...");
	}
	else {
		Log::Loger->error(__PACKAGE__, "[AFD] Daemonization sequence was already performed...");
	}
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if invoked on class

=cut

sub DESTROY()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

	# If the PID file was created, we have to remove it
	if ($self->{PID_FILE}) {
		$self->regain_privileges();
		die "Insufficient privileges to remove PID file $PID_FILE (I)\n" unless -w $PID_FILE;
		die "Can`t remove PID file $PID_FILE (I)\n" unless unlink $PID_FILE;
		Log::Loger->info(__PACKAGE__, "[AFD] PID file '$PID_FILE' succesfully removed (I)");
	}
}

=item drop_privileges($;$$) [PUBLIC,HYBRID]

  Usage     : Called from initiate() method, do not call manually
  Purpose   : Drop privileges of the process to the given account and group
  Arguments : STRING $user - user account name [MANDATORY]
              STRING $group - group account name [OPTIONAL]
              BOOL $permanent - drop privileges permanently flag [OPTIONAL]
  Returns   : Nothing
  Throws    : Dies in case of failure

=cut

sub drop_privileges($;$$)
{
	my $invocant = shift;
	my ($user, $group, $permanent) = @_;

	# Unless given explicitly, use the same group name as user name
	$group = ($group) ? $group : $user;

	# Translate given values to UID and GID
	my $uid = getpwnam($user) or die "Can`t get UID for user '$user'\n";
	my $gid = getgrnam($group) or die "Can`t get GID for group '$group'\n";

	# Let the magic happen
	$) = "$gid $gid";
	$( = $gid;
	$> = $uid;

	# Maybe we should drop the privileges permanently
	$< = $uid if ($permanent);

	my $msg = "Privileges succesfully ".(($permanent)?'!PERMANENTLY! ':'')."dropped to '$user':'$group' ('$uid':'$gid')";
	Log::Loger->info(__PACKAGE__, "[AFD] $msg");
}

=item regain_privileges() [PUBLIC,HYBRID]

  Usage     : App::Feature::Daemon->regain_privileges();
              $instance->regain_privileges();
  Purpose   : Regain back the original process privileges (unless they were dropped permanently)
  Arguments : None
  Returns   : Nothing
  Throws    : Dies in case of failure
=cut

sub regain_privileges()
{
	$> = $<;

	Log::Loger->info(__PACKAGE__, "[AFD] Privileges succesfully regained");
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init() [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new App::Feature::Daemon instance
#  Arguments : None
#  Returns   : App::Feature::Daemon reference $self
#  Throws    : Dies, if not invoked on instance

sub _init()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
}

# _setup_loging() [PROTECTED,HYBRID]
#
#  Usage     : Called from initiate() method, do not call manually
#  Purpose   : Setup handling of SIG_WARN and SIG_DIE to log the messages
#  Arguments : None
#  Returns   : Nothing
#  Throws    : Dies in case of failure

sub _setup_loging()
{
	my $invocant = shift;

	$SIG{__WARN__} = \&_log_warn;
	$SIG{__DIE__} = \&_log_die;

	Log::Loger->info(__PACKAGE__, "[AFD] SIG_WARN and SIG_DIE loging successfully configured");
}

# _detach() [PROTECTED,HYBRID]
#
#  Usage     : Called from initiate() method, do not call manually
#  Purpose   : Detach from terminal and redirect STDs to /dev/null
#  Arguments : None
#  Returns   : Nothing
#  Throws    : Dies in case of failure

sub _detach()
{
	my $invocant = shift;

	Log::Loger->debug(__PACKAGE__, "[AFD] Forking before detaching from terminal");

	# Spawn new child process, so we can severe parent relationship
	die "Can`t fork: $!\n" unless defined (my $child = fork);
	exit 0 if $child;

	Log::Loger->debug(__PACKAGE__, "[AFD] Successfully forked, detaching from terminal");

	# Become the session leader
	POSIX::setsid();

	# Reopen the STD`s to the black hole
	open(STDIN, "</dev/null") or die "Can`t redirect /dev/null to STDIN: $!\n";
	open(STDOUT, ">/dev/null") or die "Can`t redirect STDOUT to /dev/null: $!\n";
	open(STDERR, ">&STDOUT") or die "Can`t redirect STDERR to STDOUT: $!\n";

	Log::Loger->info(__PACKAGE__, "[AFD] Successfully detached from terminal");
}

# _create_pid_file($) [PROTECTED,HYBRID]
#
#  Usage     : Called from initiate() method, do not call manually
#  Purpose   : Create PID file and store the process PID in it
#  Arguments : STRING $file - name of the PID file [OPTIONAL]
#  Returns   : Nothing
#  Throws    : Dies in case of failure

sub _create_pid_file($)
{
	my $invocant = shift;
	my ($pid_file,) = @_;

	# Sanitize the PID file name (generate if not given and make the path absolute)
	$pid_file = $invocant->_get_pid_file_name($pid_file);

	# It is an error, if the file already exists, deal with it first
	if (-e $pid_file) {
		my $fh = IO::File->new($pid_file) or die "Can`t open existing PID file '$pid_file': $!\n";
		my $pid = <$fh>; close($fh);

		# There should be only single number in the PID file
		die "Invalid PID file\n" unless $pid =~ /^(\d+)$/;
		# Maybe process with this PID is running
		die "Server already running with PID $1\n" if kill 0 => $1;
		# Otherwise attempt to remove it
		die "Can`t remove PID file '$pid_file'\n" unless -w $pid_file && unlink $pid_file;
		Log::Loger->info(__PACKAGE__, "[AFD] Removed PID file '$pid_file' for defunct server process $1");
	}

	# Store the PID of current process into the
	my $fh = IO::File->new($pid_file, O_WRONLY|O_CREAT|O_EXCL, 0644) or die "Can't create PID file $pid_file: $!\n";
	print $fh $$;
	close($fh);

	# Store the name of the PID file, so we can remove it later
	if (blessed($invocant)) { $invocant->{PID_FILE} = $pid_file; }
	else                    { $PID_FILE = $pid_file; }

	Log::Loger->info(__PACKAGE__, "[AFD] Created PID file '$pid_file' containing  PID '$$'");
}

# _normalize() [PROTECTED,HYBRID]
#
#  Usage     : Called from initiate() method, do not call manually
#  Purpose   : Performs process environment normalization
#  Arguments : None
#  Returns   : Nothing
#  Throws    : Dies in case of failure

sub _normalize()
{
	my $invocant = shift;

	# Change current working directory to filesystem root
	chdir('/') or die "Can`t change current working directory to '/': $!\n";

	# Forget the file creation mask
	umask(0);

	# Delete some dangerous, or unusefull environment variables
	delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};

	# Set or reinitialize important environment variables
	$ENV{PATH} = '/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin';

	Log::Loger->info(__PACKAGE__, "[AFD] Process environment was normalized");
}

#-------------------------------------------------------------------------------

# _log_warn(@) [PROTECTED,HYBRID]
#
#  Usage     : Do not call manually, will be called as SIG_WARN handler
#  Purpose   : Log the warning message using loging framework
#  Arguments :
#  Returns   : Nothing
#  Throws    : Dies in case of failure

sub _log_warn(@)
{
	Log::Loger->log(basename($0), 'WARNING', _log_msg(@_));
}

# _log_die(@) [PROTECTED,HYBRID]
#
#  Usage     : Do not call manually, will be called as SIG_DIE handler
#  Purpose   : Log the warning message using loging framework
#  Arguments :
#  Returns   : Nothing
#  Throws    : Dies in case of failure

sub _log_die(@)
{
	Log::Loger->log_die(basename($0), 'EMERG', "UNCAUGHT EXCEPTION: " . _log_msg(@_));
}

# _get_pid_file_name($) [PROTECTED,HYBRID]
#
#  Usage     : my $pid_file = $instance->_get_pid_file_name($pid_file);
#  Purpose   : Sanitize PID file name or generate default one, if not given
#  Arguments : STRING $pid_file - name of the PID file [OPTIONAL]
#  Returns   : STRING $pid_file
#  Throws    : Dies in case of failure

sub _get_pid_file_name($)
{
	my $invocant = shift;
	my ($pid_file,) = @_;

	$pid_file = $pid_file || (PID_PATH . basename($0) . ".pid");
	$pid_file = abs_path($pid_file);
	return $1 if "$pid_file" =~ /(.*)/;
}

#-------------------------------------------------------------------------------

# _log_msg(@) [PROTECTED,HYBRID]
#
#  Usage     :
#  Purpose   : Format log message from 'warn' and 'die' messages
#  Arguments :
#  Returns   : Nothing
#  Throws    : Dies in case of failure

sub _log_msg(@)
{
	my $msg = join(" ", @_) || "Something went wrong";
	my ($pack, $filename, $line) = caller(1);
	$msg .= " at $filename line $line\n" unless $msg =~ /\n$/;
	$msg =~ s/\n$//;
	return $msg;
}

=pod

=back

=cut

1;
