package App::Feature::Loger;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

App::Feature::Loger - Short module description

=head1 SYNOPSIS

 use App::Feature::Loger;

 # Initiate the loging feature
 App::Feature::Loger->initiate(
		'std::notice',
		'file:/var/log/app.log:notice',
		'syslog::warning',
		'email:from@domain;to@domain;Subject:notice',
	);

=head1 DESCRIPTION

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use App::Feature;
use Log::Core::Essentials;
use Log::Loger;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 1;
    @ISA     = ('App::Feature');
}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant LOG_PATH 	=> '/var/log/';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#
my %VALID_LOG_MODULES = (
	'email'  => '_sw_email',
	'file'   => 1,
	'std'    => 1,
	'syslog' => '_sw_syslog',
);

#*******************************************************************************
#
# MODULE (DE)INITIALIZATIONS
#
#*******************************************************************************

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item initiate(@) [PUBLIC,HYBRID]

  Usage     : App::Feature::Loger->initiate(@log_channels);
              $instance->initiate(@log_channels);
  Purpose   : Configure and initiate the application`s loging feature
  Arguments : ARRAY @log_channels - specifications of requested log channels
  Returns   : Nothing
  Throws    : Croaks in case of any error

=cut

sub initiate(@)
{
	my $invocant = shift;
	my @log_channels = @_;

	my %cfg = $invocant->_parse_configuration(@log_channels);

	my ($cntf, $fname, $cntw, $wname, $wrs);
	foreach my $thr (keys %cfg)
	{
		$wrs = []; $fname = undef;
		if ($thr) {
			++$cntf; $fname = $invocant->_setup_filter($cntf, $thr);
		}

		foreach my $wrt (@{$cfg{$thr}})
		{
			++$cntw; $wname = $invocant->_setup_writer($cntw, $wrt);
			push(@$wrs, $wname);
		}

		Log::Loger->channel($fname, $wrs, 'accept');
	}
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if invoked on class

=cut

sub DESTROY {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init() [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new App::Feature::Loger instance
#  Arguments : None
#  Returns   : App::Feature::Loger reference $self
#  Throws    : Dies, if not invoked on instance

sub _init
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
}

# _parse_configuration() [PROTECTED]
#
#  Usage     : my %cfg = App::Feature::Daemon->_parse_configuration(@_);
#              my %cfg = $instance->_parse_configuration(@_);
#  Purpose   : Parse the new App::Feature::Loger configuration
#  Arguments : ARRAY of STRINGS @log_channels - channel configurations
#  Returns   : HASH %cfg
#  Throws    : Nothing

sub _parse_configuration(@)
{
	my $invocant = shift;
	my @log_channels = @_;

	my (%cfg, $module, $arg, $threshold, $severity);
	foreach my $c (@log_channels) {
		($module, $arg, $threshold) = split(/:/,$c);
		die "[AFL] Invalid loger configuration '$c'\n" unless $module;
		die "[AFL] Invalid log writer module '$module'\n" unless $VALID_LOG_MODULES{$module};
		$severity = ($threshold)?$threshold:'';

		if ($severity) {
			$severity = Log::Core::Essentials->severity_as_str($threshold);
			die "[AFL] Invalid log filter severity threshold '$threshold'\n" unless $severity;
		}

		push(@{$cfg{$severity}}, [$module, $arg]);
	}
	return %cfg;
}

# _setup_filter($$) [PROTECTED]
#
#  Usage     : my $fname = App::Feature::Daemon->_setup_filter($index, $threshold);
#              my $fname = $instance->_setup_filter($index, $threshold);
#  Purpose   : Setup new Log::Loger filter
#  Arguments : INTEGER $index - next sequence index for this filter
#              STRING $threshold - severity threshold for this filter
#  Returns   : STRING $fname identifier of the filter
#  Throws    : Dies if given invalid arguments

sub _setup_filter($$)
{
	my $invocant = shift;
	my ($index, $threshold) = @_;

	my $fname = "f-$index";
	Log::Loger->filter($fname, 'threshold', $threshold, 'r');
	return $fname;
}

# _setup_writer($$) [PROTECTED]
#
#  Usage     : my $fname = App::Feature::Daemon->_setup_writer($index, $threshold);
#              my $fname = $instance->_setup_writer($index, $threshold);
#  Purpose   : Setup new Log::Loger writer
#  Arguments : INTEGER $index - next sequence index for this writer
#              ARRAY REFERENCE $cfg - writer configuration
#                  STRING $module - module name
#                  STRING $args - module args
#  Returns   : STRING $fname identifier of the filter
#  Throws    : Dies if given invalid arguments

sub _setup_writer($$)
{
	my $invocant = shift;
	my ($index, $cfg) = @_;

	my $wname = "w-$index";
	my ($module, $args) = @$cfg;

	my $cbk = $VALID_LOG_MODULES{$module};
	if ($cbk =~ /^_sw_/) {
		$invocant->$cbk($wname, $module, $args);
	}
	else {
		Log::Loger->writer($wname, $module, $args, undef, 'r');
	}

	return $wname;
}

#-------------------------------------------------------------------------------
# Setup callbacks for specific writers
#-------------------------------------------------------------------------------

#  Purpose   : Setup new Log::Loger writer
#  Arguments : STRING $wid - name if the writer
#              STRING $module - log module name
#              STRING $args - log module args
#  Returns   : Nothing
#  Throws    : Dies if given invalid arguments

sub _sw_email($$$)
{
	my $invocant = shift;
	my ($wid, $module, $args) = @_;

	my @a = split(/[,;]/, $args);
	my $cfg = {
			'From'    => $a[0],
			'To'      => $a[1],
			'Subject' => $a[2],
		};
	Log::Loger->writer($wid, $module, $cfg, 'r');
}

sub _sw_syslog($$$)
{
	my $invocant = shift;
	my ($wid, $module, $args) = @_;

	Log::Loger->writer($wid, $module,  $args, 'r');
}

=pod

=back

=cut

1;
