package CTCT::Module::Test;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

CTCT::Module::Test - Short module description

=head1 SYNOPSIS

use CTCT::Module::Test;

=head1 DESCRIPTION

Extensive module description

=head1 BUGS

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Time::HiRes qw (time);
use File::Find;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use CTCT::Module;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Hash containing all valid commands supported by this module and their asociated
# callback names for indirect execution.
my %COMMANDS = (
			'coverage'	=> [
						'_command_coverage',
						'Check unit test coverage for given project root',
					]
		);

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA);
    $VERSION 	= '0.1';
    @ISA 		= ('CTCT::Module');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item get_commands() [PUBLIC]

  Usage     :
  Purpose   : Return list of all commands this module provides
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_commands()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return %COMMANDS;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new CTCT::Module::Test instance
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	return $self;
}

# _add_class_to_list [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _add_class_to_list
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	# Skip directories
	return unless -f $File::Find::name;

	# Skip everything that is not perl module
	return unless $_ =~ /^[-\d\w]+\.pm$/;

	# Apply optional log file pattern
	if ($self->{PATTERN})
	{
		my $pattern = $self->{PATTERN};
		return unless $_ =~ /$pattern/;
	}

	# Mark the file for adding to the package
	$self->{CLASS_FILES}->{$File::Find::name} = $_;
}

# _add_test_to_list [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _add_test_to_list
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	# Skip directories
	return unless -f $File::Find::name;

	# Skip everything that is not perl module
	return unless $_ =~ /^[-\d\w\.]+\.t$/;

	# Mark the file for adding to the package
	$self->{TEST_FILES}->{$File::Find::name} = $_;
}


# _find_class_and_test_files($;$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _find_class_and_test_files($;$)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($lib_dir, $tests_dir, $pattern) = @_;

	$self->{PATTERN} = $pattern || undef;

	find({'wanted' => sub { $self->_add_class_to_list }, 'untaint' => 1}, $lib_dir);
	find({'wanted' => sub { $self->_add_test_to_list }, 'untaint' => 1}, $tests_dir);

	return ($self->{CLASS_FILES}, $self->{TEST_FILES});
}

# _check_unit_test_coverage($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _check_unit_test_coverage($$)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($output, $results, $project_root) = @_;

	my $lib_dir = $project_root . '/lib';
	die "Invalid usage: '$lib_dir' is not a library directory\n" unless -d $lib_dir;

	my $tests_dir = $project_root . '/tests/unit/t';
	die "Invalid usage: '$tests_dir' is not a unit tests directory\n" unless -d $tests_dir;

	my $start = time();

	my ($files, $tests) = $self->_find_class_and_test_files($lib_dir, $tests_dir);
	$results->{'count_f'} = scalar(keys(%$files));
	$results->{'count_t'} = scalar(keys(%$tests));

	my $ptrn = $lib_dir . '/';
	foreach my $file (sort keys(%$files))
	{
		my $test_path = $file; $test_path =~ s|$ptrn||;
		$test_path =~ s|$ptrn||;
		$test_path =~ s|/|.|g;
		$test_path =~ s|\.pm|.t|g;
		$test_path =  $tests_dir . '/' . $test_path;

		if ($tests->{$test_path})
		{
			#$output .= "   - $file [PRESENT]\n";
			push(@{$results->{'present'}}, $file);
			delete $tests->{$test_path};
		}
		else {
			push(@{$results->{'missing'}}, $file);
			$output .= "   - $file [MISSING]\n";
		}
	}

	foreach my $file (sort keys(%$tests))
	{
		if ($file =~ m|/_[-\d\w\.]+\.t$|)
		{
			push(@{$results->{'special'}}, $file);
			$output .= "   - $file [SPECIAL]\n";
		}
		else {
			push(@{$results->{'unknown'}}, $file);
			$output .= "   - $file [UNKNOWN]\n";
		}
	}

	my $stop = time();
	$output .= " - Unit test coverage duration: " . ($stop - $start) . " [s]\n";
	$results->{'duration'} = $stop - $start;

	return ($output, $results);
}

#-----------------------------------------------------------------------

# _command_coverage [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _command_coverage
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($chain, @args) = @_;
	my ($project_root,) = @args;
	die "Invalid usage: [$0 @$chain] command must receive project root directory as argument\n" unless $project_root;
	die "Invalid usage: [$0 @$chain] project root must be a directory\n" unless -d $project_root;
	my ($output, $results) = ('', {});

	$output .= " - Checking unit test coverage in project root '$project_root':\n";

	return $self->_check_unit_test_coverage($output, $results, $project_root);
}

=pod

=back

=cut

1;
