package Color::Palette;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Color::Pallette - Short module description

=head1 SYNOPSIS

 use Color::Pallette;

 ##
 # Either use static class methods:
 ##

 # Translate color name to RGB hexadecimal representation
 my $rgb = Color::Palette->as_rgb('snow');        # $rgb = 'FFFAFA';
 my $rgb = Color::Palette->as_rgb('FFFAFA');      # $rgb = 'FFFAFA';

 # Translate RGB hexadecimal representation to color name
 my $color = Color::Palette->as_string('FFFAFA'); # $color = 'snow';
 my $color = Color::Palette->as_string('snow');   # $color = 'snow';

 # Fetch random color
 my ($color, $rgb) = Color::Palette->random_color();

 # Fetch random and unique color
 my ($color, $rgb) = Color::Palette->random_unique_color();

 # Fetch random and unique color and give a list of already used and therefore
 # forbidden colors
 my ($color, $rgb) = Color::Palette->random_unique_color('red','green','blue');
 my ($color, $rgb) = Color::Palette->random_unique_color('FF0000','FFFF00','0000FF');

 # Fetch random color palette of given length (number of colors)
 my @palette = Color::Palette->random_palette(10);

 # Fetch next color from static (hardcoded) palette
 my ($color, $rgb) = Color::Palette->color();

 # Fetch next unique color from static (hardcoded) palette
 my ($color, $rgb) = Color::Palette->unique_color();

 # Fetch next unique color from static (hardcoded) palette and give a list of
 # already used and therefore forbidden colors
 my ($color, $rgb) = Color::Palette->unique_color('red','green','blue');
 my ($color, $rgb) = Color::Palette->unique_color('FF0000','FFFF00','0000FF');

 # Fetch color palette (chunk of static palette of given lenght)
 my @palette = Color::Palette->palette(10);

 ##
 # Or create instance and use object methods:
 ##

 my $instance = Color::Palette->new();

 my $rgb = $instance->as_rgb('snow');        # $rgb = 'FFFAFA';
 my $rgb = $instance->as_rgb('FFFAFA');      # $rgb = 'FFFAFA';

 my $color = $instance->as_string('FFFAFA'); # $color = 'snow';
 my $color = $instance->as_string('snow');   # $color = 'snow';

 my ($color, $rgb) = $instance->random_color();

 my ($color, $rgb) = $instance->random_unique_color();
 my ($color, $rgb) = $instance->random_unique_color('red','green','blue');
 my ($color, $rgb) = $instance->random_unique_color('FF0000','FFFF00','0000FF');

 my @palette = $instance->random_palette(10);

 my ($color, $rgb) = $instance->color();

 my ($color, $rgb) = $instance->unique_color();
 my ($color, $rgb) = $instance->unique_color('red','green','blue');
 my ($color, $rgb) = $instance->unique_color('FF0000','FFFF00','0000FF');

 my @palette = $instance->palette(10);

=head1 DESCRIPTION

Extensive module description.

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use List::Util 'shuffle';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

my %COLOR_PALETTE;		# Color palette ('color' => 'RGB')
my %COLOR_PALETTE_R;	# Reversed color palette ('RGB' => 'color')
my @COLORS_STR;			# List of all color names
my @COLORS_HEX;			# List of all color RGBs
my @COLORS_OPT;			# Optimal color palette

my $PALETTE_OPT = undef;
my $PALETTE_RND = undef;
my $PALETTE_UNQ = undef;

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION 	= '0.1';
    $DEVEL		= 0;

    %COLOR_PALETTE = (
		"aliceblue" 			=> "F0F8FF",
		"antiquewhite" 			=> "FAEBD7",
		"aqua" 					=> "00FFFF",
		"aquamarine" 			=> "7FFFD4",
		"azure" 				=> "F0FFFF",
		"beige" 				=> "F5F5DC",
		"bisque" 				=> "FFE4C4",
		"black" 				=> "000000",
		"blanchedalmond" 		=> "FFEBCD",
		"blue" 					=> "0000FF",
		"blueviolet" 			=> "8A2BE2",
		"brown" 				=> "A52A2A",
		"burlywood" 			=> "DEB887",
		"cadetblue" 			=> "5F9EA0",
		"chartreuse" 			=> "7FFF00",
		"chocolate" 			=> "D2691E",
		"coral" 				=> "FF7F50",
		"cornflowerblue" 		=> "6495ED",
		"cornsilk" 				=> "FFF8DC",
		"crimson" 				=> "DC143C",
		"cyan" 					=> "00FFFF",
		"darkblue" 				=> "00008B",
		"darkcyan" 				=> "008B8B",
		"darkgoldenrod" 		=> "B8860B",
		"darkgray" 				=> "A9A9A9",
		"darkgreen" 			=> "006400",
		"darkkhaki" 			=> "BDB76B",
		"darkmagenta" 			=> "8B008B",
		"darkolivegreen" 		=> "556B2F",
		"darkorange" 			=> "FF8C00",
		"darkorchid" 			=> "9932CC",
		"darkred" 				=> "8B0000",
		"darksalmon" 			=> "E9967A",
		"darkseagreen" 			=> "8FBC8F",
		"darkslateblue" 		=> "483D8B",
		"darkslategray" 		=> "2F4F4F",
		"darkturquoise" 		=> "00CED1",
		"darkviolet" 			=> "9400D3",
		"deeppink" 				=> "FF1493",
		"deepskyblue" 			=> "00BFFF",
		"dimgray" 				=> "696969",
		"dodgerblue" 			=> "1E90FF",
		"firebrick" 			=> "B22222",
		"floralwhite" 			=> "FFFAF0",
		"forestgreen" 			=> "228B22",
		"fuchsia" 				=> "FF00FF",
		"gainsboro" 			=> "DCDCDC",
		"ghostwhite" 			=> "F8F8FF",
		"gold" 					=> "FFD700",
		"goldenrod" 			=> "DAA520",
		"gray" 					=> "808080",
		"green" 				=> "008000",
		"greenyellow" 			=> "ADFF2F",
		"honeydew" 				=> "F0FFF0",
		"hotpink" 				=> "FF69B4",
		"indianred" 			=> "CD5C5C",
		"indigo" 				=> "4B0082",
		"ivory" 				=> "FFFFF0",
		"khaki" 				=> "F0E68C",
		"lavender" 				=> "E6E6FA",
		"lavenderblush" 		=> "FFF0F5",
		"lawngreen" 			=> "7CFC00",
		"lemonchiffon" 			=> "FFFACD",
		"lightblue" 			=> "ADD8E6",
		"lightcoral" 			=> "F08080",
		"lightcyan" 			=> "E0FFFF",
		"lightgoldenrodyellow" 	=> "FAFAD2",
		"lightgreen" 			=> "90EE90",
		"lightgrey" 			=> "D3D3D3",
		"lightpink" 			=> "FFB6C1",
		"lightsalmon" 			=> "FFA07A",
		"lightseagreen" 		=> "20B2AA",
		"lightskyblue" 			=> "87CEFA",
		"lightslategray" 		=> "778899",
		"lightsteelblue" 		=> "B0C4DE",
		"lightyellow" 			=> "FFFFE0",
		"lime" 					=> "00FF00",
		"limegreen" 			=> "32CD32",
		"linen" 				=> "FAF0E6",
		"magenta" 				=> "FF00FF",
		"maroon" 				=> "800000",
		"mediumauqamarine" 		=> "66CDAA",
		"mediumblue" 			=> "0000CD",
		"mediumorchid" 			=> "BA55D3",
		"mediumpurple" 			=> "9370D8",
		"mediumseagreen" 		=> "3CB371",
		"mediumslateblue" 		=> "7B68EE",
		"mediumspringgreen" 	=> "00FA9A",
		"mediumturquoise" 		=> "48D1CC",
		"mediumvioletred" 		=> "C71585",
		"midnightblue" 			=> "191970",
		"mintcream" 			=> "F5FFFA",
		"mistyrose" 			=> "FFE4E1",
		"moccasin" 				=> "FFE4B5",
		"navajowhite" 			=> "FFDEAD",
		"navy" 					=> "000080",
		"oldlace" 				=> "FDF5E6",
		"olive" 				=> "808000",
		"olivedrab" 			=> "688E23",
		"orange" 				=> "FFA500",
		"orangered" 			=> "FF4500",
		"orchid" 				=> "DA70D6",
		"palegoldenrod" 		=> "EEE8AA",
		"palegreen" 			=> "98FB98",
		"paleturquoise" 		=> "AFEEEE",
		"palevioletred" 		=> "D87093",
		"papayawhip" 			=> "FFEFD5",
		"peachpuff" 			=> "FFDAB9",
		"peru" 					=> "CD853F",
		"pink" 					=> "FFC0CB",
		"plum" 					=> "DDA0DD",
		"powderblue" 			=> "B0E0E6",
		"purple" 				=> "800080",
		"red" 					=> "FF0000",
		"rosybrown" 			=> "BC8F8F",
		"royalblue" 			=> "4169E1",
		"saddlebrown" 			=> "8B4513",
		"salmon" 				=> "FA8072",
		"sandybrown" 			=> "F4A460",
		"seagreen" 				=> "2E8B57",
		"seashell" 				=> "FFF5EE",
		"sienna" 				=> "A0522D",
		"silver" 				=> "C0C0C0",
		"skyblue" 				=> "87CEEB",
		"slateblue" 			=> "6A5ACD",
		"slategray" 			=> "708090",
		"snow" 					=> "FFFAFA",
		"springgreen" 			=> "00FF7F",
		"steelblue" 			=> "4682B4",
		"tan" 					=> "D2B48C",
		"teal" 					=> "008080",
		"thistle" 				=> "D8BFD8",
		"tomato" 				=> "FF6347",
		"turquoise" 			=> "40E0D0",
		"violet" 				=> "EE82EE",
		"wheat" 				=> "F5DEB3",
		"white" 				=> "FFFFFF",
		"whitesmoke" 			=> "F5F5F5",
		"yellow" 				=> "FFFF00",
		"yellowGreen"			=> "9ACD32",
	);

	%COLOR_PALETTE_R = reverse(%COLOR_PALETTE);

	@COLORS_STR = keys(%COLOR_PALETTE);
	@COLORS_HEX = keys(%COLOR_PALETTE_R);
	@COLORS_OPT = (
		"red",
		"yellow",
		"blue",
		"green",
		"midnightblue",
		"salmon",
		"orange",
		"gray",
		"lightblue",
		"peachpuff",
		"indigo",

		"aliceblue",
		"antiquewhite",
		"aqua",
		"aquamarine",
		"azure",
		"beige",
		"bisque",
		"black",
		"blanchedalmond",
		"blueviolet",
		"brown",
		"burlywood",
		"cadetblue",
		"chartreuse",
		"chocolate",
		"coral",
		"cornflowerblue",
		"cornsilk",
		"crimson",
		"cyan",
		"darkblue",
		"darkcyan",
		"darkgoldenrod",
		"darkgray",
		"darkgreen",
		"darkkhaki",
		"darkmagenta",
		"darkolivegreen",
		"darkorange",
		"darkorchid",
		"darkred",
		"darksalmon",
		"darkseagreen",
		"darkslateblue",
		"darkslategray",
		"darkturquoise",
		"darkviolet",
		"deeppink",
		"deepskyblue",
		"dimgray",
		"dodgerblue",
		"firebrick",
		"floralwhite",
		"forestgreen",
		"fuchsia",
		"gainsboro",
		"ghostwhite",
		"gold",
		"goldenrod",
		"greenyellow",
		"honeydew",
		"hotpink",
		"indianred",
		"ivory",
		"khaki",
		"lavender",
		"lavenderblush",
		"lawngreen",
		"lemonchiffon",
		"lightcoral",
		"lightcyan",
		"lightgoldenrodyellow",
		"lightgreen",
		"lightgrey",
		"lightpink",
		"lightsalmon",
		"lightseagreen",
		"lightskyblue",
		"lightslategray",
		"lightsteelblue",
		"lightyellow",
		"lime",
		"limegreen",
		"linen",
		"magenta",
		"maroon",
		"mediumauqamarine",
		"mediumblue",
		"mediumorchid",
		"mediumpurple",
		"mediumseagreen",
		"mediumslateblue",
		"mediumspringgreen",
		"mediumturquoise",
		"mediumvioletred",
		"mintcream",
		"mistyrose",
		"moccasin",
		"navajowhite",
		"navy",
		"oldlace",
		"olive",
		"olivedrab",
		"orangered",
		"orchid",
		"palegoldenrod",
		"palegreen",
		"paleturquoise",
		"palevioletred",
		"papayawhip",
		"peru",
		"pink",
		"plum",
		"powderblue",
		"purple",
		"rosybrown",
		"royalblue",
		"saddlebrown",
		"sandybrown",
		"seagreen",
		"seashell",
		"sienna",
		"silver",
		"skyblue",
		"slateblue",
		"slategray",
		"snow",
		"springgreen",
		"steelblue",
		"tan",
		"teal",
		"thistle",
		"tomato",
		"turquoise",
		"violet",
		"wheat",
		"white",
		"whitesmoke",
		"yellow",
		"yellowGreen",
	);
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC ,STATIC]

  Usage     : my $instance = Color::Pallette->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Color::Pallette reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
	my $self = bless ({}, $class);
	return $self->_init(@_);
}

=item as_rgb($) [PUBLIC, HYBRID]

  Usage     : my $rgb = $palette->as_rgb($color); OR
              my $rgb = Color::Palette->as_rgb($color);
  Purpose   : Translate color name into RGB hexadecimal representation
  Arguments : STRING $color - name of the color
  Returns   : STRING $rgb - color in RGB format (6 hexadecimal characters)
  Throws    : Croaks, if not given color as argument
  Comments  : Hybrid method, may be called as class, or instance method

=cut

sub as_rgb($)
{
	my $invocant = shift;
	my ($color,) = @_;
	croak ((caller(0))[3] . ": color must be given as argument") unless $color;

	# Perhaps the color is already in hex format
	return $color if $COLOR_PALETTE_R{$color};

	# Otherwise attempt to translate string to hex
	return $COLOR_PALETTE{$color};
}

=item as_string($) [PUBLIC, HYBRID]

  Usage     : my ($color, $rgb) = $palette->as_string($color); OR
              my ($color, $rgb) = Color::Palette->as_string($color);
  Purpose   : Translate RGB hexadecimal representation into color name
  Arguments : STRING $rgb - color in RGB format (6 hexadecimal characters) [MANDATORY]
  Returns   : STRING $color - name of the color
  Throws    : Croaks, if not given color as argument
  Comments  : Hybrid method, may be called as class, or instance method

=cut

sub as_string($)
{
	my $invocant = shift;
	my ($color,) = @_;
	croak ((caller(0))[3] . ": color must be given as argument") unless $color;

	# Perhaps the color is already in string format
	return $color if $COLOR_PALETTE{$color};

	# Otherwise attempt to translate hex to string
	return $COLOR_PALETTE_R{$color};
}

=item color() [PUBLIC, HYBRID]

  Usage     : my ($color, $rgb) = $palette->color(); OR
              my ($color, $rgb) = Color::Palette->color();
  Purpose   : Fetch next color from static color palette
  Arguments : None
  Returns   : STRING $color - name of the color
              STRING $rgb - color in RGB format (6 hexadecimal characters)
  Throws    : Nothing
  Comments  : Hybrid method, may be called as class, or instance method. After
              fetching all colors starts from beginning

=cut

sub color()
{
	my $invocant = shift;
	my $palette;

	# Fetch or initialize the palette either from the instance, or class storage
	if (blessed($invocant)) {
		$invocant->{PALETTE_OPT} = [@COLORS_OPT] unless $invocant->{PALETTE_OPT} and scalar(@{$invocant->{PALETTE_OPT}});
		$palette = $invocant->{PALETTE_OPT};
	}
	else {
		$PALETTE_OPT = [@COLORS_OPT] unless $PALETTE_OPT and scalar(@$PALETTE_OPT);
		$palette = $PALETTE_OPT;
	}

	# Pick the next color in the palette
	my $color = shift(@$palette);

	return ($color, $COLOR_PALETTE{$color});
}

=item next_color(;@) [PUBLIC]

  Usage     : my ($color, $rgb) = $palette->next_color(@list); OR
              my ($color, $rgb) = Color::Palette->next_color(@list);
  Purpose   : Fetch next unique color from static palette
  Arguments : ARRAY OF STRINGS @forbidden_colors - list of forbidden colors [OPTIONAL]
  Returns   : STRING $color - name of the color
              STRING $rgb - color in RGB format (6 hexadecimal characters)
  Throws    : Nothing
  Comments  : Hybrid method, may be called as class, or instance method

=cut

sub next_color(;@)
{
	my $invocant = shift;
	my @forbidden_colors = @_ or ();

	# Some colors may be designated as forbidden
	my %forbidden_colors;
	map { $forbidden_colors{$invocant->as_string($_)} = 1; } @forbidden_colors;

	# Fetch the next unique optimal color
	foreach my $color (@COLORS_OPT) {
		return ($color, $COLOR_PALETTE{$color}) unless exists $forbidden_colors{$color};
	}

	return (undef,undef);
}

=item random_color() [PUBLIC, HYBRID]

  Usage     : my ($color, $rgb) = $palette->random_color(); OR
              my ($color, $rgb) = Color::Palette->random_color();
  Purpose   : Fetch random color from color palette
  Arguments : None
  Returns   : STRING $color - name of the color
              STRING $rgb - color in RGB format (6 hexadecimal characters)
  Throws    : Nothing
  Comments  : Hybrid method, may be called as class, or instance method

=cut

sub random_color()
{
	my $invocant = shift;

	# Generate random color
	my $color = $COLORS_STR[int(rand(@COLORS_STR))];

	return ($color, $COLOR_PALETTE{$color});
}

=item unique_color(;@) [PUBLIC]

  Usage     : my ($color, $rgb) = $palette->unique_color(@list); OR
              my ($color, $rgb) = Color::Palette->unique_color(@list);
  Purpose   : Fetch next unique color from static palette
  Arguments : ARRAY OF STRINGS @forbidden_colors - list of forbidden colors [OPTIONAL]
  Returns   : STRING $color - name of the color
              STRING $rgb - color in RGB format (6 hexadecimal characters)
  Throws    : Nothing
  Comments  : Hybrid method, may be called as class, or instance method

=cut

sub unique_color(;@)
{
	my $invocant = shift;
	my @forbidden_colors = @_ or ();

	# Some colors may be designated as forbidden
	my %forbidden_colors;
	map { $forbidden_colors{$invocant->as_string($_)} = 1; } @forbidden_colors;

	# Fetch or initialize the palette either from the instance, or class storage
	my $palette;
	if (blessed($invocant)) {
		$invocant->{PALETTE_UNQ} = [@COLORS_OPT] unless $invocant->{PALETTE_UNQ} and scalar(@{$invocant->{PALETTE_UNQ}});
		$palette = $invocant->{PALETTE_UNQ};
	}
	else {
		$PALETTE_UNQ = [@COLORS_OPT] unless $PALETTE_UNQ and scalar(@$PALETTE_UNQ);
		$palette = $PALETTE_UNQ;
	}

	while (my $color = shift(@$palette)) {
		return ($color, $COLOR_PALETTE{$color}) unless exists $forbidden_colors{$color};
	}

	return (undef,undef);
}

=item random_unique_color(;@) [PUBLIC]

  Usage     : my ($color, $rgb) = $palette->random_unique_color(@list); OR
              my ($color, $rgb) = Color::Palette->random_unique_color(@list);
  Purpose   : Fetch random and unique color
  Arguments : ARRAY OF STRINGS @forbidden_colors - list of forbidden colors [OPTIONAL]
  Returns   : STRING $color - name of the color
              STRING $rgb - color in RGB format (6 hexadecimal characters)
  Throws    : Nothing
  Comments  : Hybrid method, may be called as class, or instance method

=cut

sub random_unique_color(;@)
{
	my $invocant = shift;
	my @forbidden_colors = @_ or ();

	# Some colors may be designated as forbidden
	my %forbidden_colors;
	map { $forbidden_colors{$invocant->as_string($_)} = 1; } @forbidden_colors;

	# Fetch or initialize the palette either from the instance, or class storage
	my $palette;
	if (blessed($invocant)) {
		$invocant->{PALETTE_RND} = [shuffle(@COLORS_OPT)] unless $invocant->{PALETTE_RND} and scalar(@{$invocant->{PALETTE_RND}});
		$palette = $invocant->{PALETTE_RND};
	}
	else {
		$PALETTE_RND = [shuffle(@COLORS_OPT)] unless $PALETTE_RND and scalar(@$PALETTE_RND);
		$palette = $PALETTE_RND;
	}

	while (my $color = shift(@$palette)) {
		return ($color, $COLOR_PALETTE{$color}) unless exists $forbidden_colors{$color};
	}

	return (undef,undef);
}

=item palette($) [PUBLIC, HYBRID]

  Usage     : my @palette = $palette->palette($size); OR
              my @palette = Color::Palette->palette($size);
  Purpose   : Fetch static color palette of given size
  Arguments : INTEGER $size - number of colors in the palette [MANDATORY]
  Returns   : ARRAY of ARRAYS containing colors:
                STRING $color - name of the color
                STRING $rgb - color in RGB format (6 hexadecimal characters)
  Throws    : Croaks, if not given palette size as argument
              Croaks, if requested palette size is too big
  Comments  : Hybrid method, may be called as class, or instance method

=cut

sub palette($)
{
	my $invocant = shift;
	my ($size,) = @_;
	croak ((caller(0))[3] . ": color palette size must be given as argument") unless ($size and (int($size) > 0));
	croak ((caller(0))[3] . ": requested palette size is too big") if int($size) > scalar(@COLORS_OPT);

	my @slice = @COLORS_OPT[0 .. --$size];
	my @palette = ();

	map { push(@palette, [$_, $COLOR_PALETTE{$_}]); } @slice;

	# Otherwise attempt to translate string to hex
	return @palette;
}

=item random_palette($) [PUBLIC, HYBRID]

  Usage     : my @palette = $palette->random_palette($size); OR
              my @palette = Color::Palette->random_palette($size);
  Purpose   : Fetch random color palette of given size
  Arguments : INTEGER $size - number of colors in the palette [MANDATORY]
  Returns   : ARRAY of ARRAYS containing colors:
                STRING $color - name of the color
                STRING $rgb - color in RGB format (6 hexadecimal characters)
  Throws    : Croaks, if not given palette size as argument
              Croaks, if requested palette size is too big
  Comments  : Hybrid method, may be called as class, or instance method

=cut

sub random_palette($)
{
	my $invocant = shift;
	my ($size,) = @_;
	croak ((caller(0))[3] . ": color palette size must be given as argument") unless ($size and (int($size) > 0));
	croak ((caller(0))[3] . ": requested palette size is too big") if int($size) > scalar(@COLORS_OPT);

	# Generate random color palette
	my @slice = (shuffle(@COLORS_STR))[0 .. --$size];
	my @palette = ();

	map { push(@palette, [$_, $COLOR_PALETTE{$_}]); } @slice;

	# Otherwise attempt to translate string to hex
	return @palette;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Color::Pallette instance
#  Arguments :
#  Returns   : Color::Pallette reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

	return $self;
}

=pod

=back

=cut

1;
