package Config::JSON;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Config::JSON - Simple, but powerful config file parser

=head1 SYNOPSIS

 use Config::JSON;

 my $parser = Config::JSON->new($params);

 # Attempt to parse and get the configurations
 my ($config, $error) = $parser->config("/config/file/name.cfg", "/schema/file/name.cfg");

 # Shortcut via static class method
 my ($config, $error) = Config::JSON->load("/config/file/name.cfg", "/schema/file/name.cfg");

=head1 DESCRIPTION

This is the implementation of simple JSON config file parser with very easy
to use syntax.

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Config::Validator(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Storable qw(freeze);
use Digest::MD5 qw(md5_hex);

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;

#-- Custom application modules ------------------------------------------------#
use JSON::Schema::Validator;

#-- Module initializations ----------------------------------------------------#
BEGIN {
	use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item load($;@) [PUBLIC,STATIC]

  Usage     : my $config = Config::JSON->load($config_file, $schema_file);
  Purpose   : Load configurations from given JSON file
  Returns   : Same as config() method
  Arguments : Same as config() method
  Throws    : Croaks, if invoked on object
              Croaks, if given bogus parameters
  Comments  : Internally creates parser instance and uses config() method
  See Also  : config() method

=cut

sub load($;@)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my ($cfg_file, $schema, @params) = @_;

	# Create private instance of a parser and try to parse the given file
	my $parser = $class->new($schema, @params);
	return $parser->config($cfg_file, $schema);
}

=item new(;$) [PUBLIC,STATIC]

  Usage     : my $parser = Config::JSON->new($params);
  Purpose   : Create new instance of a configuration parser
  Returns   : Config::JSON reference to new configuration parser
  Arguments : All arguments are passed to the _init() method
  Throws    : Croaks, if invoked on object
              Croaks, if given bogus parameters
  Comments  : Internally uses _init_params() method
  See Also  : _init_params() method

=cut

sub new(;$)
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my $self = {};
	bless($self, $class);
	return $self->_init(@_);
}

=item config($;$) [PUBLIC]

  Usage     : my $config = $parser->config("/config/file/name.json", "/schema/file/name.json");
  Purpose   : Parse given JSON config file and return configurations in a hash reference
  Arguments : STRING $cfg_file - config file name [MANDATORY]
              STRING $schema - JSON schema [OPTIONAL]
  Returns   : ARRAY containing:
				(REFERENCE containing all configurations, UNDEF) on SUCCESS
				(UNDEF, STRING description of error) on FAILURE
  Throws    : Croaks, if invoked on object
              Croaks, if given bogus config file
              Croaks, if given invalid parameters

=cut

sub config($;$)
{
	my $self = shift;
	confess "Instance method not invoked on object" unless blessed($self);
	my ($cfg_file, $schema) = @_;

	# Perform some sanity checks first
	die "Name of the config file name must be given as argument\n" unless $cfg_file;
	die "Config file '$cfg_file' is not ordinary file\n" unless -f $cfg_file;

	my ($object, $validator, $error);
	eval {
		# Parse the object from given JSON file
		$object = $self->_load_json_file($cfg_file);

		# Schema validation is optional
		if ($schema) {
			$validator = $self->_validator($schema);
			$error = $validator->validate($object);
			$object = undef if $error;
		}
	};
	if ($@) {
		$error = $@ ;
	}
	return ($object, $error);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Config::Validator instance
#  Arguments : STRING $schema - JSON schema [OPTIONAL]
#  Returns   : Config::JSON reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init(;$)
{
	my $self = shift;
	confess "Instance method not invoked on object" unless blessed($self);
	my ($schema, %params) = @_;

	# Prepare internal storage for schema validators
	$self->{VALIDATORS} = {};

	# Create the validator instance in advance
	$self->_validator($schema) if ($schema);

	return $self;
}

# _load_json_file($) [PROTECTED]
#
#  Usage     : my $json = $self->_load_json_file($file);
#  Purpose   : Load JSON structure from external file
#  Arguments : STRING $file - name of the JSON file
#  Returns   : REFERENCE $json

sub _load_json_file($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($file,) = @_;
	confess "JSON file must be given as argument" unless $file;
	confess "JSON file must be ordinary file" unless -f $file;

	# Load JSON string from external file
	open (my $fh, $file) or die "Can`t open JSON file '$file'\n";
	my $json;
	while (my $line = <$fh>) {
		# Custom feature - remove comments
		next if $line =~ /^\s*(#|;|\/\/)/;
		$json .= $line;
	}
	close($fh);

	# Untaint the loaded string
	if ($json =~ /^(.+)$/s) { $json = $1; }

	# Parse out the JSON object and return it
	return decode_json($json);
}

# _validator($) [PROTECTED]
#
#  Usage     : my $validator = $self->_validator($schema);
#  Purpose   :
#  Arguments :
#  Returns   :

sub _validator($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($schema,) = @_;

	my $key = $schema;
	$key = freeze $schema if ref $schema;
	$key = md5_hex("$key");

	unless ($self->{VALIDATORS}->{$key}) {
		$self->{VALIDATORS}->{$key} = JSON::Schema::Validator->new($schema);
	}
	return $self->{VALIDATORS}->{$key};
}

=pod

=back

=cut

1;
