package Config::Parser;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Config::Parser - Simple, but powerful config file parser

=head1 SYNOPSIS

 use Config::Parser;

 my $parser = Config::Parser->new($params);

 # Attempt to parse and get the configurations
 my $config = $parser->config("/config/file/name.cfg");
 if ($config) {
	... (process the configuration somehow)
 }
 else {
 	my $errors = $parser->get_errors();
 }

 # Shortcut via static class method
 my $config = Config::Parser->load("/config/file/name.cfg", $params);
 if ($config) {
 	... (process the configuration somehow)
 }
 else {
	my $errors = Config::Parser->get_errors();
 }

 # Only parse the config file and check for errors
 my $errors = $parser->parse("/config/file/name.cfg");

=head1 DESCRIPTION

This is the implementation of simple config file parser with very easy
to use syntax.

=head1 USAGE

Basic usage was described in the SYNOPSIS section. Howewer, there are advanced
means of use of this module, which give the user more control about how the configurations
are parsed fromthe file and storeg in the resulting hash.

Constructor method new() and static shortcut method load() may receive optional
$param argument. This argument must be hash reference, which may contain following
keys:

=over

=item B<cb_directive_validator>

	- function reference to the callback, which will be called for each directive
	  found in the configuration file
	- this callback may perform some additional validation of the directive
	  key and value
	- this callback will receive following arguments:
		($section_name, $section_id, $directive, $value)
	- this callback should return following values:
		($error, $errmsg)
	- $error = 0 means success, 1 failure. $errmsg is optional, if not given,
	  error message will be generated automatically

=item B<cb_directive_processor>

	- function reference to the callback, which will be called for each directive
	  found in the configuration file
	- this callback may perform some additional processing with the directive
	  key and value, even altering them
	- this callback will receive following arguments:
		($section_name, $section_id, $directive, $value)
	- this callback should return following values:
		($directive, $value, $error, $errmsg)
	- for the explanation of error reporting mechanism, see cb_directive_validator above

=item B<cb_section_validator>

	- function reference to the callback, which will be called for each section
	  found in the configuration file
	- this callback may perform some additional validation of the section
	  name and id
	- this callback will receive following arguments:
		($section_name, $section_id)
	- this callback should return following values:
		($error, $errmsg)
	- for the explanation of error reporting mechanism, see cb_directive_validator above

=item B<cb_section_processor>

	- function reference to the callback, which will be called for each section
	  found in the configuration file
	- this callback may perform some additional processing with the section
	  name and id, even altering them
	- this callback will receive following arguments:
		($section_name, $section_id)
	- this callback should return following values:
		($section_name, $section_id, $error, $errmsg)
	- for the explanation of error reporting mechanism, see cb_directive_validator above

=item B<cb_postprocessing>

	- function reference to the callback, which will be called after the file is
	  successfully parsed
	- this callback may perform some additional processing with the whole
	  configuration hash, even altering it
	- this callback will receive following arguments:
		($configs)
	- this callback should return following values:
		($configs)

=item B<stack_values>

	- by default, if there are multiple occurences of the same directive key in
	  the configuration file, value of the latter will overwrite the former one
	- this option will alter this default behaviour - values will be stored in the
	  array, even if there is only one occurence of that directive key
	- this parameter may be either of type boolean TRUE (1|y|yes|true), in that
	  case values of all directive keys will be stored in array
	- or this parameter may be [,;\t ]+ separated string, in that case only
	  values for that particular directive keys will be stored in array
	  (e.g. string 'config1, config2' will make the values of directives with
	  keys 'config1' and 'config2' go to the array)

=item B<compress>

	- by default, resulting hash has the following structure:
		{'section_a' => {'section_a_id1' => {'key1' => 'value1'}}}
	- this default structure is very complex for some simple config files, because
	  they are not using all syntax features (sections, section identifiers, hierarchical keys)
	- this optional parameter performs following simplifications:
		- remove all empty subsections and sections
		- if the section has only one default instance, move configurations one layer up for that section
		- if there is only default section on top level, move the configurations one layer up
	- this parameter must be positive INTEGER (1|2)

=item B<unique_sections>

	- if this parameter is provided, each section and section ID may occur only once
	  in the configuration file (with exception of the default section)
	- this parameter must be of type boolean TRUE (1|y|yes|true)

=item B<structure>

	- this parameter turns on the structure checking for the configuration file
	- this parameter must be a hash reference with following structure:
		{'section_name' => {'possible_key1' => datatype_specification,
							'possible_key2' => datatype_specification,
						}
		}
	- datatype_specification - specification of the configuration value datatype,
	  see the Config::Validator(3) manual page for details about the specification
	  syntax

=back

Here is the example of the advanced usage:

 # Define optional callbacks

 # Directive validator callback is called for each DIRECTIVE (key = value) in the config file
 sub directive_validator($$$$)
 {
	my ($section_name, $section_id, $directive, $value) = @_;
	my ($error, $errmsg) = (0, undef);

	...
	Perform some validation here, return $error = 0 for success, 1 for failure.
	$errmsg is optional, if not given, error message will be generated.
	...

	return ($error, $errmsg);
 }

 # Section validator callback is called for each SECTION defined in the config file
 sub section_validator($$)
 {
	my ($section_name, $section_id) = @_;
	my ($error, $errmsg) = (0, undef);

	...
	Perform some validation here, return $error = 0 for success, 1 for failure.
	$errmsg is optional, if not given, error message will be generated.
	...

	return ($error, $errmsg);
 }

 # Directive processor callback is called after successfull DIRECTIVE validation, it may be used
 # to somehow further process the directive KEY, or VALUE.
 sub directive_processor($$$$)
 {
	my ($section_name, $section_id, $directive, $value) = @_;
	my ($error, $errmsg) = (0, undef);

	...
	Perform some additional processing of the directive KEY or VALUE, the mechanism
	$error and $errmsg used in validators is used here for reporting errors as well.
	...

	return ($directive, $value, $error, $errmsg);
 }

 # Section processor callback is called after successfull SECTION validation, it may be used
 # to somehow further process the section NAME or ID.
 sub section_processor($$)
 {
	my ($section_name, $section_id) = @_;
	my ($error, $errmsg) = (0, undef);

	...
	Perform some additional processing of the section NAME or ID, the mechanism
	$error and $errmsg used in validators is used here for reporting errors as well.
	...

	return ($section_name, $section_id, $error, $errmsg);
 }

 # Postprocessing callback is called before returning the configuration to the caller.
 sub postprocessing($)
 {
	my $CONFIGS = shift;

	...
	Perform some postprocessing of the configurations
	...

	return $CONFIGS;
 }

 my $params = {
		'cb_directive_validator'	=> \&directive_validator,
		'cb_section_validator'		=> \&section_validator,
		'cb_directive_processor'	=> \&directive_processor,
		'cb_section_processor'		=> \&section_processor,
		'cb_postprocessing'			=> \&postprocessing,
		'stack_values'				=> "config1 config2",
		'compress'					=> 1,
		'unique_sections'			=> 1,
		'structure'					=> {__general__	=> {'config1' => 'int:5:10'},
										'section'	=> {'config2' => 'int::1000',
														'config3' => 'string',
														'config4' => 'bool',
														'config5' => 'hash:a,b,c'
													   },
										},
			};

 my $parser = Config::Parser->new($params);

 # Attempt to parse and get the configurations
 my $config = $parser->config("/config/file/name.cfg");
 if ($config) {
	... (process the configuration somehow)
 }
 else {
 	my $errors = $parser->get_errors();
 }

 # Shortcut via static class method
 my $config = Config::Parser->load("/config/file/name.cfg", $params);
 if ($config) {
 	... (process the configuration somehow)
 }
 else {
	my $errors = Config::Parser->get_errors();
 }

 # Only parse the config file and check for errors
 my $errors = $parser->parse("/config/file/name.cfg");

=head1 CONFIG FILE SYNTAX

 #--------------------
 # Example config file
 #--------------------

 # General configurations

 config1            = value1
 config2.subconfig1 = value2
 config2.subconfig2 = value3

 # Define section
 [logging]

 target = logging_target

 # Define section with id
 [host example.host.tld]

 config1 = updated_value1

 # Include sub-config file
 {include /path/to/other/config/file.cfg}

 # Define section with another id
 [host example.anotherhost.tld]

 config3 = updated_value1
 config4 = a,b,c,d
 config5 = a=>b,c=>d,e=>f

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Config::Validator(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Storable qw(dclone);

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Utils::String;
use Config::Validator;

#-- Module initializations ----------------------------------------------------#
BEGIN {
	use vars qw($VERSION $DEVEL $LAST_INSTANCE @DATA_TYPES);
    $VERSION = '0.1';
    $DEVEL   = 0;

	# Storage for last created parser instance
	$LAST_INSTANCE = undef;

	# List of valid datatypes
	@DATA_TYPES = Config::Validator->valid_datatypes();
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

# Name of the default section in the config file
use constant DFLT_SECTION_NAME	=> "__general__";

# Identifier of the default section in the config file
use constant DFLT_SECTION_ID	=> "__undef__";

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# List of names of all validator and processor callbacks
my @CALLBACK_NAMES 	= (
		'cb_directive_validator',
		'cb_section_validator',
		'cb_directive_processor',
		'cb_section_processor',
		'cb_postprocessing',
	);
my %CALLBACK_NAMES_SH = map { $_ => 1 } @CALLBACK_NAMES;

# List of names of all boolean operation flags
my @FLAG_NAMES = ('unique_sections');
my %FLAG_NAMES_SH = map { $_ => 1 } @FLAG_NAMES;

# List of other processing parameters
my @PARAMETERS = ('compress');
my %PARAMETERS_SH = map { $_ => 1 } @PARAMETERS;

# RE pattern chunks
my $RE_CHUNK_KEY = '[_a-zA-Z][-_a-zA-Z0-9]*';
my $RE_CHUNK_ID	 = '[-:@._a-zA-Z0-9]+';

# Compiled RE patterns for config file analysis
my $RE_PTRN_COMMENT    = qr/^[#;].*/;
my $RE_PTRN_EMPTY_LINE = qr/^\s*$/;
my $RE_PTRN_INCLUDE    = qr/^\{include\s+(.+)\}$/;
my $RE_PTRN_SECTION    = qr/^\[($RE_CHUNK_KEY)(?:\s+($RE_CHUNK_ID))?\]$/;
my $RE_PTRN_DIRECTIVE  = qr/^($RE_CHUNK_KEY(?:\.$RE_CHUNK_KEY)*)\s*=\s*(.+)$/;
my $RE_PTRN_TRUE       = qr/^((?i)1|y|yes|t|true|on)$/;
my $RE_PTRN_FALSE      = qr/^((?i)0|n|no|f|false|off)$/;

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item load($;@) [PUBLIC,STATIC]

  Usage     : my $config = Config::Parser->load($config_file_name, $params);
  Purpose   : Load configurations from given file
  Returns   : Hash reference containing all configurations (SUCCESS)
              undef (FAILURE)
  Arguments : string $cfg_file - name of the config file to parse
              array @params - optional parameters, will be passed to the new() method
  Throws    : Croaks, if invoked on object
              Croaks, if given bogus config file
              Croaks, if given bogus parameters
  Comments  : Internally creates parser instance and uses config() method
  See Also  : config() method

=cut

sub load($;@)
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
	my ($cfg_file, @params) = @_;

	# Create private instance of a parser and try to parse the given file
	my $parser = $class->new(@params);
	return $parser->config($cfg_file);
}

=item new(;$) [PUBLIC,STATIC]

  Usage     : my $parser = Config::Parser->new($params);
  Purpose   : Create new instance of a configuration parser
  Returns   : Config::Parser reference to new configuration parser
  Arguments : All arguments are passed to the _init() method
  Throws    : Croaks, if invoked on object
              Croaks, if given bogus parameters
  Comments  : Internally uses _init_params() method
  See Also  : _init_params() method

=cut

sub new(;$)
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
	my $self = {};
	bless($self, $class);

	# Store the reference to the last instance
	$LAST_INSTANCE = $self;

	# Initialize only processing parameters, inner storage variables related
	# to the parsing will be initialized separately before the parsing. This
	# way one instance can be reused multiple times (in case the parsing
	# parameters are the same).
	return $self->_init(@_);
}

=item config($) [PUBLIC]

  Usage     : my $config = $parser->config("/config/file/name.cfg");
  Purpose   : Parse given config file and return configurations in a hash reference
  Returns   : HASH REFERENCE containing all configurations (SUCCESS)
              undef (FAILURE)
  Arguments : Same as parse() method
  Throws    : Same as parse() method
  Comments  : Internally uses parse() method
  See Also  : parse() method

=cut

sub config($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

	# Parse the given file, throw away array reference containing the errors
	$self->parse(@_);

	# If any errors occured during the parsing process, return UNDEF
	if ($self->_any_errors()) { return undef; }

	# Otherwise return the successfully parsed configurations
	return $self->_get_config();
}

=item parse($) [PUBLIC]

  Usage     : my $errors = $parser->parse("/config/file/name.cfg");
  Purpose   : Parse given config file and return all errors, that have occured
  Returns   : ARRAY REFERENCE containing all errors as strings
  Arguments : string $cfg_file - config file name
  Throws    : Croaks, if invoked on object
              Croaks, if given bogus config file
              Croaks, if given invalid parameters
  Comments  : Internally used by method config().
              This method may be used separately for syntax validation.
  See Also  : config() method

=cut

sub parse($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($cfg_file,) = @_;

	# Perform some sanity checks first
	croak ((caller(0))[3] . ": you have to provide config file name\n") unless $cfg_file;
	croak ((caller(0))[3] . ": config file '$cfg_file' is not ordinary file\n") unless -f $cfg_file;

	# Initialize inner storage variables related to parsing
	$self->_init_before_parsing();

	# Try to parse the given file
	$self->_parse_config_file($cfg_file);

	# And return any errors, that might have occured
	return $self->get_errors();
}

=item get_errors() [PUBLIC]

  Usage     : my $errors = $parser->get_errors() or
              my $errors = Config::Parser->get_errors()
  Purpose   : Get descriptions of all errors, that occured during the parsing of config file so far
  Returns   : ARRAY REFERENCE containing all errors as strings
  Arguments : None
  Throws    : Croaks, if invoked on class
              Croaks, if invoked before any configuration file was parsed

=cut

sub get_errors()
{
	my $invocant = shift;

	# If the method is invoked as class method, invoke it on last created instance
	unless (ref $invocant) {
		croak ((caller(0))[3] . ": method invoked too soon, no parser has been created yet") unless $LAST_INSTANCE;
		return $LAST_INSTANCE->get_errors() unless ref $invocant;
	}

	croak ((caller(0))[3] . ": method invoked too soon, no configuration file has been parsed yet") unless exists($invocant->{ERRORS});
	return $invocant->{ERRORS};
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Config::Validator instance
#  Arguments : None
#  Returns   : Config::Parser reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init(;$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($params,) = @_;

	# Optional parameters, that might be given
	if ($params) {
		croak ((caller(0))[3] . ": optional parameters must be given in hash reference\n") unless ref $params eq 'HASH';
		foreach my $param (keys %{$params}) {
			# Optional validator and processing callbacks
			if (exists $CALLBACK_NAMES_SH{lc($param)}) {
				croak ((caller(0))[3] . ": '$param' callback must be function reference\n") unless ref $params->{$param} eq 'CODE';
				$self->{uc($param)} = $params->{$param};
			}
			# Optional bool operation flags
			elsif (exists $FLAG_NAMES_SH{lc($param)}) {
				# All bool operation flags are off by default, they can be only turned on
				if ($params->{$param} =~ /$RE_PTRN_TRUE/) {
					$self->{uc($param)} = 1;
				}
				elsif ($params->{$param} =~ /$RE_PTRN_FALSE/) {
					$self->{uc($param)} = 0;
				}
				else {
					croak ((caller(0))[3] . ": invalid boolean value for '$param'\n");
				}
			}
			# Optional operation parameters
			elsif (exists $PARAMETERS_SH{lc($param)}) {
				$self->{uc($param)} = $params->{$param};
			}
			# Optional config structure definition
			elsif (lc($param) eq 'structure') {
				$self->_validate_param_structure($params->{$param});
				$self->{uc($param)} = $params->{$param};
			}
			# Optional stack values parameter
			elsif (lc($param) eq 'stack_values') {
				# If the parameter is BOOL true, values of all directive keys will be stored into array
				if ($params->{$param} =~ /$RE_PTRN_TRUE/) {
					$self->{uc($param)} = 1;
				}
				# Otherwise only values of explicitly given directive keys will be stored into array
				else {
					my @keys = split(/[,;\t ]+/, $params->{$param});
					my %sh = map { $_ => 1 } @keys;
					$self->{uc($param)} = \%sh;
				}
			}
			else {
				croak ((caller(0))[3] . ": unknown parameter '$param'\n");
			}
		}
	}

	return $self;
}

# _validate_param_structure($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _validate_param_structure($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($structure,) = @_;

	# PARAM VALIDATION: Config structure must be passed as not empty hash reference
	croak ((caller(0))[3] . ": config structure must be given in hash reference\n") unless ref $structure eq 'HASH';
	croak ((caller(0))[3] . ": config structure must not be empty\n") unless scalar(keys(%{$structure}));
	map {
		# Keys of the outer hash are 'section' names, values must be not empty hash references as well
		croak ((caller(0))[3] . ": config section '$_' structure must be given in hash reference\n") unless ref $structure->{$_} eq 'HASH';
		my $section = $_;
		map {
			# Keys of the inner hash are 'directive' key names, values are the 'datatypes' of the directives
			croak ((caller(0))[3] . ": invalid datatype '$structure->{$section}->{$_}' for directive key '$_' in config section '$section', possible values are: @DATA_TYPES\n") unless Config::Validator->is_valid_datatype($structure->{$section}->{$_});
		} keys(%{$structure->{$_}});
	} keys(%{$structure});
}

# _init_before_parsing() [PROTECTED]
#
#  Usage     :
#  Purpose   : Initialize instance variables related to parsing before parsing actually
#              happens. This way one instance may be reused multiple times.
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _init_before_parsing()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

	# General initializations
	$self->{CONFIG}			= {};	# Parsed configurations
	$self->{ERRORS}			= [];	# Errors, that occured during the parsing
	$self->{CFG_FILES_HND}	= [];	# Handles of currently opened config files
	$self->{CFG_FILES_CNT}	= [];	# Line counter for current config file
	$self->{CFG_FILES_CUR}	= [];	# Name of current config file
	$self->{CFG_FILES_INC}	= {};	# Hash of all already included files

	return $self;
}

# _any_errors() [PROTECTED]
#
#  Usage     :
#  Purpose   : Check, if any errors occured during the parsing of config file so far
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _any_errors()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

	return 1 if (scalar @{$self->{ERRORS}} > 0);
	return 0;
}

#-------------------------------------------------------------------------------

# _compress_config($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Compress parsed configurations:
#                - remove all empty subsections and sections
#	             - if the section has only one default instance, move configurations one layer up for that section
#	             - if there is only default section on top level, move the configurations one layer up
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _compress_config($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($compression,) = @_;

	# Remove the empty subsections and sections
	foreach my $section_name (keys(%{$self->{CONFIG}})) {
		foreach my $section_id (keys(%{$self->{CONFIG}->{$section_name}})) {
			# Remove empty subsection
			unless (scalar keys(%{$self->{CONFIG}->{$section_name}->{$section_id}})) {
				delete $self->{CONFIG}->{$section_name}->{$section_id};
			}
		}
		# Remove empty section
		unless (scalar keys(%{$self->{CONFIG}->{$section_name}})) {
			delete $self->{CONFIG}->{$section_name};
		}
	}

	# Compression level must be at least 1
	if ($compression >= 1) {
		# If the section has only default instance, move configurations up one layer for that section
		foreach my $section (keys(%{$self->{CONFIG}})) {
			my @ids = keys(%{$self->{CONFIG}->{$section}});
			my $dflt_section_id = DFLT_SECTION_ID;
			if (scalar @ids == 1 and $ids[0] eq $dflt_section_id) {
				$self->{CONFIG}->{$section} = dclone($self->{CONFIG}->{$section}->{$dflt_section_id});
			}
		}
	}

	# Compression level must be at least 2
	if ($compression >= 2) {
		# If there is only default section on top level, move the configurations one layer up
		my @names = keys(%{$self->{CONFIG}});
		my $dflt_section_name = DFLT_SECTION_NAME;
		if (scalar @names == 1 and $names[0] eq $dflt_section_name) {
			$self->{CONFIG} = dclone($self->{CONFIG}->{$dflt_section_name});
		}
	}
}

# _get_config() [PROTECTED]
#
#  Usage     :
#  Purpose   : Return parsed configurations
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _get_config()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

	# Optionally compress the configurations before returning them
	if ($self->{COMPRESS}) {
		$self->_compress_config($self->{COMPRESS});
	}
	# Optionally compress the configurations before returning them
	if ($self->{CB_POSTPROCESSING}) {
		(my $config, $self->{CONFIG}) = ($self->{CONFIG}, undef);
		$self->{CONFIG} = $self->{CB_POSTPROCESSING}->($config);
	}
	return $self->{CONFIG};
}

#-------------------------------------------------------------------------------

# _insert_configuration($$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Insert parsed configuration directive into inner configuration storage tree
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _insert_configuration($$$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($section_name, $section_id, $directive_key, $value) = @_;
	$section_id = ($section_id) ? $section_id : DFLT_SECTION_ID;

	my $ptr	= $self->{CONFIG}->{$section_name}->{$section_id};

	# Split directive key at dots to get path
	my @keys = split(/[.:]/, $directive_key);

	# Now walk down that path
	for (my $i = 0; $i < scalar(@keys); $i++) {
		# More keys in the path available, create the storage
		if (($i + 1) < scalar(@keys)) {
			# Be carefull not to clear previously created storage, it may already contain some values
			unless ($ptr->{$keys[$i]}) { $ptr->{$keys[$i]} = {}; }
			# Move pointer to the created key
			$ptr = $ptr->{$keys[$i]};
		}
		# Last directive, assign value
		else {
			# Some directive key values might be stacked together into array
			if ((ref $self->{STACK_VALUES} eq '' and $self->{STACK_VALUES}) or
				(ref $self->{STACK_VALUES} eq 'HASH' and $self->{STACK_VALUES}->{$directive_key}))
			{
				push(@{$ptr->{$keys[$i]}}, $value);
			}
			# Otherwise rewrite the previous values
			else {
				$ptr->{$keys[$i]} = $value;
			}
		}
	}
}

#-------------------------------------------------------------------------------

sub _check_directive_structure($$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($section_name, $directive, $value) = @_;

	return ($value, 1, "Invalid directive '$directive' in section '$section_name'") unless $self->{STRUCTURE}->{$section_name}->{$directive};

	my $specification = $self->{STRUCTURE}->{$section_name}->{$directive};
	my ($error, $result) = Config::Validator->validate($specification, $value);

	return ($result, 1, "Invalid value '$value' for datatype '$specification' in '$directive', section '$section_name': $error") if $error;
	return ($result, 0, undef);
}

sub _check_section_structure($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my $section_name = shift;

	return (1, "Invalid section '$section_name'") unless $self->{STRUCTURE}->{$section_name};
	return (0, undef);
}

sub _check_directive($$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($section_name, $section_id, $directive, $value, $error, $errmsg) = @_;

	# Call the optional directive structure check
	if ($self->{STRUCTURE}) {
		($value, $error, $errmsg) = $self->_check_directive_structure($section_name, $directive, $value);
		return ($directive, $value, $error, $errmsg) if $error;
	}
	# Call the optional directive validator callback, if it was defined
	if ($self->{CB_DIRECTIVE_VALIDATOR}) {
		($error, $errmsg) = $self->{CB_DIRECTIVE_VALIDATOR}->($section_name, $section_id, $directive, $value);
		return ($directive, $value, $error, $errmsg) if $error;
	}
	# Call the optional directive processor callback, if it was defined
	if ($self->{CB_DIRECTIVE_PROCESSOR}) {
		($directive, $value, $error, $errmsg) = $self->{CB_DIRECTIVE_PROCESSOR}->($section_name, $section_id, $directive, $value);
	}
	return ($directive, $value, $error, $errmsg);
}

sub _check_section($$$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($section_name, $section_id, $error, $errmsg) = @_;

	# Default section may not have any subsections identified with IDs
	if ($section_name eq DFLT_SECTION_NAME and $section_id ne DFLT_SECTION_ID) {
		return ($section_name, $section_id, 1, "Default section may not have any subsections identified with IDs")
	}
	# Call the optional section structure check
	if ($self->{STRUCTURE}) {
		($error, $errmsg) = $self->_check_section_structure($section_name);
		return ($section_name, $section_id, $error, $errmsg) if $error;
	}
	# Call the optional section validator callback, if it was defined
	if ($self->{CB_SECTION_VALIDATOR}) {
		($error, $errmsg) = $self->{CB_SECTION_VALIDATOR}->($section_name, $section_id);
		return ($section_name, $section_id, $error, $errmsg) if $error;
	}
	# Call the optional section processor callback, if it was defined
	if ($self->{CB_SECTION_PROCESSOR}) {
		($section_name, $section_id, $error, $errmsg) = $self->{CB_SECTION_PROCESSOR}->($section_name, $section_id);
	}
	# If the sections must be unique, check for it
	if ($self->{UNIQUE_SECTIONS} and $section_name ne DFLT_SECTION_NAME) {
		if ($self->{CONFIG}->{$section_name} and $self->{CONFIG}->{$section_name}->{$section_id}) {
			return ($section_name, $section_id, 1, "Section '$section_name' with id '$section_id' is already defined");
		}
	}
	# Initialize the storage for the current section
	unless ($self->{CONFIG}->{$section_name}->{$section_id}) {
		$self->{CONFIG}->{$section_name}->{$section_id} = {};
	}
	return ($section_name, $section_id, $error, $errmsg);
}

#-------------------------------------------------------------------------------

# _open_config_file($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Add new config file to the file stack
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _open_config_file($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($cfg_file,) = @_;

	# Is it really ordinary file
	unless (-T $cfg_file) {
		return "Config file '$cfg_file' is not ordinary plain text file";
	}

	# If the file was already included, it is an error
	if ($self->{CFG_FILES_INC}->{$cfg_file}) {
		return "Config file '$cfg_file' was already included. Multiple inclusions of the same file are not allowed";
	}

	# Try to open the config file
	open(my $input, $cfg_file) or return "Can`t open config file '$cfg_file'";

	# Push the handle to the stack and mark the filename in the registry
	unshift(@{$self->{CFG_FILES_HND}}, $input);
	unshift(@{$self->{CFG_FILES_CNT}}, 0);
	unshift(@{$self->{CFG_FILES_CUR}}, $cfg_file);
	$self->{CFG_FILES_INC}->{$cfg_file}++;
	return 0;
}

# _readline() [PROTECTED]
#
#  Usage     :
#  Purpose   : Read one line from the config file on top of the stack
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _readline()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

	# RECURSE BREAKPOINT, there are no more filehandles left
	return 0 unless ($self->{CFG_FILES_HND}->[0]);

	# Try to read and return line from current file
	my $fh = $self->{CFG_FILES_HND}->[0];
	my $line = <$fh>;
	if ($line) {
		return ($line, ++$self->{CFG_FILES_CNT}->[0], $self->{CFG_FILES_CUR}->[0]);
	}

	# Current file is empty, pop the stack and try to read from the next file
	$fh = shift(@{$self->{CFG_FILES_HND}});
	shift(@{$self->{CFG_FILES_CNT}});
	shift(@{$self->{CFG_FILES_CUR}});
	close($fh);
	return $self->_readline();
}

# _analyze_line($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Analyze given line
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _analyze_line($$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($line, $section_name, $section_id) = @_;

	my ($directive, $value, $error, $errmsg);

	# Section switching
	if ($line =~ /$RE_PTRN_SECTION/) {
		($section_name, $section_id, $error, $errmsg) = $self->_check_section($1, (defined($2) ? $2 : DFLT_SECTION_ID), 0, undef);
		return ($section_name, $section_id, $error, $errmsg);
	}
	# Configuration directive syntax
	elsif ($line =~ /$RE_PTRN_DIRECTIVE/) {
		($directive, $value, $error, $errmsg) = $self->_check_directive($section_name, $section_id, $1, $2, 0, undef);
		return ($section_name, $section_id, $error, $errmsg) if $error;

		$self->_insert_configuration($section_name, $section_id, $directive, $value);
		return ($section_name, $section_id, $error, $errmsg);
	}
	# Unknown line, return the ERROR flag SET
	else {
		return ($section_name, $section_id, 1);
	}
}

# _parse_config_file($) [PROTECTED]
#
#  Usage     :
#  Purpose   : Perform the analysis of a given file line by line
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _parse_config_file($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($cfg_file,) = @_;

	my $section_name = DFLT_SECTION_NAME;
	my $section_id   = DFLT_SECTION_ID;
	my $error        = 0;
	my $errmsg       = 0;
	my $counter      = 0;
	my $line         = 0;

	# Open and parse the master configuration file line by line
	$error = $self->_open_config_file($cfg_file);
	if($error) { croak ((caller(0))[3] . ": could not open toplevel config file '$cfg_file'\n"); }

	($section_name, $section_id, $error, $errmsg) = $self->_check_section($section_name, $section_id, $error, $errmsg);
	if($error) { croak ((caller(0))[3] . ": general section '$section_name' is not defined in the config structure\n"); }

	($line, $counter, $cfg_file) = $self->_readline();
	while ($line) {
		# Strip the white characters from the beginning and end of the line
		$line = trim($line);

		# Skip comments and empty lines
		if ($line =~ /$RE_PTRN_COMMENT/)		{ next; }
		if ($line =~ /$RE_PTRN_EMPTY_LINE/)	{ next; }

		# Special line - include subfile
		if ($line =~ /$RE_PTRN_INCLUDE/) {
			$errmsg	= $self->_open_config_file($1);
			$error  = ($errmsg) ? 1 : 0;
		}
		# Process the cofiguration directive line
		else {
			($section_name, $section_id, $error, $errmsg) = $self->_analyze_line($line, $section_name, $section_id);
		}

		# Log the detected syntax errors
		if($error) {
			push(@{$self->{ERRORS}}, ($errmsg) ? "[$cfg_file:$counter] $errmsg" : "[$cfg_file:$counter] Unknown line: '$line'");
		}
	}
	continue {
		($line, $counter, $cfg_file) = $self->_readline();
	}
}

=pod

=back

=cut

1;
