package Config::Validator;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Config::Validator - Configuration value validator

=head1 SYNOPSIS

 use Config::Validator;

 # Usage via class interface
 # Validate given configuration value
 my ($error, $value) = Config::Validator->validate($validator, $value);

 # Extract complex data types from configuration values
 my $result = Config::Validator->extract_list($value);
 my $result = Config::Validator->extract_hash($value);
 my $result = Config::Validator->extract_flags($value);

 # Usage via class interface
 # Validate given configuration value
 my $validator = Config::Validator->new();
 my ($error, $value) = $validator->validate($validator, $value);

 # Extract complex data types from configuration values
 my $result = $validator->extract_list($value);
 my $result = $validator->extract_hash($value);
 my $result = $validator->extract_flags($value);

=head1 DESCRIPTION

This class can be used for simple configuration value validation. The most important
key method called validate(), which can be called either on instance, or class. It
takes two arguments (see description in FUNCTION REFERENCE section for more details),
validation specification and value to be validated. Validation specification is a
description of the validator to be used along with optional arguments. validator name
corresponds with possible data types, that may be suplied in the configuration.
Currently following data types are defined:

	Simple: bool, flag, int, float, string, date, time, datetime, ipv4, ipv6, ip, file, enum
	Complex: list, hash, flags

For these datatypes appropriate validators are available:

=over

=item B<bool>

 Value is checked to have boolean TRUE or FALSE meaning. Check is CASE-INSENSITIVE.

 For boolean TRUE following values are recognized: 1,y,yes,t,true,on
 For boolean FALSE following values are recognized: 0,n,no,f,false,off

 Use case:
	Simple on/off switches

 Examples:
	my ($error, $value) = Config::Validator->validate('bool', 'y');

=item B<flag>

 Value is checked to be a flag (positive integers and zero).

 Use case:
	Repeated configuration increases the power (e.g. verbosity)

 Examples:
	my ($error, $value) = Config::Validator->validate('flag', 'y');

=item B<int[:min][:max]>

 Value is checked to have integer meaning. Optionally minimal and maximal value
 can be specified.

 Examples:
	my ($error, $value) = Config::Validator->validate('int',     '5');
	my ($error, $value) = Config::Validator->validate('int:4',   '5');
	my ($error, $value) = Config::Validator->validate('int::6',  '5');
	my ($error, $value) = Config::Validator->validate('int:4:6', '5');

=item B<float[:min][:max]>

 Value is checked to have float meaning. Optionally minimal and maximal value
 can be specified.

 Examples:
	my ($error, $value) = Config::Validator->validate('float',         '5.2');
	my ($error, $value) = Config::Validator->validate('float:4.1',     '5.2');
	my ($error, $value) = Config::Validator->validate('float::6.1',    '5.2');
	my ($error, $value) = Config::Validator->validate('float:4.1:6.1', '5.2');

=item B<string[:min][:max]>

 Value is checked to have string meaning. Optionally minimal and maximal length
 can be specified.

 Examples:
	my ($error, $value) = Config::Validator->validate('string',     'abcde');
	my ($error, $value) = Config::Validator->validate('string:4',   'abcde');
	my ($error, $value) = Config::Validator->validate('string::6',  'abcde');
	my ($error, $value) = Config::Validator->validate('string:4:6', 'abcde');

=item B<date>

 Value is checked to be the date in 'YYYY-MM-DD' format.

 Examples:
	my ($error, $value) = Config::Validator->validate('date', '2013-01-01');

=item B<time>

 Value is checked to be the time in 'HH:MM(:SS)' format.

 Examples:
	my ($error, $value) = Config::Validator->validate('time', '22:22');
	my ($error, $value) = Config::Validator->validate('time', '22:22:22');

=item B<datetime>

 Value is checked to be the datetime in 'YYYY-MM-DD(T)HH:MM(:SS)(Z|+-TZ)' format.

 Examples:
	my ($error, $value) = Config::Validator->validate('datetime', '2013-05-08T12:22Z');
	my ($error, $value) = Config::Validator->validate('datetime', '2013-05-08 12:22Z');
	my ($error, $value) = Config::Validator->validate('datetime', '2013-05-08T12:22:22Z');
	my ($error, $value) = Config::Validator->validate('datetime', '2013-05-08 12:22:22Z');
	my ($error, $value) = Config::Validator->validate('datetime', '2013-05-08T12:22:22+0100');
	my ($error, $value) = Config::Validator->validate('datetime', '2013-05-08 12:22:22-0100');

=item B<ipv4>

 Value is checked to be the IPv4 address in decimal dotted notation.

 Examples:
	my ($error, $value) = Config::Validator->validate('ipv4', '192.168.1.1');

=item B<ipv6>

 Value is checked to be the IPv6 address in hexadecimal colon spaced notation.

 Examples:
	my ($error, $value) = Config::Validator->validate('ipv6', '192:168::1:1');

=item B<ip>

 Value is checked to be the IPv4 or IPv6 address.

 Examples:
	my ($error, $value) = Config::Validator->validate('ip', '192.168.1.1');
	my ($error, $value) = Config::Validator->validate('ip', '192:168::1:1');

=item B<hostname>

 Value is checked to be the DNS hostname.

 Examples:
    my ($error, $value) = Config::Validator->validate('hostname', 'www.google.com');

=item B<host>

 Value is checked to be the DNS hostname or IPv4 or IPv6 address.

 Examples:
	my ($error, $value) = Config::Validator->validate('host', 'www.google.com');
	my ($error, $value) = Config::Validator->validate('host', '192.168.1.1');
	my ($error, $value) = Config::Validator->validate('host', '192:168::1:1');

=item B<email>

 Value is checked to be the email address.

 Examples:
	my ($error, $value) = Config::Validator->validate('email', 'info@gmail.com');

=item B<file[:[test][,test]...>

 Value is checked to be the reasonably sane filename. Optionally various file tests
 can be performed on the file.

 For the filename to be valid, it must consist on of the following characters:
	[-_./a-zA-Z0-9]
 Currently following file test are available:
	'f' - file must be ordinary file
	'd' - file must be directory
	'l' - file must be symbolic link
	'S' - file must be socket
	'T' - file must be ordinary text file
	'r' - file must be readable for effective UID/GID
	'w' - file must be writable for effective UID/GID
	'x' - file must be executable for effective UID/GID
	'o' - file must be owned by effective UID/GID
	'R' - file must be readable for real UID/GID
	'W' - file must be writable for real UID/GID
	'X' - file must be executable for real UID/GID
	'O' - file must be owned by real UID/GID
	'e' - file must exist
	's' - file must not be empty (non zero size)
	'z' - file must be empty (zero size)
	'a' - file name must be absolute path

 More tests can be specified at once, one of the following characters can be used
 as separator: [;, ].

 Examples:
	my ($error, $value) = Config::Validator->validate('file',         '/var/log/syslog');
	my ($error, $value) = Config::Validator->validate('file:f',       '/var/log/syslog');
	my ($error, $value) = Config::Validator->validate('file:e,f,w,s', '/var/log/syslog');

=item B<enum:val,val,...>

 Value is checked against the list of allowed options, one of the following characters
 can be used as separator: [;,].

 Examples:
	my ($error, $value) = Config::Validator->validate('enum:a,b,c,d', 'a');
	my ($error, $value) = Config::Validator->validate('enum:a;b;c;d', 'a');

=item B<list:(validator)>

 This validator can be used for value list validation. Every item in the list will
 be checked with selected validator. Any of the validators mentioned above can be used.

 Examples:
	my ($error, $value) = Config::Validator->validate('list:bool',          'y,n,0,1');
	my ($error, $value) = Config::Validator->validate('list:int::6',        '5,4,3,5,2');
	my ($error, $value) = Config::Validator->validate('list:float:4.1:6.1', '5.2;4.3;5.8');
	my ($error, $value) = Config::Validator->validate('list:string',        'abcde;123abc');
	my ($error, $value) = Config::Validator->validate('list:date',          '2013-01-01,2013-02-02');
	my ($error, $value) = Config::Validator->validate('list:time',          '22:22,11:11');
	my ($error, $value) = Config::Validator->validate('list:datetime',      '2013-05-08T12:22Z,2013-05-08T22:22Z');
	my ($error, $value) = Config::Validator->validate('list:ipv4',          '192.168.1.1,192.168.1.2');
	my ($error, $value) = Config::Validator->validate('list:ipv6',          '192:168::1:1;192:168::1:1');
	my ($error, $value) = Config::Validator->validate('list:ip',            '192.168.1.1;192:168::1:1');
	my ($error, $value) = Config::Validator->validate('list:file:e,f,w,s',  '/var/log/auth.log;/var/log/syslog');
	my ($error, $value) = Config::Validator->validate('list:enum:a;b;c;d',  'a;a;b;b;d');

 Items in the value list are expected to be separated by one of the following
 characters: [;,].

=item B<hash:key1,key2,...>

 This validator can be used for hash key validation. It internally uses the ENUM
 validation, so see the appropriate description for more details.

 Examples:
	my ($error, $value) = Config::Validator->validate('hash:a,c,e', 'a:b,c:d,e:f');
	my ($error, $value) = Config::Validator->validate('hash:a,c,e', 'a=>b;c=>d;e=>f');

 Items in the key list are expected to be separated by one of the following
 characters: [;,]. Hash specification uses [;,] characters for pair separation and
 (=>|:) for key-value separation.

=back

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Utils::String;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant DT_BOOL     => 'bool';
use constant DT_FLAG     => 'flag';
use constant DT_INT      => 'int';
use constant DT_FLOAT    => 'float';
use constant DT_STRING   => 'string';
use constant DT_DATE     => 'date';
use constant DT_TIME     => 'time';
use constant DT_DATETIME => 'datetime';
use constant DT_IPV4     => 'ipv4';
use constant DT_IPV6     => 'ipv6';
use constant DT_IP       => 'ip';
use constant DT_HOSTNAME => 'hostname';
use constant DT_HOST     => 'host';
use constant DT_EMAIL    => 'email';
use constant DT_FILE     => 'file';
use constant DT_ENUM     => 'enum';
use constant DT_LIST     => 'list';
use constant DT_HASH     => 'hash';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Compiled RE patterns for config value analysis
my $RE_PTRN_BOOL     = qr/^((?i)1|y|yes|t|true|on|0|n|no|f|false|off)$/;
my $RE_PTRN_TRUE     = qr/^((?i)1|y|yes|t|true|on)$/;
my $RE_PTRN_FALSE    = qr/^((?i)0|n|no|f|false|off)$/;
my $RE_PTRN_INT      = qr/^[-+]?\d+$/;
my $RE_PTRN_FLOAT    = qr/^[-+]?\d+(?:\.\d+)?$/;
my $RE_PTRN_DATE     = qr/^\d{4}-\d{2}-\d{2}$/;
my $RE_PTRN_TIME     = qr/^\d{2}:\d{2}(:\d{2})?$/;
my $RE_PTRN_DATETIME = qr/^(\d{4}-\d{2}-\d{2}(?:T| )\d{2}:\d{2}(?::\d{2})?(?:Z|\+\d{2}:?\d{2}|-\d{2}:?\d{2})?)$/;
my $RE_PTRN_IPV4ADDR = qr/^([\d]+\.[\d]+\.[\d]+\.[\d]+)$/;
my $RE_PTRN_IPV6ADDR = qr/^([:a-zA-Z0-9]+:[:a-zA-Z0-9]*)$/;
my $RE_PTRN_HOSTNAME = qr/^[-a-zA-Z0-9]+(\.[-a-zA-Z0-9]+)*$/;
my $RE_PTRN_EMAIL    = qr/^[-._a-zA-Z0-9]+@[-a-zA-Z0-9]+(\.[-a-zA-Z0-9]+)*/;
my $RE_PTRN_FILE     = qr{^/?((/|\./|\.\./)[-_.a-zA-Z0-9]*)+$};

my @DATA_TYPES = (
		DT_BOOL,DT_FLAG,DT_INT,DT_FLOAT,DT_STRING,DT_DATE,DT_TIME,DT_DATETIME,
		DT_IPV4,DT_IPV6,DT_IP,DT_HOSTNAME,DT_HOST,DT_EMAIL,DT_FILE,DT_ENUM,
		DT_LIST,DT_HASH,
	);
my %DATA_TYPES_SH = map { $_ => 1 } @DATA_TYPES;

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL %VALIDATORS %OPTIONS);
    $VERSION = '0.1';
    $DEVEL   = 0;

	# Storage for precalculated options
	%OPTIONS = ();

	# Available validators
    %VALIDATORS = (
		DT_BOOL() => sub {
			my ($value,) = @_;
			return ('',1) if $value =~ /$RE_PTRN_TRUE/;
			return ('',0) if $value =~ /$RE_PTRN_FALSE/;
			return "Invalid boolean value '$value', possible values are: [1,y,yes,t,true,on,0,n,no,f,false,off]";
		},
		DT_FLAG() => sub {
			my ($value,) = @_;
			return "Invalid flag value '$value', possible values are: [0,1,2,...]" unless $value =~ /$RE_PTRN_INT/ and $value >= 0;
			return ('', $value);
		},
		DT_INT() => sub {
			my ($value, $min, $max) = @_;
			$min = undef if defined $min and "A$min" eq "A"; $max = undef if defined $max and "A$max" eq "A";
			return "Invalid integer value '$value'" unless $value =~ /$RE_PTRN_INT/;
			return "Invalid integer value '$value', must be greater than or equal '$min'" if defined($min) and not $min <= $value;
			return "Invalid integer value '$value', must be less than or equal '$max'" if defined($max) and not $max >= $value;
			return ('', $value);
		},
		DT_FLOAT() => sub {
			my ($value, $min, $max) = @_;
			$min = undef if defined $min and "A$min" eq "A"; $max = undef if defined $max and "A$max" eq "A";
			return "Invalid float value '$value'" unless $value =~ /$RE_PTRN_FLOAT/;
			return "Invalid float value '$value', must be greater than or equal '$min'" if defined($min) and not $min <= $value;
			return "Invalid float value '$value', must be less than or equal '$max'" if defined($max) and not $max >= $value;
			return ('', $value);
		},
		DT_STRING() => sub {
			my ($value, $min, $max) = @_;
			$min = undef if defined $min and "A$min" eq "A"; $max = undef if defined $max and "A$max" eq "A";
			my $l = length($value);
			return "Invalid string value '$value', must be longer than or equal '$min' characters" if defined($min) and not $min <= $l;
			return "Invalid string value '$value', must be shorter than or equal '$max' characters" if defined($max) and not $max >= $l;
			return ('', _clean($value));
		},
		DT_DATE() => sub {
			my ($value,) = @_;
			return "Invalid date value '$value', expecting date in format 'YYYY-MM-DD'" unless $value =~ /$RE_PTRN_DATE/;
			return ('', $value);
		},
		DT_TIME() => sub {
			my ($value,) = @_;
			return "Invalid time value '$value', expecting time in format 'HH:MM(:SS)'" unless $value =~ /$RE_PTRN_TIME/;
			return ('', $value);
		},
		DT_DATETIME() => sub {
			my ($value,) = @_;
			return "Invalid datetime value '$value', expecting datetime in format 'YYYY-MM-DD(T)HH:MM(:SS)(Z|+-TZ)'" unless $value =~ /$RE_PTRN_DATETIME/;
			return ('', $value);
		},
		DT_IPV4() => sub {
			my ($value,) = @_;
			return "Invalid IPv4 address value '$value'" unless $value =~ /$RE_PTRN_IPV4ADDR/;
			return ('', $value);
		},
		DT_IPV6() => sub {
			my ($value,) = @_;
			return "Invalid IPv6 address value '$value'" unless $value =~ /$RE_PTRN_IPV6ADDR/;
			return ('', $value);
		},
		DT_IP() => sub {
			my ($value,) = @_;
			return "Invalid IP address value '$value'" unless $value =~ /$RE_PTRN_IPV4ADDR/ or $value =~ /$RE_PTRN_IPV6ADDR/;
			return ('', $value);
		},
		DT_HOSTNAME() => sub {
			my ($value,) = @_;
			return "Invalid hostname value '$value'" unless $value =~ /$RE_PTRN_HOSTNAME/;
			return ('', $value);
		},
		DT_HOST() => sub {
			my ($value,) = @_;
			return "Invalid host value '$value'" unless $value =~ /$RE_PTRN_HOSTNAME/ or $value =~ /$RE_PTRN_IPV4ADDR/ or $value =~ /$RE_PTRN_IPV6ADDR/;
			return ('', $value);
		},
		DT_EMAIL() => sub {
			my ($value,) = @_;
			return "Invalid email address value '$value'" unless $value =~ /$RE_PTRN_EMAIL/;
			return ('', $value);
		},
		DT_FILE() => sub {
			my ($value, $tests) = @_;
			return "Invalid filename value '$value'" unless $value =~ /$RE_PTRN_FILE/;
			my @tests = ();	@tests = split(/[,;]/, $tests) if defined($tests);
			foreach my $p (@tests) {
				return "File '$value' is not ordinary file" if $p eq 'f' and not -f $value;
				return "File '$value' is not directory" if $p eq 'd' and not -d $value;
				return "File '$value' is not symbolic link" if $p eq 'l' and not -l $value;
				return "File '$value' is not socket file" if $p eq 'S' and not -S $value;
				return "File '$value' is not plain text file" if $p eq 'T' and not -T $value;

				return "File '$value' is not readable by effective UID/GID" if $p eq 'r' and not -r $value;
				return "File '$value' is not writable by effective UID/GID" if $p eq 'w' and not -w $value;
				return "File '$value' is not executable by effective UID/GID" if $p eq 'x' and not -x $value;
				return "File '$value' is not owned by effective UID/GID" if $p eq 'o' and not -o $value;
				return "File '$value' is not readable by real UID/GID" if $p eq 'R' and not -R $value;
				return "File '$value' is not writable by real UID/GID" if $p eq 'W' and not -W $value;
				return "File '$value' is not executable by real UID/GID" if $p eq 'X' and not -X $value;
				return "File '$value' is not owned by real UID/GID" if $p eq 'O' and not -O $value;

				return "File '$value' does not exist" if $p eq 'e' and not -e $value;
				return "File '$value' is empty" if $p eq 's' and not -s $value;
				return "File '$value' is not empty" if $p eq 'z' and not -z $value;

				return "File '$value' is not absolute path" if $p eq 'a' and not $value =~ /^\//;
			}
			return ('', _clean($value));
		},
		DT_ENUM() => sub {
			my ($value, $options) = @_;
			return ('', $value) unless $options;
			my $o = _get_options($options);
			return ('', _clean($value)) if $o->{$value};
			return "Invalid value '$value', possible values are: [" . join(',',sort keys(%$o)) . "]";
		},
		# There is another possibility for the recursion, maybe it is better?
		#DT_LIST() => sub {
			#my ($value, $validation) = @_;
			#$value = _extract_list($value);
			#return ('', $value) unless $validation;
			#my ($validator, $rest) = split(/:/, $validation, 2);
			# ...
		#},
	);
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Config::Validator->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Config::Validator reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
	my $self = bless ({}, $class);
	return $self->_init(@_);
}

#-------------------------------------------------------------------------------

=item valid_datatypes() [PUBLIC,STATIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :
  Comments  :
  See Also  :

=cut

sub valid_datatypes()
{
	my $invocant = shift;
	return @DATA_TYPES;
}

=item is_valid_datatype($) [PUBLIC,STATIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :
  Comments  :
  See Also  :

=cut

sub is_valid_datatype($)
{
	my $invocant = shift;
	my ($specification,) = @_;
	my ($dt, @args) = split(/:/, $specification);

	$dt = lc($dt);
	return $dt if $DATA_TYPES_SH{$dt};
	return undef;
}

#-------------------------------------------------------------------------------

=item extract_list($) [PUBLIC,HYBRID]

  Usage     : my $result = Config::Validator->extract_list($value);
  Purpose   : Extract the list from given string serialization
  Arguments : STRING $serialization - hash serialization
  Returns   : ARRAY REFERENCE $list
  Throws    : Croaks, if not given all necessary arguments

=cut

sub extract_list($)
{
	my $invocant = shift;
	my ($serialization,) = @_;
	croak ((caller(0))[3] . ": serialized list must be given as first argument") unless $serialization;

	return $serialization if ref $serialization;

	my @v = ();
	push(@v, split(/[,;]/, $serialization));
	@v = map { trim($_) } @v;
	return \@v;
}

=item extract_hash($) [PUBLIC,HYBRID]

  Usage     : my $result = Config::Validator->extract_hash($value);
  Purpose   : Extract the hash from given string serialization
  Arguments : STRING $serialization - hash serialization
  Returns   : HASH REFERENCE $hash
  Throws    : Croaks, if not given all necessary arguments

=cut

sub extract_hash($)
{
	my $invocant = shift;
	my ($serialization,) = @_;
	croak ((caller(0))[3] . ": values to be extracted must be given as first argument") unless $serialization;

	return $serialization if ref $serialization;

	my @key_value_pairs = split(/[,;]/, $serialization);
	my $hash = {};
	foreach my $key_value_pair (@key_value_pairs) {
		my ($key, $val) = split(/=>|:/, $key_value_pair);
		$key = trim($key); $val = trim($val);
		$hash->{$key} = $val;
	}
	return $hash;
}

=item extract_flags($) [PUBLIC,HYBRID]

  Usage     : my $result = Config::Validator->extract_flags($value);
  Purpose   : Extract the hash from given string serialization
  Arguments : STRING $serialization - flags serialization
  Returns   : HASH REFERENCE $hash
  Throws    : Croaks, if not given all necessary arguments

=cut

sub extract_flags($)
{
	my $invocant = shift;
	my ($serialization,) = @_;
	croak ((caller(0))[3] . ": values to be extracted must be given as first argument") unless $serialization;

	return $serialization if ref $serialization;

	my @values = split(/[,;]/, $serialization);
	my $hash = {};
	map { $hash->{trim($_)}++; } @values;
	return $hash;
}

#-------------------------------------------------------------------------------

=item validate($$) [PUBLIC,HYBRID]

  Usage     : my ($error, $value) = $instance->validate($validation, $value); OR
              my ($error, $value) = Config::Validator->validate($validation, $value);
  Purpose   : Validate given configuration value using given validator specification
  Arguments : STRING $validation - validator specification
              STRING $value - value to be validated
  Returns   : EMPTY STRING on validation SUCCESS, STRING $error on FAILURE
  Throws    : Croaks, if not given all necessary arguments

=cut

sub validate($$)
{
	my $invocant = shift;
	my ($validation, $value) = @_;
	croak ((caller(0))[3] . ": validation specification must be given as first argument") unless $validation;
	croak ((caller(0))[3] . ": value to be validated must be given as second argument") unless defined $value;

	#$value = _clean($value);

	my ($validator, @args) = split(/:/, $validation);
	die "Validator must be given in validation specification\n" unless $validator;

	if ($validator eq DT_LIST()) {
		my $result = [];
		my $error;
		$value = $invocant->extract_list($value);
		if (@args) {
			$validation = join(':', @args);
			foreach my $v (@$value) {
				($error, $v) = $invocant->validate($validation, $v);
				return ($error, $v) if $error;
				push(@$result, $v);
			}
			return ('', $result);
		}
		return ('', $value);
	}
	elsif ($validator eq DT_HASH()) {
		my $result = {};
		my $error;
		$value = $invocant->extract_hash($value);
		if (@args) {
			$validation = join(':', DT_ENUM(), @args);
			foreach my $k (keys %$value) {
				($error, $k) = $invocant->validate($validation, $k);
				return ($error, $k) if $error;
				$result->{$k} = clean($value->{$k});
			}
			return ('', $result);
		}
		return ('', $value);
	}
	else {
		die "Invalid validator '$validator'\n" unless exists $VALIDATORS{$validator};
		return &{$VALIDATORS{$validator}}(clean($value), @args);
	}
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Config::Validator instance
#  Arguments : None
#  Returns   : Config::Validator reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	return $self;
}

# _get_options($) [PROTECTED]
#
#  Usage     : my $hash = _get_options($options);
#  Purpose   : Parse and cache given options list and return it as search hash
#  Arguments : STRING $options - options list
#  Returns   : HASH REFERENCE $hash
#  Throws    : Croaks, if not given mandatory arguments

sub _get_options($)
{
	my ($options,) = @_;
	confess "Options must be given as argument" unless $options;

	unless ($OPTIONS{$options}) {
		my @o = (); push(@o, split(/[,;]/, $options)) if defined($options);
		my %o = map { $_ => 1 } @o;
		$OPTIONS{$options} = \%o;
	}
	return $OPTIONS{$options};
}

# _clean($) [PROTECTED]
#
#  Usage     : my $value = _clean($value);
#  Purpose   : Clean and untaint given value
#  Arguments : STRING $value - value to be cleaned [MANDATORY]
#  Returns   : STRING $value
#  Throws    : Nothing

sub _clean($)
{
	my ($value,) = @_;
	if ($value =~ /(.*)/) { $value = $1; }
	return $value;
}

# Recursive cleaning function, currently unused
#sub _clean
#{
	#my ($value,) = @_;
	#return undef unless defined $value;

	#unless (ref $value) {
		#return clean($value);
	#}
	#elsif (ref $value eq 'ARRAY') {
		#my $result = [];
		#foreach my $v (@$value) {
			#push(@$result, _clean($v));
		#}
		#return $result;
	#}
	#elsif (ref $value eq 'HASH') {
		#my $result = {};
		#foreach my $k (keys %$value) {
			#$result->{_clean($k)} = _clean($value->{$k});
		#}
		#return $result;
	#}
	#croak ((caller(0))[3] . ": invalid value given to be cleaned: $value");
#}

=pod

=back

=cut

1;
