#
# 1. Změna v RE pro date-time, sekundy by měly být optional
# 2. pattern a format jsou exclusive, problem u duration
#
package JSON::Schema::Validator;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

JSON::Schema::Validator - JSON Schema Validator

=head1 SYNOPSIS

 use JSON::Schema::Validator;

 # Constructor takes name of the schema file as argument
 my $validator = JSON::Schema::Validator->new('/path/to/json/schema.json');

 # OR you may pass the JSON schema as string
 my $schema = '
 {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "type": "object"
 }
 ';
 my $validator = JSON::Schema::Validator->new($schema);

 # OR you may pass the JSON schema as data structure
 my $schema = {
		"$schema": "http://json-schema.org/draft-04/schema#",
		"type": "object"
	};
 my $validator = JSON::Schema::Validator->new($schema);

 # Validate given object structure
 my $error = $validator->validate($object);

 # Validate object stored within given file
 my $error = $validator->validate_file('/path/to/json/file.json');

=head1 DESCRIPTION

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;

#-- Custom application modules ------------------------------------------------#
use Config::Validator;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
# General keywords
use constant KEY_SCHEMA      => '$schema';
use constant KEY_REF         => '$ref';
use constant KEY_DEFINITIONS => 'definitions';
use constant KEY_TITLE       => 'title';
use constant KEY_DESCRIPTION => 'description';
use constant KEY_DEFAULT     => 'default';
use constant KEY_ENUM        => 'enum';
use constant KEY_TYPE        => 'type';

# Schema combination keywords
use constant KEY_ANYOF => 'anyOf';
use constant KEY_ALLOF => 'allOf';
use constant KEY_ONEOF => 'oneOf';
use constant KEY_NOT   => 'not';

# JSON data types
use constant T_NULL    => 'null';
use constant T_BOOLEAN => 'boolean';
use constant T_INTEGER => 'integer';
use constant T_NUMBER  => 'number';
use constant T_STRING  => 'string';
use constant T_ARRAY   => 'array';
use constant T_OBJECT  => 'object';
use constant T_NUMERIC => 'numeric'; # Custom internal data type, not part of the JSON Schema specs
use constant T_CUSTOM  => 'custom';  # Custom internal data type, not part of the JSON Schema specs
use constant T_UNDEF   => 'null';    # Custom internal data type, not part of the JSON Schema specs

# STRING format types
use constant F_DATE     => 'date';      # Custom format, not part of the JSON Schema specs
use constant F_TIME     => 'time';      # Custom format, not part of the JSON Schema specs
use constant F_DATETIME => 'date-time';
use constant F_DURATION => 'duration';  # Custom format, not part of the JSON Schema specs
use constant F_EMAIL    => 'email';
use constant F_HOSTNAME => 'hostname';
use constant F_IPV4     => 'ipv4';
use constant F_IPV6     => 'ipv6';
use constant F_IP       => 'ip';        # Custom format, not part of the JSON Schema specs
use constant F_HOST     => 'host';      # Custom format, not part of the JSON Schema specs
use constant F_URI      => 'uri';
use constant F_FILE     => 'file';      # Custom format, not part of the JSON Schema specs

# INTEGER and NUMBER datatype properties
use constant KEY_MULTIPLE_OF => 'multipleOf';
use constant KEY_MINIMUM     => 'minimum';
use constant KEY_MAXIMUM     => 'maximum';
use constant KEY_EXMINIMUM   => 'exclusiveMinimum';
use constant KEY_EXMAXIMUM   => 'exclusiveMaximum';

# STRING datatype properties
use constant KEY_MINLENGTH => 'minLength';
use constant KEY_MAXLENGTH => 'maxLength';
use constant KEY_PATTERN   => 'pattern';
use constant KEY_FORMAT    => 'format';

# ARRAY datatype properties
use constant KEY_ITEMS    => 'items';
use constant KEY_ADTITEMS => 'additionalItems';
use constant KEY_MINITEMS => 'minItems';
use constant KEY_MAXITEMS => 'maxItems';
use constant KEY_UNQITEMS => 'uniqueItems';

# OBJECT datatype properties
use constant KEY_PROPERTIES    => 'properties';
use constant KEY_ADTPROPERTIES => 'additionalProperties';
use constant KEY_PTRPROPERTIES => 'patternProperties';
use constant KEY_REQUIRED      => 'required';
use constant KEY_MINPROPERTIES => 'minProperties';
use constant KEY_MAXPROPERTIES => 'maxProperties';
use constant KEY_DEPENDENCIES  => 'dependencies';
use constant KEY_PROPPATTERN   => 'proppattern';  # Custom internal property, not part of the JSON Schema specs

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Schema validators
my @SCHEMA_VALIDATORS = (
	# NUMERIC specific validators
	[KEY_MULTIPLE_OF(), '_validate_multipleof',       T_NUMERIC()],
	[KEY_MINIMUM(),     '_validate_minimum',          T_NUMERIC()],
	[KEY_MAXIMUM(),     '_validate_maximum',          T_NUMERIC()],
	[KEY_EXMINIMUM(),   '_validate_exclusiveminimum', T_NUMERIC()],
	[KEY_EXMAXIMUM(),   '_validate_exclusivemaximum', T_NUMERIC()],

	# STRING specific validators
	[KEY_MINLENGTH(), '_validate_minlength', T_STRING()],
	[KEY_MAXLENGTH(), '_validate_maxlength', T_STRING()],
	[KEY_PATTERN(),   '_validate_pattern',   T_STRING()],
	[KEY_FORMAT(),    '_validate_format',    T_STRING()],

	# ARRAY specific validators
	[KEY_ITEMS(),    '_validate_items',           T_ARRAY()],
	[KEY_ADTITEMS(), '_validate_additionalitems', T_ARRAY()],
	[KEY_MINITEMS(), '_validate_minitems',        T_ARRAY()],
	[KEY_MAXITEMS(), '_validate_maxitems',        T_ARRAY()],
	[KEY_UNQITEMS(), '_validate_uniqueitems',     T_ARRAY()],

	# OBJECT specific validators
	[KEY_REQUIRED(),      '_validate_required',             T_OBJECT()],
	[KEY_PROPERTIES(),    '_validate_properties',           T_OBJECT()],
	[KEY_PTRPROPERTIES(), '_validate_patternproperties',    T_OBJECT()],
	[KEY_ADTPROPERTIES(), '_validate_additionalproperties', T_OBJECT()],
	[KEY_MINPROPERTIES(), '_validate_minproperties',        T_OBJECT()],
	[KEY_MAXPROPERTIES(), '_validate_maxproperties',        T_OBJECT()],
	[KEY_DEPENDENCIES (), '_validate_dependencies',         T_OBJECT()],

	# Common validators
	[KEY_ENUM(),  '_validate_enum'],

	# Combined schema validators
	[KEY_ANYOF(), '_validate_anyof'],
	[KEY_ALLOF(), '_validate_allof'],
	[KEY_ONEOF(), '_validate_oneof'],
	[KEY_NOT(),   '_validate_not'],
);
my %SCHEMA_VALIDATORS_H = map { $_->[0] => $_->[1] } @SCHEMA_VALIDATORS;

# Datatype validators
my %DT_VALIDATORS = (
	T_NULL()    => '_validate_null',
	T_BOOLEAN() => '_validate_boolean',
	T_INTEGER() => '_validate_integer',
	T_NUMBER()  => '_validate_number',
	T_STRING()  => '_validate_string',
	T_ARRAY()   => '_validate_array',
	T_OBJECT()  => '_validate_object',
);

# Format validators
my %FM_VALIDATORS = (
	#F_DATE()     => '_vf_date',     # Custom format, not part of the JSON Schema specs
	#F_TIME()     => '_vf_time',     # Custom format, not part of the JSON Schema specs
	F_DATETIME() => '_vf_datetime',
	#F_DURATION() => '_vf_duration', # Custom format, not part of the JSON Schema specs
	F_EMAIL()    => '_vf_email',
	F_HOSTNAME() => '_vf_hostname',
	F_IPV4()     => '_vf_ipv4',
	F_IPV6()     => '_vf_ipv6',
	#F_IP()       => '_vf_ip',       # Custom format, not part of the JSON Schema specs
	#F_HOST()     => '_vf_host',     # Custom format, not part of the JSON Schema specs
	F_URI()      => '_vf_uri',
	#F_FILE()     => '_vf_file',     # Custom format, not part of the JSON Schema specs
);

# Regular expressions for datatype checks
my $RE_BOOLEAN = qr/^1|0|true|false|t|f|yes|no|y|n|on|off$/i;
my $RE_INTEGER = qr/^-?\d+$/i;
my $RE_NUMBER  = qr/^-?\d+(?:\.\d+)?(?:e-?\d+)?$/i;

# Regular expressions for string format checks
my $RE_F_DATETIME = qr/^[0-9]{4}-[0-9]{2}-[0-9]{2}[Tt ][0-9]{2}:[0-9]{2}(?::[0-9]{2})?(?:\\.[0-9]+)?(?:[Zz]|(?:[+-][0-9]{2}:[0-9]{2}))?$/;
my $RE_F_EMAIL    = qr/^[-._a-zA-Z0-9]+@[-a-zA-Z0-9]+(\.[-a-zA-Z0-9]+)*/;
my $RE_F_HOSTNAME = qr/^[-a-zA-Z0-9]+(\.[-a-zA-Z0-9]+)*$/;
my $RE_F_IPV4     = qr/^([\d]+\.[\d]+\.[\d]+\.[\d]+)$/;
my $RE_F_IPV6     = qr/^([:a-zA-Z0-9]+:[:a-zA-Z0-9]*)$/;
my $RE_F_URI      = qr/^[a-zA-Z][a-zA-Z0-9+.-]*:[][a-zA-Z0-9._~:\/?#@*'&'()*+,;=%-]*$/;

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $validator = JSON::Schema::Validator->new(...);
  Purpose   : Create and return reference to the new instance
  Arguments : Same as _init() method
  Returns   : Mentat::Message::Validator reference
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my $self = bless ({}, $class);
	$self->_init(@_);
	return $self;
}

=item validate($) [PUBLIC]

  Usage     : my $error_string = $validator->validate($object);
  Purpose   : Validate the given object against the internal schema
  Arguments : HASH REFERENCE $object - object reference
  Returns   : STRING $error - string describing error on validation FAILURE
              UNDEF on validation SUCCESS
  Throws    : Croaks, if not invoked on object

=cut

sub validate($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($object,) = @_;
	return "Validated object must be given via reference" unless $object and ref $object;

	return $self->_validate($object, $self->{SCHEMA},'#');
}

=item validate_file($) [PUBLIC]

  Usage     : my $error_string = $validator->validate_file($object_file);
  Purpose   : Validate object located in given file against the internal schema
  Arguments : STRING $object_file - name of the file containing the validated object
  Returns   : Same as validate() method
  Throws    : Croaks, if not invoked on object

=cut

sub validate_file($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($object_file,) = @_;
	return "Object file must be ordinary file" unless -f $object_file;

	# Load object from external file
	my $object = $self->load_json_file($object_file);

	# Perform the validation
	return $self->validate($object);
}

=item schema($) [PUBLIC]

  Usage     : my $schema = $validator->schema();
  Purpose   : Get the and preprocessed JSON schema against which the objects
              will be validated
  Arguments : None
  Returns   : HASH REFERENCE $schema
  Throws    : Croaks, if not invoked on object

=cut

sub schema()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	return $self->{SCHEMA};
}

#===============================================================================
# Private interface
#===============================================================================

# _init($) [PROTECTED]
#
#  Usage     : In constructor: return $self->_init(@_);
#  Purpose   : Initialization callback for subclassess
#  Arguments : STRING $schema - name of the file containing the JSON schema, OR
#              STRING $schema - JSON schema as string, OR
#              HASH REFERENCE $schema - schema as data structure
#  Returns   : JSON::Schema::Validator reference

sub _init($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($schema,) = @_;
	confess "JSON schema must be given as argument" unless $schema;

	# JSON schema may be given via external file
	if (-f $schema) {
		$schema = $self->_load_json_file($schema);
	}
	# Or it may be plain string
	elsif (not ref $schema) {
		$schema = $self->_decode_json($schema);
	}
	$self->{SCHEMA} = $schema;

	# Preprocess the loaded schema
	$self->_process_schema($self->{SCHEMA}, $self->{SCHEMA}, '#');

	return $self;
}

# _load_json_file($) [PROTECTED]
#
#  Usage     : my $json = $self->_load_json_file($file);
#  Purpose   : Load JSON structure from external file
#  Arguments : STRING $file - name of the JSON file
#  Returns   : REFERENCE $json

sub _load_json_file($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($file,) = @_;
	confess "JSON file must be given as argument" unless $file;
	confess "JSON file must be ordinary file" unless -f $file;

	# Load JSON string from external file
	open (my $fh, $file) or die "Can`t open JSON file '$file'\n";
	my $json;
	while (my $line = <$fh>) {
		# Custom feature - remove comments
		next if $line =~ /^\s*(#|;|\/\/)/;
		$json .= $line;
	}
	close($fh);

	# Untaint the loaded string
	if ($json =~ /^(.+)$/s) { $json = $1; }

	# Parse out the JSON object and return it
	return $self->_decode_json($json);
}

# _decode_json($) [PROTECTED]
#
#  Usage     : my $json = $self->_decode_json($string);
#  Purpose   : Decode JSON string into object
#  Arguments : STRING $string - JSON string
#  Returns   : REFERENCE $json

sub _decode_json($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($string,) = @_;

	# Parse out the JSON object and return it
	return decode_json($string);
}

#-------------------------------------------------------------------------------
# Schema postprocessing
#-------------------------------------------------------------------------------

# _process_schema($$) [PROTECTED]
#
#  Usage     : my $object = $self->_process_schema($schema, $current, $path);
#  Purpose   : Process JSON schema
#  Arguments : OBJECT $schema - JSON schema
#              OBJECT $current - current position in JSON schema
#              STRING $path - path to the current position in JSON schema
#  Returns   : Nothing

sub _process_schema($$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($schema, $current, $path) = @_;

	# Schema is expected to be HASH REFERENCE
	return unless ref $current eq 'HASH';

	# First of all expand references with target schema
	$current = $self->_expand_ref($schema, $current, $path) if ($current->{KEY_REF()});

	# Now process all keys within the schema
	my $path_current;
	foreach my $key (sort keys %$current) {
		$path_current = __mp($path, $key);

		# Precompile string regular expression patterns
		if ($key eq KEY_PATTERN()) {
			unless (ref $current->{$key} eq 'Regexp') {
				my $ptrn = $current->{$key};
				print STDERR "Preprocessing '$path_current': precompiling string pattern '$ptrn'\n" if $DEVEL > 1;
				$current->{$key} = qr/$ptrn/;
			}
			else {
				print STDERR "Preprocessing '$path_current': string pattern was already precompiled\n" if $DEVEL > 1;
			}
		}
		# Precompile object property regular expression patterns
		elsif ($key eq KEY_PTRPROPERTIES()) {
			foreach my $prop (keys %{$current->{$key}}) {
				unless ($current->{$key}->{$prop}->{KEY_PROPPATTERN()}) {
					print STDERR "Preprocessing '$path_current': precompiling property pattern '$prop'\n" if $DEVEL > 1;
					$current->{$key}->{$prop}->{KEY_PROPPATTERN()} = qr/$prop/;
				}
				else {
					print STDERR "Preprocessing '$path_current': property pattern '$prop' was already precompiled\n" if $DEVEL > 1;
				}
			}
		}
		# Transform enum lists into search hashes
		elsif ($key eq KEY_ENUM()) {
			unless (ref $current->{$key} eq 'HASH') {
				my @list = @{$current->{$key}};
				print STDERR "Preprocessing '$path_current': compiling enum '".join(', ', __lv(@list))."'\n" if $DEVEL > 1;
				my %h = map { __v($_) => 1 } @list;
				$current->{$key} = \%h;
			}
			else {
				print STDERR "Preprocessing '$path_current': enum was already compiled\n" if $DEVEL > 1;
			}
		}
		# Recursively postprocess subhashes
		elsif (ref $current->{$key} eq 'HASH') {
			print STDERR "Preprocessing '$path_current': recursive descend into hash\n" if $DEVEL > 1;
			$self->_process_schema($self->{SCHEMA}, $current->{$key}, $path_current);
		}
		# Recursively postprocess subarrays
		elsif (ref $current->{$key} eq 'ARRAY') {
			my $cnt = 0;
			foreach my $item (@{$current->{$key}}) {
				$path_current = __mp($path, $key.'['.($cnt++).']');
				print STDERR "Preprocessing '$path_current': recursive descend into array\n" if $DEVEL > 1;
				$self->_process_schema($self->{SCHEMA}, $item, $path_current) if ref $item;
			}
		}
		# Just display anything else
		else {
			print STDERR "Preprocessing '$path_current': nothing to be done\n" if $DEVEL > 1;
		}
	}
}

# _expand_ref($$$) [PROTECTED]
#
#  Usage     : my $object = $self->_expand_ref($schema, $current, $path);
#  Purpose   : Expand references with target schema
#  Arguments : OBJECT $schema - JSON schema
#              OBJECT $current - current position in JSON schema
#              STRING $path - path to the current position in JSON schema
#  Returns   : REFERENCE $object

sub _expand_ref($$$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($schema, $current, $path) = @_;

	my $target = $current->{KEY_REF()};
	my $replacement;

	# Fetch the referenced schema
	if ($target =~ /^#\/(.+)/) {
		print STDERR "Preprocessing '$path': expanding reference to '$target'\n" if $DEVEL > 1;
		$replacement = $self->_fetch_ref_hash($schema, $1);
	}
	else {
		die "Unknown reference for schema key '$path' to '$target'\n";
	}

	# Replace the reference with referenced schema
	delete $current->{KEY_REF()};
	foreach my $k (keys %$replacement) {
		unless (exists $current->{$k}) {
			print STDERR "Preprocessing '$path': adding schema key '$k'\n" if $DEVEL > 1;
			$current->{$k} = $replacement->{$k};
		}
	}

	return $current;
}

# _fetch_ref_hash($$) [PROTECTED]
#
#  Usage     : my $object = $self->_fetch_ref_hash($schema, $target);
#  Purpose   : Fetch referenced schema within given hash
#  Arguments : OBJECT $schema - JSON schema
#              STRING $target - referenced target
#  Returns   : REFERENCE $object

sub _fetch_ref_hash($$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($schema, $target) = @_;

	my $result = $schema;
	my @nodes = split(/\//, $target);
	foreach my $n (@nodes) {
		die "Invalid reference path '$target', target must be object\n" unless $result and ref $result eq 'HASH';
		die "Invalid reference path '$target', node '$n' does not exist\n" unless $result->{$n};
		$result = $result->{$n};
	}
	return $result;
}

#*******************************************************************************
# Validation related methods
#*******************************************************************************

# _validate($$$) [PROTECTED]
#
#  Usage     : my $object = $self->_validate($object, $schema, $path);
#  Purpose   : Master validation method, validate given object against the
#              given schema
#  Arguments : OBJECT $object - object to be validated
#              STRING $schema - validation schema
#              STRING $path - path to the current object
#  Returns   : STRING $error on validation FAILURE, UNDEF on SUCCESS

sub _validate($$$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($object, $schema, $path) = @_;
	confess "Object schema must be given via hash reference" unless ref $schema;

	# First try to validate according to the TYPE keyword
	my ($type, $validator, $e);
	if ($schema->{KEY_TYPE()}) {
		$type      = $schema->{KEY_TYPE()};
		$validator = $DT_VALIDATORS{$type};
		$validator = '_validate_custom' unless $validator;

		# First perform the datatype validation
		$e = $self->$validator($object, $schema, $path, $type);
		return $e if $e;
		# Use the validation dispatcher method to execute additional validations
		$e = $self->_validation_dispatcher($object, $schema, $path, $type);
		return $e if $e;
	}
	# Without knowing the datatype walk through all defined keys and make best effort
	else {
		# Use the validation dispatcher method to execute validations in best order
		$e = $self->_validation_dispatcher($object, $schema, $path);
		return $e if $e;
	}
	return undef;
}

# _validation_dispatcher($$$$) [PROTECTED]
#
#  Usage     : my $result = $self->_validation_dispatcher($object, $schema, $path, $type);
#  Purpose   : Execute validations in best order
#  Arguments : OBJECT $object - object to be validated
#              STRING $schema - validation schema
#              STRING $path - path to the current object
#  Returns   : STRING $error on validation FAILURE, UNDEF on SUCCESS

sub _validation_dispatcher($$$$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($object, $schema, $path, $type) = @_;
	$type = __tt($type) if $type;

	my ($key, $validator, $validator_type, $e);
	foreach my $validation (@SCHEMA_VALIDATORS) {
		($key, $validator, $validator_type) = @$validation;

		# If the datatype is known, some validations does not make sense, so do
		# not even bother
		next if $type and $validator_type and $validator_type ne $type;

		# Proceed only if the validation is defined within schema for currently
		# validated object
		if (exists $schema->{$key}) {
			$e = $self->$validator($object, $schema, $path);
			return $e if $e;
		}
	}
	return undef;
}

#-------------------------------------------------------------------------------
# Datatype validators
#
#  Purpose   : Validate, that the given object has proper datatype
#  Arguments : OBJECT $object - object to be validated
#              STRING $schema - validation schema
#              STRING $path - path to the current object
#              STRING $type - validated datatype (usable for custom validation)
#  Returns   : STRING $error on validation FAILURE, UNDEF on SUCCESS
#-------------------------------------------------------------------------------

sub _validate_null($$$$)
{
	my $self = shift;
	my ($object, $schema, $path, $type) = @_;
	print STDERR "Validation '$path': value '".__v($object)."' against datatype 'null'\n" if $DEVEL;

	return "'$path': invalid value '".__v($object)."', expected '$type'" if defined $object;
	return undef;
}
sub _validate_boolean($$$$)
{
	my $self = shift;
	my ($object, $schema, $path, $type) = @_;
	print STDERR "Validation '$path': value '".__v($object)."' against datatype 'boolean'\n" if $DEVEL;

	return "'$path': expected '$type', received undefined value instead" unless defined $object;
	return "'$path': invalid value '".__v($object)."', expected '$type'" unless $object =~ /$RE_BOOLEAN/;
	return undef;
}
sub _validate_integer($$$$)
{
	my $self = shift;
	my ($object, $schema, $path, $type) = @_;
	print STDERR "Validation '$path': value '".__v($object)."' against datatype 'integer'\n" if $DEVEL;

	return "'$path': received 'null' value, expected '$type'" unless defined $object;
	return "'$path': invalid value '".__v($object)."', expected '$type'" unless $object =~ /$RE_INTEGER/;
	return undef;
}
sub _validate_number($$$$)
{
	my $self = shift;
	my ($object, $schema, $path, $type) = @_;
	print STDERR "Validation '$path': value '".__v($object)."' against datatype 'number'\n" if $DEVEL;

	return "'$path': received 'null' value, expected '$type'" unless defined $object;
	return "'$path': invalid value '".__v($object)."', expected '$type'" unless $object =~ /$RE_NUMBER/;
	return undef;
}
sub _validate_string($$$$)
{
	my $self = shift;
	my ($object, $schema, $path, $type) = @_;
	print STDERR "Validation '$path': value '".__v($object)."' against datatype 'string'\n" if $DEVEL;

	return "'$path': received 'null' value, expected '$type'" unless defined $object;
	#return "'$path': invalid value '".__v($object)."', expected '$type'" if ref $object;
	return undef;
}
sub _validate_array($$$$)
{
	my $self = shift;
	my ($object, $schema, $path, $type) = @_;
	print STDERR "Validation '$path': value '".__v($object)."' against datatype 'array'\n" if $DEVEL;

	return "'$path': received 'null' value, expected '$type'" unless defined $object;
	return "'$path': invalid value '".__v($object)."', expected '$type'" unless ref $object eq 'ARRAY';
	return undef;
}
sub _validate_object($$$$)
{
	my $self = shift;
	my ($object, $schema, $path, $type) = @_;
	print STDERR "Validation '$path': value '".__v($object)."' against datatype 'object'\n" if $DEVEL;

	return "'$path': received 'null' value, expected '$type'" unless defined $object;
	return "'$path': invalid value '".__v($object)."', expected '$type'" unless ref $object eq 'HASH';
	return undef;
}
sub _validate_custom($$$$)
{
	my $self = shift;
	my ($object, $schema, $path, $type) = @_;
	print STDERR "Validation '$path': value '".__v($object)."' against custom datatype '$type'\n" if $DEVEL;

	return "'$path': received 'null' value, expected '$type'" unless defined $object;
	return "'$path': invalid value '".__v($object)."', expected '$type'" unless ref $object eq $type or (blessed($object) and $object->isa($type));
	return undef;
}

#-------------------------------------------------------------------------------
# Schema combination validators
#-------------------------------------------------------------------------------

sub _validate_allof($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': 'allOf' schema combination for '".__v($object)."'\n" if $DEVEL;

	my $schemas = $schema->{KEY_ALLOF()};
	my $e;

	# allOf algorithm: every schema must match, so report first error
	foreach my $s (@$schemas) {
		$e = $self->_validate($object, $s, $path);
		return $e." [allOf]" if $e;
	}
	return undef;
}
sub _validate_anyof($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': 'anyOf' schema combination for '".__v($object)."'\n" if $DEVEL;

	my $schemas = $schema->{KEY_ANYOF()};
	my $e;

	# anyOf algorithm: at least one schema must match, so report first success
	foreach my $s (@$schemas) {
		$e = $self->_validate($object, $s, $path);
		return undef unless $e;
	}
	return $e." [anyOf]";
}
sub _validate_oneof($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': 'oneOf' schema combination for '".__v($object)."'\n" if $DEVEL;

	my $schemas = $schema->{KEY_ONEOF()};
	my $e;
	my $cnt;

	# oneOf algorithm: exactly one schema must match, so count successes
	foreach my $s (@$schemas) {
		$e = $self->_validate($object, $s, $path);
		$cnt++ unless $e;
	}
	return undef if $cnt == 1;
	return "'$path': invalid value '".__v($object)."', exactly one of the schemas may match [oneOf]";
}
sub _validate_not($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': 'not' schema combination for '".__v($object)."'\n" if $DEVEL;

	my $sub_schema = $schema->{KEY_NOT()};
	my $e = $self->_validate($object, $sub_schema, $path);

	return undef if $e;
	return "'$path': invalid value '".__v($object)."', schema must not match";
}

#-------------------------------------------------------------------------------
# Common validators
#-------------------------------------------------------------------------------

sub _validate_enum($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $options = $schema->{KEY_ENUM()};
	my $v;
	unless (ref $object) {
		$v = __v($object);
		print STDERR "Validation '$path': value '$v' against 'enum'\n" if $DEVEL;
		unless (exists $options->{$v}) {
			return "'$path': invalid enum value '$v', allowed values are: '".join(', ', sort keys(%$options))."'";
		}
	}
	elsif (ref $object eq 'ARRAY') {
		foreach my $val (@$object) {
			$v = __v($val);
			print STDERR "Validation '$path': value '$v' against 'enum'\n" if $DEVEL;
			unless (exists $options->{$v}) {
				return "'$path': invalid enum value '$v', allowed values are: '".join(', ', sort keys(%$options))."'";
			}
		}
	}
	return undef;
}

#-------------------------------------------------------------------------------
# NUMERIC specific validators
#-------------------------------------------------------------------------------

sub _validate_multipleof($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MULTIPLE_OF()};
	my $v  = __v($object);
	print STDERR "Validation '$path': value '$v' against 'multipleOf':'$arg'\n" if $DEVEL;

	return "'$path': received 'null' value, value must be multiple of '$arg'" unless defined $object;
	return undef if ($object - (int($object/$arg) * $arg)) == 0;
	return "'$path': invalid value '$v', value must be multiple of '$arg'";
}
sub _validate_minimum($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MINIMUM()};
	my $v = __v($object);
	print STDERR "Validation '$path': value '$v' against 'minimum':'$arg'\n" if $DEVEL;

	return "'$path': received 'null' value, value must be greater than or equal '$arg'" unless defined $object;
	return undef if $object >= $arg;
	return "'$path': invalid value '$v', value must be greater than or equal '$arg'";
}
sub _validate_maximum($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MAXIMUM()};
	my $v = __v($object);
	print STDERR "Validation '$path': value '$v' against 'maximum':'$arg'\n" if $DEVEL;

	return "'$path': received 'null' value, value must be less than or equal '$arg'" unless defined $object;
	return undef if $object <= $arg;
	return "'$path': invalid value '$v', value must be less than or equal '$arg'";
}
sub _validate_exclusiveminimum($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MINIMUM()};
	my $mod = $schema->{KEY_EXMINIMUM()};
	my $v = __v($object);
	print STDERR "Validation '$path': value '$v' against 'exclusiveMinimum':'$mod':'$arg'\n" if $DEVEL;

	return "'$path': received 'null' value, value must be less than or equal '$arg'" unless defined $object;
	return undef if not $mod and $object >= $arg;
	return undef if $mod and $object > $arg;
	return "'$path': invalid value '$v', value must be greater than or equal '$arg'" if not $mod;
	return "'$path': invalid value '$v', value must be greater than '$arg'" if $mod;
}
sub _validate_exclusivemaximum($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MAXIMUM()};
	my $mod = $schema->{KEY_EXMAXIMUM()};
	my $v = __v($object);
	print STDERR "Validation '$path': value '$v' against 'exclusiveMaximum':'$mod':'$arg'\n" if $DEVEL;

	return "'$path': received 'null' value, value must be less than or equal '$arg'" unless defined $object;
	return undef if not $mod and $object <= $arg;
	return undef if $mod and $object < $arg;
	return "'$path': invalid value '$v', value must be less than or equal '$arg'" if not $mod;
	return "'$path': invalid value '$v', value must be less than '$arg'" if $mod;
}

#-------------------------------------------------------------------------------
# STRING specific validators
#-------------------------------------------------------------------------------

sub _validate_minlength($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MINLENGTH()};
	my $v = __v($object);
	print STDERR "Validation '$path': value '$v' against 'minLength':'$arg'\n" if $DEVEL;

	return undef if length($object) >= $arg;
	return "'$path': invalid value '$v', value must be longer than or equal '$arg'";
}
sub _validate_maxlength($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MAXLENGTH()};
	my $v = __v($object);
	print STDERR "Validation '$path': value '$v' against 'maxLength':'$arg'\n" if $DEVEL;

	return undef if length($object) <= $arg;
	return "'$path': invalid value '$v', value must be shorter than or equal '$arg'";
}
sub _validate_pattern($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_PATTERN()};
	my $v = __v($object);
	print STDERR "Validation '$path': value '$v' against 'pattern':'$arg'\n" if $DEVEL;

	return undef if $object =~ /$arg/;
	return "'$path': invalid value '$v', value must match the '$arg' pattern";
}
sub _validate_format($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg       = $schema->{KEY_FORMAT()};
	my $validator = $FM_VALIDATORS{$arg};
	my $v = __v($object);
	print STDERR "Validation '$path': value '$v' against 'format':'$arg'\n" if $DEVEL;

	# Ignore unknown formats
	return undef unless $validator;

	my $e = $self->$validator($object);
	return undef unless $e;
	return "'$path': invalid value '$v', value must match the '$arg' format";
}

#-------------------------------------------------------------------------------
# ARRAY specific validators
#-------------------------------------------------------------------------------

sub _validate_items($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': 'items'\n" if $DEVEL;

	my ($e, $p, $s);
	my $cnt = -1;
	foreach my $item (@$object) {
		++$cnt;
		$p = join('/', $path."[".($cnt)."]");
		# List validation mode - all objects in list must match the same schema
		if (ref $schema->{KEY_ITEMS()} eq 'HASH') {
			print STDERR "Validation '$path': list validation for item '$cnt'\n" if $DEVEL;
			$s = $schema->{KEY_ITEMS()};
		}
		# Tuple validation mode - each item in hash has its own schema
		elsif (ref $schema->{KEY_ITEMS()} eq 'ARRAY') {
			print STDERR "Validation '$path': tuple validation for item '$cnt'\n" if $DEVEL;
			$s = $schema->{KEY_ITEMS()}->[$cnt];
		}
		# Schema will not be found for additional items
		if ($s) {
			$e = $self->_validate($item, $s, $p);
		}
		return $e if $e;
	}
	return undef;
}
sub _validate_additionalitems($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': 'additionalItems'\n" if $DEVEL;

	my ($e, $p, $s);
	my $cnt = -1;
	# Additional items make sense only for tuples
	if (ref $schema->{KEY_ITEMS()} eq 'ARRAY') {
		# Additional items may have their own schema defined
		if ($schema->{KEY_ADTITEMS()}) {
			foreach my $item (@$object) {
				++$cnt;
				$p = join('/', $path."[".($cnt)."]");
				next if exists $schema->{KEY_ITEMS()}->[$cnt];
				$s = $schema->{KEY_ADTITEMS()};
				$e = $self->_validate($item, $s, $p);
				return $e if $e;
			}
		}
		# Additional items are disabled via 'additionalItems' = false
		else {
			print STDERR "Validation '$path': 'additionalItems' false\n" if $DEVEL;
			my $limit = scalar @{$schema->{KEY_ITEMS()}};
			my $count = scalar @$object;
			return "'$path': no additional items are allowed in array [$count/$limit]" if ($limit < $count);
		}
	}
	else {
		die "Schema key 'additionalItems' makes no sense for list validation\n";
	}
	return undef;
}
sub _validate_minitems($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MINITEMS()};
	print STDERR "Validation '$path': list size against 'minItems':'$arg'\n" if $DEVEL;

	return undef if scalar @$object >= $arg;
	return "'$path': invalid array length, expecting at least '$arg' items";
}
sub _validate_maxitems($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MAXITEMS()};
	print STDERR "Validation '$path': list size against 'maxItems':'$arg'\n" if $DEVEL;

	return undef if scalar @$object <= $arg;
	return "'$path': invalid array length, expecting at most '$arg' items";
}
sub _validate_uniqueitems($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_UNQITEMS()};
	print STDERR "Validation '$path': list items against 'uniqueItems':'$arg'\n" if $DEVEL;

	if ($arg) {
		my (%seen, $p);
		my $cnt = -1;
		foreach my $item (@$object) {
			++$cnt;
			$p = join('/', $path."[".($cnt)."]");
			return "'$p': invalid value '$item', array items must be unique" if $seen{$item};
			$seen{$item}++;
		}
	}
	return undef;
}

#-------------------------------------------------------------------------------
# OBJECT specific validators
#-------------------------------------------------------------------------------

sub _validate_required($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	foreach my $req (@{$schema->{KEY_REQUIRED()}}) {
		print STDERR "Validation '$path': checking for required property '$req'\n" if $DEVEL;
		return "'$path': missing required property '$req'" unless exists $object->{$req};
	}
	return undef;
}
sub _validate_properties($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': checking object 'properties'\n" if $DEVEL;

	my ($e, $p, $s);
	foreach my $prop (sort keys %$object) {
		$p = join('/', $path, $prop);
		if ($schema->{KEY_PROPERTIES()}->{$prop}) {
			$s = $schema->{KEY_PROPERTIES()}->{$prop};
			$e = $self->_validate($object->{$prop}, $s, $p);
			return $e if $e;
		}
	}
	return undef;
}
sub _validate_patternproperties($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': checking object 'patternProperties'\n" if $DEVEL;

	foreach my $prop (sort keys %$object) {
		# 'patternProperties' schema does not apply on exactly defined properties
		next if $schema->{KEY_PROPERTIES()}->{$prop};

		my ($e, $p, $s);
		my @prop_schemas = __pc($prop, $schema->{KEY_PTRPROPERTIES()});
		foreach my $prop_schema (@prop_schemas) {
			$p    = join('/', $path, $prop);
			$s    = $schema->{KEY_PTRPROPERTIES()}->{$prop_schema};

			print STDERR "Validation '$path': checking property '$prop' against property pattern schema '$prop_schema'\n" if $DEVEL;
			$e = $self->_validate($object->{$prop}, $s, $p);
			return $e if $e;
		}
	}
	return undef;
}
sub _validate_additionalproperties($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;
	print STDERR "Validation '$path': checking object 'additionalProperties'\n" if $DEVEL;

	my (@ps, $e, $p, $s);
	foreach my $prop (sort keys %$object) {
		# 'additionalProperties' schema does not apply on exactly defined properties
		next if $schema->{KEY_PROPERTIES()}->{$prop};

		# 'additionalProperties' schema does not apply on pattern properties
		@ps = __pc($prop, $schema->{KEY_PTRPROPERTIES()});
		next if @ps;

		# Additional properties can be disabled via 'additionalProperties' = false
		unless ($schema->{KEY_ADTPROPERTIES()}) {
			unless ($schema->{KEY_PROPERTIES()}->{$prop}) {
				return "'$path': invalid property '$prop', no additional properties are allowed";
			}
		}
		# Otherwise validate the property against explicit schema
		else {
			$p = join('/', $path, $prop);
			$s = $schema->{KEY_ADTPROPERTIES()};
			$e = $self->_validate($object->{$prop}, $s, $p);
			return $e if $e;
		}
	}
	return undef;
}
sub _validate_minproperties($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MINPROPERTIES()};
	print STDERR "Validation '$path': object size against 'minProperties':'$arg'\n" if $DEVEL;

	return undef if scalar keys %$object >= $arg;
	return "'$path': invalid number of object properties, expecting at least '$arg' properties";
}
sub _validate_maxproperties($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_MAXPROPERTIES()};
	print STDERR "Validation '$path': object size against 'maxProperties':'$arg'\n" if $DEVEL;

	return undef if scalar keys %$object <= $arg;
	return "'$path': invalid number of object properties, expecting at most '$arg' properties";
}
sub _validate_dependencies($$$)
{
	my $self = shift;
	my ($object, $schema, $path) = @_;

	my $arg = $schema->{KEY_DEPENDENCIES()};
	foreach my $dependency (keys %$arg) {
		# Process the simple 'property' dependencies
		if (ref $arg->{$dependency} eq 'ARRAY') {
			foreach my $target (@{$arg->{$dependency}}) {
				print STDERR "Validation '$path': checking property dependency '$dependency' => '$target'\n" if $DEVEL;
				return "'$path': unsatisfied property dependency '$dependency' => '$target'" unless exists $object->{$target};
			}
		}
		# Process the 'schema' dependencies
		elsif (ref $arg->{$dependency} eq 'HASH') {
			print STDERR "Validation '$path': checking schema dependency '$dependency'\n" if $DEVEL;
			my $e = $self->_validate($object, $arg->{$dependency}, $path);
			return "'$path': unsatisfied schema dependency '$dependency' => $e" if $e;
		}
	}
	return undef;
}

#-------------------------------------------------------------------------------
# Format validators
#
#  Purpose   : Validate given string against appropriate format
#  Arguments : STRING $arg - validated string
#  Returns   : STRING $error on validation FAILURE, UNDEF on SUCCESS
#-------------------------------------------------------------------------------

sub _vf_datetime($)
{
	my ($self, $arg) = @_;

	return "Invalid value '$arg' for 'datetime' format" unless $arg =~ /$RE_F_DATETIME/;
	return undef;
}
sub _vf_email($)
{
	my ($self, $arg) = @_;

	return "Invalid value '$arg' for 'email' format" unless $arg =~ /$RE_F_EMAIL/;
	return undef;
}
sub _vf_hostname($)
{
	my ($self, $arg) = @_;

	return "Invalid value '$arg' for 'hostname' format" unless $arg =~ /$RE_F_HOSTNAME/;
	return undef;
}
sub _vf_ipv4($)
{
	my ($self, $arg) = @_;

	return "Invalid value '$arg' for 'ipv4' format" unless $arg =~ /$RE_F_IPV4/;
	return undef;
}
sub _vf_ipv6($)
{
	my ($self, $arg) = @_;

	return "Invalid value '$arg' for 'ipv6' format" unless $arg =~ /$RE_F_IPV6/;
	return undef;
}
sub _vf_uri($)
{
	my ($self, $arg) = @_;

	return "Invalid value '$arg' for 'uri' format" unless $arg =~ /$RE_F_URI/;
	return undef;
}

#-------------------------------------------------------------------------------
# Helpers
#-------------------------------------------------------------------------------

# Convert given value into printable form
sub __v($)
{
	my $value = shift;
	return T_UNDEF()  unless defined $value;
	return T_OBJECT() if ref $value eq 'HASH';
	return T_ARRAY()  if ref $value eq 'ARRAY';
	return $value;
}
# Convert given list of values into printable form
sub __lv(@)
{
	return map { __v($_) } @_;
}
# Merge given path chunks
sub __mp(@)
{
	return join('/', @_);
}
# Translate given data type into internal form
sub __tt($)
{
	my $type = shift;
	return T_NUMERIC if $type eq T_NUMBER or $type eq T_INTEGER;
	return $type     if $DT_VALIDATORS{$type};
	return T_CUSTOM;
}
# Check given value against given property patterns
sub __pc($$)
{
	my ($value, $schema) = @_;

	my @patterns;
	return @patterns unless $schema;

	my ($s, $ptrn);
	foreach my $ps (keys %$schema) {
		$s    = $schema->{$ps};
		$ptrn = $s->{KEY_PROPPATTERN()};

		push(@patterns, $ps) if $value =~ /$ptrn/;
	}
	return @patterns;
}

=pod

=back

=cut

1;
