package Log::Channel::Channel;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Log::Channel::Channel - Default implementation of the Log::Channel::Module interface

=head1 SYNOPSIS

  use Log::Channel::Channel;

  my $channel = new Log::Channel::Channel([$filter1,$filter2,$filter3],[$writer1, $writer2]);
  my $channel = new Log::Channel::Channel(undef,[$writer1, $writer2],'accept');
  $channel->write($source, $severity, $message);

=head1 DESCRIPTION

This is the default implementation of the Log::Channel::Module interface. Channel may
contain one or more filters, which are used to restrict the messages, that will be
written to one or more writers. See Log::Filter::(modules) for more information about
log filters, and Log::Writer::(modules) for more information about log writers.

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Channel::Module;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA);
    $VERSION 	= '0.01';
    @ISA 		= ('Log::Channel::Module');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#
use constant ACCEPT => Log::Filter::Module::ACCEPT;
use constant REJECT => Log::Filter::Module::REJECT;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item write($$$) [PUBLIC]

  Usage     : $channel->write($source, $severity, $message);
  Purpose   : Handle given message
  Returns   : 1, if the message passed filters, and 0, if it was rejected
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
  Throws    : Croaks, if invoked on class, or if given invalid arguments
  Comments  : Internally uses pass_filters() method
  See Also  : pass_filters() method
			  Log::Core::Essentials module for permited severity values

=cut

sub write($$$) {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($source, $severity, $message) = @_;
	croak ((caller(0))[3] . ": invalid arguments: $source, $severity, $message\n") unless (defined($source) and defined($severity) and defined($message));

	# If the message passes all channel filters
	if ($self->pass_filters($source, $severity, $message))
	{
		# Write it to all writers in this channel
		foreach my $writer (@{$self->{WRITERS}})
		{
			$writer->write($source, $severity, $message);
		}
		return 1;
	}
	return 0;
}

=item pass_filters($$$) [PUBLIC]

  Usage     : $channel->pass_filters($source, $severity, $message)
  Purpose   : Test, if the given message passess all filters in this channel
  Returns   : Log::Filter::Module::ACCEPT on success, Log::Filter::Module::REJECT on failure
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
  Throws    : Croaks, if invoked on class, or if given invalid arguments
  Comments  : Internally used by write() method
  See Also  : Log::Core::Essentials module for permited severity values

=cut

sub pass_filters($$$) {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($source, $severity, $message) = @_;
	croak ((caller(0))[3] . ": invalid arguments: $source, $severity, $message\n") unless (defined($source) and defined($severity) and defined($message));

	my ($result, $continue);
	foreach my $filter (@{$self->{FILTERS}})
	{
		($result, $continue) = $filter->accept($source, $severity, $message);
		unless ($continue)
		{
			return $result;
		}
	}
	return $self->{DEFAULT_FILTER_RESULT};
}

=item _init() [PROTECTED]

  Usage     : Used from constructor as follows: return $self->_init(@_);
  Purpose   : Initialize newly created channel instance
  Returns   : $self
  Arguments : array reference $filters - Chain of channel filters
			  array reference $writers - List of channel writers
			  enum $dfr - Default filter chain result (Log::Filter::Module::ACCEPT|REJECT)
  Throws    : Croaks, if invoked on class, or if given invalid arguments
  Comments  : Internally used by parent`s new() method
  See Also  : Log::Filter::Module module for default filter chain result values

=cut

sub _init {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($filters, $writers, $dfr) = @_;

	# Filters may be undefined, it means no filters in the chain at all
	unless (defined($filters))
	{
		$self->{FILTERS} =  [];
	}
	else {
		croak ((caller(0))[3] . ": invalid arguments, filters must be array reference") unless (ref($filters) eq 'ARRAY');
		map { croak ((caller(0))[3] . ": invalid filter") unless ($_->isa('Log::Filter::Module')); } @{$filters};
		$self->{FILTERS} = $filters;
	}

	# Writers must be given, otherwiseit would not make sense
	croak ((caller(0))[3] . ": invalid arguments, writers must be array reference") unless (ref($writers) eq 'ARRAY');
	croak ((caller(0))[3] . ": invalid arguments, empty writers") unless (scalar @{$writers});
	map { croak ((caller(0))[3] . ": invalid writer") unless ($_->isa('Log::Writer::Module')); } @{$writers};
	$self->{WRITERS} = $writers;

	# Default filter chain result is to reject the message
	$self->{DEFAULT_FILTER_RESULT} = REJECT;
	if (defined($dfr) and uc($dfr) eq 'ACCEPT')
	{
		$self->{DEFAULT_FILTER_RESULT} = ACCEPT;
	}
	return $self;
}

=pod

=back

=cut

1;
