package Log::Channel::Module;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Log::Channel::Module - Base module defining mandatory interface for all log channel modules

=head1 SYNOPSIS

  use Log::Channel::Module;

  BEGIN {
    @ISA = ('Log::Channel::Module');
  }

=head1 DESCRIPTION

Classes extending this base class may be used as CHANNELS in loging framework.
Channels are used for message processing. Loging framework dispatches each
message to all registered channels, which can then perform filtering, thresholding
and write message to different destinations. See default channel implementation in
class Log::Channel::Channel for details.

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Filter::Module;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION);
    $VERSION = '0.01';
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item new() [PUBLIC,STATIC]

  Usage     : my $channel = Log::Channel::(module)->new($channel_filters, $channel_writers, $dfr);
  Purpose   : Create new instance of log channel
  Returns   : Reference to the new instance
  Arguments : All arguments are passed to the _init() function
  Throws    : Croaks if invoked on object
  Comments  : This method should not be overloaded, descendants should implement their version of _init() method instead
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my $self = bless ({}, $class);
	return $self->_init(@_);
}


=item write($$$) [ABSTRACT, PUBLIC]

  Usage     : $channel->write($source, $severity, $message);
  Purpose   : Handle given message
  Returns   : Must return 1, if the message passed filters, and 0, if it was rejected
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : Log::Core::Essentials module for permited severity values

=cut

sub write($$$) {
	my $self = shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

=item pass_filters() [ABSTRACT, PUBLIC]

  Usage     : $channel->pass_filters($source, $severity, $message)
  Purpose   : Test, if the given message passess all filters in this channel
  Returns   : Log::Filter::Module::ACCEPT on success, Log::Filter::Module::REJECT on failure
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : Log::Core::Essentials module for permited severity values

=cut

sub _pass_filters($$$) {
	my $self = shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

=item _init() [ABSTRACT, PROTECTED]

  Usage     : Used from constructor as follows: return $self->_init(@_);
  Purpose   : Initialize newly created channel instance
  Returns   : Must return $self
  Arguments : Unknown, depend on the implementation
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : new() method

=cut

sub _init {
	my $self = shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

=pod

=back

=cut

1;
