package Log::Core::Essentials;
use strict;
use warnings;

#*******************************************************************************
#
# DOCUMENTATION SECTION
#
#*******************************************************************************

=head1 NAME

Log::Core::Essentials - Essential methods and constants for whole logging framework

=head1 SYNOPSIS

 use Log::Core::Essentials;

 # Get list of all valid severity values
 my @severities = Log::Core::Essentials->severities();

 # Translate given severity into integer representation
 my $result = Log::Core::Essentials->severity_as_int('EMERG');

 # Translate given severity into string representation
 my $result = Log::Core::Essentials->severity_as_str(4);

=head1 DESCRIPTION

This class contains essential contants for whole logging framework, and
methods for converting message severities from string to their integer
representations, and vice versa.

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
#
# INITIALIZATION AND CLEANUP SECTION
#
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.01';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

# Log severity integer values
use constant EMAIL 		=> 8;	# special severity - send email
use constant EMERG 		=> 7;	# system is unusable
use constant ALERT		=> 6;	# action must be taken immediately
use constant CRIT		=> 5;	# critical conditions
use constant ERROR		=> 4;	# error conditions
use constant WARNING	=> 3;	# warning conditions
use constant NOTICE		=> 2;	# normal, but significant, condition
use constant INFO		=> 1;	# informational message
use constant DEBUG		=> 0;	# debug-level message

# Log severity string values
use constant EMAIL_STR 		=> 'EMAIL';		# special severity - send email
use constant EMERG_STR 		=> 'EMERG';		# system is unusable
use constant ALERT_STR		=> 'ALERT';		# action must be taken immediately
use constant CRIT_STR		=> 'CRIT';		# critical conditions
use constant ERROR_STR		=> 'ERROR';		# error conditions
use constant WARNING_STR	=> 'WARNING';	# warning conditions
use constant NOTICE_STR		=> 'NOTICE';	# normal, but significant, condition
use constant INFO_STR		=> 'INFO';		# informational message
use constant DEBUG_STR		=> 'DEBUG';		# debug-level message

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Translation table from string severity labels to integer representations
# We MUST NOT USE '=>' symbol in the hash definition, otherwise the constant
# names left of it would be automatically quoted and interpreted as strings.
my %SEVERITY2INT = ( EMAIL_STR		, EMAIL,
					 EMERG_STR		, EMERG,
					 ALERT_STR 		, ALERT,
					 CRIT_STR		, CRIT,
					 ERROR_STR		, ERROR,
					 WARNING_STR 	, WARNING,
					 NOTICE_STR		, NOTICE,
					 INFO_STR		, INFO,
					 DEBUG_STR		, DEBUG,
					);

# Translation table from integer representations to string severity labels
my @INT2SEVERITY = ( DEBUG_STR,
					 INFO_STR,
					 NOTICE_STR,
					 WARNING_STR,
					 ERROR_STR,
					 CRIT_STR,
					 ALERT_STR,
					 EMERG_STR,
					 EMAIL_STR,
					);

#*******************************************************************************
#
# IMPLEMENTATION SECTION
#
#*******************************************************************************

=item severities() [PUBLIC,STATIC]

  Usage     : Log::Core::Essentials->severities();
  Purpose   : Get list of all valid severities
  Arguments : None
  Returns   : ARRAY of STRINGS @severities
  Throws    : Croaks, if not invoked on class

=cut

sub severities()
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
	return @INT2SEVERITY;
}

=item severity_as_int($) [PUBLIC,STATIC]

  Usage     : Log::Core::Essentials->severity_as_int('DEBUG')
  Purpose   : Convert severity value in integer or string into corresponding integer value
  Arguments : MIXED $severity - integer or string severity to convert, strings are case insensitive
  Returns   : UNDEF on FAILURE, INTEGER representation of given severity on SUCCESS
  Throws    : Croaks, if not invoked on class

=cut

sub severity_as_int($)
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
	my ($severity,) = @_;

	# Return undefined value if severity was not specified
	return undef unless (defined($severity));

	# Perform conversion to integer only if the argument is empty string or not a number
	if ($severity =~ /^$|[^\d]+/) {
		$severity = $SEVERITY2INT{uc($severity)};
	}

	# Return undefined value if the integer is not a valid severity value
	return undef unless (defined($severity) and defined($INT2SEVERITY[$severity]));

	# Othervise return valid value
	return $severity;
}

=item severity_as_str [PUBLIC,STATIC]

  Usage     : Log::Core::Essentials->severity_as_str(7)
  Purpose   : Convert severity value in integer or string into corresponding string value
  Arguments : MIXED $severity - integer or string severity to convert, strings are case insensitive
  Returns   : UNDEF on FAILURE, STRING representation of given severity on SUCCESS
  Throws    : Croaks, if not invoked on class

=cut

sub severity_as_str($)
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if blessed($class);
	my ($severity,) = @_;

	# Return undefined value if severity was not specified
	return undef unless (defined($severity));

	# Normalize the severity to the uppercase
	$severity = uc($severity);

	# Perform conversion only if the argument is a number
	if ($severity =~ /[\d]+/) {
		$severity = $INT2SEVERITY[int($severity)];
	}

	# Return undefined value if the integer is not a valid severity value
	return undef unless (defined($severity) and defined($SEVERITY2INT{$severity}));

	# Othervise return valid value
	return $severity;
}

=pod

=back

=cut

1;
