package Log::Filter::All;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Log::Filter::All - Filter, that accepts all log messages (AUTHORITATIVE)

=head1 SYNOPSIS

  use Log::Filter::All;

  my $filter = new Log::Filter::All();
  my ($result, $continue) = $filter->accept($source, $severity, $message);

=head1 DESCRIPTION

This filter accepts all messages. It is authoritative, so it`s result is
final and evaluation will not continue.

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Filter::Module;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA);
    $VERSION 	= '0.01';
    @ISA 		= ('Log::Filter::Module');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item accept($$$) [PUBLIC]

  Usage     : my ($result, $continue) = $filter->accept($source, $severity, $message);
  Purpose   : Check, if the filter will accept the given message
  Returns   : Array (Log::Filter::Module::ACCEPT, Log::Filter::Module::FINISH)
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
  Throws    : Croaks, if invoked on class, or if given invalid arguments
  Comments  :
  See Also  :

=cut

sub accept($$$) {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($source, $severity, $message) = @_;
	croak ((caller(0))[3] . ": invalid arguments") unless (defined($source) and defined($severity) and defined($message));

	# Always return success, our result is authoritative
	return (Log::Filter::Module::ACCEPT, Log::Filter::Module::FINISH);
}

=item _init() [PROTECTED]

  Usage     : Used from constructor as follows: return $self->_init(@_);
  Purpose   : Initialize newly created filter instance
  Returns   : $self
  Arguments : None
  Throws    : Croaks, if invoked on class
  Comments  : Internally used by parent`s new() method
  See Also  :

=cut

sub _init {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self;
}

=pod

=back

=cut

1;
