package Log::Filter::Module;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Log::Filter::Module - Base module defining mandatory interface for all log filtering modules

=head1 SYNOPSIS

  use Log::Filter::Module;

  BEGIN {
    @ISA = ('Log::Filter::Module');
  }

=head1 DESCRIPTION

Classes extending this base class may be used as FILTERS in loging framework.
Filters are used in channels for restricting messages, that will be passed
to the writers.

Filters are intended to be used in chains. Most important is the accept()
method, which returns array containing the result of the filtering (message
is ACCEPTed or REJECTed) and flag determining, if the evaluation should
CONTINUE, or FINISH. The caller is responsible for appropriate reaction.

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION);
    $VERSION 	= '0.01';
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#
use constant ACCEPT 		=> 1;		# Message passed the filter
use constant REJECT 		=> 0;		# Message did not pass the filter

use constant REQUIRED 		=> 'req';	# Filter is required
use constant SUFFICIENT		=> 'suf';	# Filter is sufficient

use constant CONTINUE 		=> 1;		# Evaluation must continue with next filter in chain
use constant FINISH 		=> 0;		# Evaluation is completed

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item new() [PUBLIC,STATIC]

  Usage     : my $filter = Log::Filter::(module)->new(... arguments);
  Purpose   : Create new instance of log filter
  Returns   : Reference to the new instance
  Arguments : All arguments are passed to the _init() method
  Throws    : Croaks, if invoked on object
  Comments  : This method should not be overloaded, descendants should implement their version of _init() method instead
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my $self = bless ({}, $class);
	return $self->_init(@_);
}

=item accept($$$) [ABSTRACT,PUBLIC]

  Usage     : my ($result, $continue) = $filter->accept($source, $severity, $message);
  Purpose   : Check, if the filter will accept the given message
  Returns   : Array (ACCEPT|REJECT,CONTINUE|FINISH)
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  :

=cut

sub accept($$$) {
	my $self = shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

=item _init() [ABSTRACT,PROTECTED]

  Usage     : Used from constructor as follows: return $self->_init(@_);
  Purpose   : Initialize newly created filter instance
  Returns   : Must return $self
  Arguments : Unknown, depend on the implementation
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : new() method

=cut

sub _init {
	my $self = shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

=pod

=back

=cut

1;
