package Log::Filter::Re;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Log::Filter::Re - Test given message field for values according the given PCRE patterns

=head1 SYNOPSIS

  use Log::Filter::Re;

  my $filter = new Log::Filter::Re('source','modul.*[a|b]');
  my $filter = new Log::Filter::Re('severity','error');
  my $filter = new Log::Filter::Re('message',['127.0.0.1','^notice'],'i r');

  my ($result, $continue) = $filter->accept($source, $severity, $message);

=head1 DESCRIPTION

This module accepts messages, whose given field comply to the at least one of the given
PCRE patterns.

There are two flags, that are altering the avaluation process:

 i - Invert the result of the evaluation -
		ACCEPT => REJECT
		REJECT => ACCEPT
 r - By default, module`s evaluation is SUFFICIENT -
		ACCEPT => FINISH
		REJECT => CONTINUE
	 By setting the flag 'r', module`s evaluation will be REQUIRED -
		ACCEPT => CONTINUE
		REJECT => FINISH

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Filter::Module;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA);
    $VERSION 	= '0.01';
    @ISA 		= ('Log::Filter::Module');
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#
use constant ACCEPT 		=> Log::Filter::Module::ACCEPT;
use constant REJECT 		=> Log::Filter::Module::REJECT;

use constant REQUIRED 		=> Log::Filter::Module::REQUIRED;
use constant SUFFICIENT		=> Log::Filter::Module::SUFFICIENT;

use constant CONTINUE 		=> Log::Filter::Module::CONTINUE;
use constant FINISH 		=> Log::Filter::Module::FINISH;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Translation table from message fields to indexes
my %FIELD2INDEX = (	'SOURCE' 	=> 0,
					'SEVERITY'	=> 1,
					'MESSAGE'	=> 2,
					);
# Translation table from indexes to message fields
my @INDEX2FIELD = ( 'SOURCE',
					'SEVERITY',
					'MESSAGE',
					);

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item accept($$$) [PUBLIC]

  Usage     : my ($result, $continue) = $filter->accept($source, $severity, $message);
  Purpose   : Check, if the filter will accept the given message
  Returns   : Array (Log::Filter::Module::ACCEPT|REJECT, Log::Filter::Module::FINISH|CONTINUE)
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
  Throws    : Croaks, if invoked on class, or if given invalid arguments
  Comments  :
  See Also  :

=cut

sub accept($$$) {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($source, $severity, $message) = @_;
	croak ((caller(0))[3] . ": invalid arguments") unless (defined($source) and defined($severity) and defined($message));

	my @fields = @_;
	my $resolution;
	my $continue;

	# Walk through all PCRE patterns
	foreach my $pattern (@{$self->{PATTERNS}})
	{
		# If the given field complies to the current pattern
		if (@fields[$self->{FIELD}] =~ /$pattern/)
		{
			unless ($self->{INVERT})
			{
				$resolution = ACCEPT;
				$continue 	= ($self->{MODE} eq SUFFICIENT) ? FINISH : CONTINUE;
			}
			else {
				$resolution = REJECT;
				$continue 	= ($self->{MODE} eq REQUIRED) ? FINISH : CONTINUE;
			}
			return ($resolution, $continue);
		}
	}

	unless ($self->{INVERT})
	{
		$resolution = REJECT;
		$continue 	= ($self->{MODE} eq REQUIRED) ? FINISH : CONTINUE;
	}
	else {
		$resolution = ACCEPT;
		$continue 	= ($self->{MODE} eq SUFFICIENT) ? FINISH : CONTINUE;
	}
	return ($resolution, $continue);
}

=item _init() [PROTECTED]

  Usage     : Used from constructor as follows: return $self->_init(@_);
  Purpose   : Initialize newly created filter instance
  Returns   : $self
  Arguments : enum $field - name of the field to test (source | severity | message)
			  array reference $patterns - list of PCRE patterns
			  string $flags - flags separated with spaces
					i - invert the result
					r - required
  Throws    : Croaks, if invoked on class or invalid arguments given
  Comments  : Internally used by parent`s new() method
  See Also  : Log::Filter::Re->_init() method

=cut

sub _init {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($field, $patterns, $flags) = @_;
	croak ((caller(0))[3] . ": invalid arguments") unless ($field and $patterns);

	# First try to parse and init mandatory arguments
	$self->{FIELD} 		= $self->_field_int($field);
	$self->{PATTERNS} 	= [];
	my @ptrns;
	if(ref($patterns) eq 'ARRAY') 	{ push(@ptrns, @{$patterns}); }
	else 							{ push(@ptrns, $patterns);	}
	push(@{$self->{PATTERNS}}, map { qr/$_/ } @ptrns);
	croak ((caller(0))[3] . ": invalid arguments") unless (defined($self->{FIELD}) and scalar @{$self->{PATTERNS}});

	# Set the engine flags to default falues
	$self->{MODE} = Log::Filter::Module::SUFFICIENT;
	$self->{INVERT} = 0;

	# Setup the aditional engine flags
	if ($flags)
	{
		foreach my $flag (split(/[,; ]+/, $flags))
		{
			if 		($flag =~ /^r$|^req$/) { $self->{MODE} = Log::Filter::Module::REQUIRED; }
			elsif 	($flag =~ /^i$|^inv$/) { $self->{INVERT} = 1; }
			else 	{ croak ((caller(0))[3] . ": unknown flag '$flag'"); }
		}
	}
	return $self;
}

=item _field_int($) [PROTECTED]

  Usage     : $self->_field_int($field)
  Purpose   : Internal method for converting log field to integer index
  Returns   : integer index
  Arguments : enum $field - name of the field to translate (source | severity | message)
  Throws    : Croaks, if invoked on class
  Comments  : Should be used only internally, used from _init() method
  See Also  :

=cut

sub _field_int($) {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my $field = uc(shift);

	# Return undefined value if field was not specified
	return undef unless (defined($field));

	# Perform conversion only if the argument is not a number
	if ($field =~ /^$|[^\d]+/)
	{
		$field = $FIELD2INDEX{uc($field)};
	}

	# Return undefined value if the integer is not a valid field value
	return undef unless (defined($field) and defined($INDEX2FIELD[$field]));

	# Othervise return valid value
	return $field;
}


=pod

=back

=cut

1;
