package Log::Filter::Threshold;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Log::Filter::Threshold - Filter log messages based on minimal and maximal severity threshold

=head1 SYNOPSIS

  use Log::Filter::Threshold;

  my $filter = new Log::Filter::Threshold('error');

  my ($result, $continue) = $filter->accept($source, $severity, $message);

=head1 DESCRIPTION

This module accepts messages, whose severity level is greater or equal than
the given threshold.

There are two flags, that are altering the avaluation process:

 i - Invert the result of the evaluation -
		ACCEPT => REJECT
		REJECT => ACCEPT
 r - By default, module`s evaluation is SUFFICIENT -
		ACCEPT => FINISH
		REJECT => CONTINUE
	 By setting the flag 'r', module`s evaluation will be REQUIRED -
		ACCEPT => CONTINUE
		REJECT => FINISH

Note.: Inverted eveluation is lower than threshold (not lower or equal).

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Core::Essentials;
use Log::Filter::Module;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA);
    $VERSION 	= '0.01';
    @ISA 		= ('Log::Filter::Module');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#
use constant ACCEPT 		=> Log::Filter::Module::ACCEPT;
use constant REJECT 		=> Log::Filter::Module::REJECT;

use constant REQUIRED 		=> Log::Filter::Module::REQUIRED;
use constant SUFFICIENT		=> Log::Filter::Module::SUFFICIENT;

use constant CONTINUE 		=> Log::Filter::Module::CONTINUE;
use constant FINISH 		=> Log::Filter::Module::FINISH;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item accept($$$) [PUBLIC]

  Usage     : my ($result, $continue) = $filter->accept($source, $severity, $message);
  Purpose   : Check, if the filter will accept the given message
  Returns   : Array (Log::Filter::Module::ACCEPT|REJECT, Log::Filter::Module::FINISH|CONTINUE)
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
  Throws    : Croaks, if invoked on class, or if given invalid arguments
  Comments  :
  See Also  :

=cut

sub accept($$$) {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($source, $severity, $message) = @_;
	croak ((caller(0))[3] . ": invalid arguments") unless (defined($source) and defined($severity) and defined($message));

	my $resolution;
	my $continue;
	my $level = Log::Core::Essentials->severity_as_int($severity);
	croak ((caller(0))[3] . ": invalid severity value: '$severity'") unless (defined($level));

	# Threshold is lower or equal than the severity level of the log message
	if ($self->{THRESHOLD} <= $level)
	{
		unless ($self->{INVERT})
		{
			$resolution = ACCEPT;
			$continue 	= ($self->{MODE} eq SUFFICIENT) ? FINISH : CONTINUE;
		}
		else {
			$resolution = REJECT;
			$continue 	= ($self->{MODE} eq REQUIRED) ? FINISH : CONTINUE;
		}
		return ($resolution, $continue);
	}

	unless ($self->{INVERT})
	{
		$resolution = REJECT;
		$continue 	= ($self->{MODE} eq REQUIRED) ? FINISH : CONTINUE;
	}
	else {
		$resolution = ACCEPT;
		$continue 	= ($self->{MODE} eq SUFFICIENT) ? FINISH : CONTINUE;
	}
	return ($resolution, $continue);
}

=item _init() [PROTECTED]

  Usage     : Used from constructor as follows: return $self->_init(@_);
  Purpose   : Initialize newly created filter instance
  Returns   : $self
  Arguments : string $threshold - threshold
			  string $flags - flags separated with spaces
					i - invert the result
					r - required
  Throws    : Croaks, if invoked on class or invalid arguments given
  Comments  : Internally used by parent`s new() method
  See Also  :

=cut

sub _init {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($threshold, $flags) = @_;
	croak ((caller(0))[3] . ": wrong arguments") unless (defined($threshold));

	# Convert threshold into it`s corresponding integer value
	$threshold = Log::Core::Essentials->severity_as_int($threshold);
	croak ((caller(0))[3] . ": wrong arguments") unless (defined($threshold));

	# Setup threshold
	$self->{THRESHOLD} = $threshold;

	# Set the engine flags to default falues
	$self->{MODE} = SUFFICIENT;
	$self->{INVERT} = 0;

	# Setup the aditional engine flags
	if ($flags)
	{
		foreach my $flag (split(/[,; ]+/, $flags))
		{
			if 		($flag =~ /^r$|^req$/) { $self->{MODE} = REQUIRED; }
			elsif 	($flag =~ /^i$|^inv$/) { $self->{INVERT} = 1; }
			else 	{ croak ((caller(0))[3] . ": unknown flag '$flag'"); }
		}
	}
	return $self;
}

=pod

=back

=cut

1;
