package Log::Loger;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Log::Loger - Powerful, but lightweight and easy to use Perl loging framework

=head1 SYNOPSIS

  use Log::Loger;

  # Usage through static methods
  Log::Loger->filter('filter1','threshold','WARNING','r');
  Log::Loger->filter('filter2','source','module_a, module_b');
  Log::Loger->writer('writer1','std');
  my $ch_id = Log::Loger->channel('filter1,filter2','writer1');
  Log::Loger->channel_remove($ch_id);

  Log::Loger->log('source', 'ERROR', 'message');
  Log::Loger->log_warn('source', 'ERROR', 'message');
  Log::Loger->log_die('source', 'ERROR', 'message');
  Log::Loger->log_exit('source', 'ERROR', 'message');

  # Quickly turn on the logging of all messages to STDERR
  Log::Loger->on();

  # Usage through instance methods
  my $instance = new Log::Loger();

  $instance->filter('filter1','threshold','WARNING','r');
  $instance->filter('filter2','source','module_a, module_b');
  $instance->filter('filter3','message','9');
  $instance->writer('writer1','std');
  my $ch_id = $instance->channel(['filter1','filter2','filter3'],['writer1']);
  $instance->channel_remove($ch_id);

  $instance->log('source', 'ERROR', 'message');
  $instance->log_warn('source', 'ERROR', 'message');
  $instance->log_die('source', 'ERROR', 'message');
  $instance->log_exit('source', 'ERROR', 'message');

  # Quickly turn on the logging of all messages to STDERR
  $instance->on();

=head1 DESCRIPTION

This log framework is designed to be easy to use and extend to your needs.
It may be used through static or instance methods. The design was inspired
by syslog-ng loger.

All incoming messages are directed to the CHANNELS. CHANNELS are composed
of one or more chained FILTERS and one or more WRITERS. When message
enters the CHANNEL, all FILTERS are consulted in order to determine, if the
message will be passed to the WRITERS to be written to the destinations.

By default, FILTERS work in the SUFFICIENT mode - their result is authoritative
and if the message passes particular filter, result is sufficient and no
other filters are considered. If the message is rejected, default channel
policy aplies. This default behaviour may be altered using the 'r' flag
passed to most of the filters. In the REQUIRED mode, if the message is rejected,
result is authoritative and no more filters are considered. If the message
passes, next filter in row is considered.

Most of the filters also accept the 'i' flag, which inverts the result. It means,
that the filter will return REJECT instead of ACCEPT and vice versa. Please
see the documentation of particular filter for more information.

=head1 AUTHOR

    Jan Mach
    Cesnet, z.s.p.o
    jan.mach@cesnet.cz
    http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp qw(carp croak cluck confess longmess);
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

# Core logging contants and utilities
use Log::Core::Essentials;

# Log filtering modules
use Log::Filter::All;
use Log::Filter::None;
use Log::Filter::Set;
use Log::Filter::Source;
use Log::Filter::Severity;
use Log::Filter::Re;
use Log::Filter::Message;
use Log::Filter::Threshold;

# Log writing modules
use Log::Writer::Email;
use Log::Writer::Handle;
use Log::Writer::File;
use Log::Writer::Std;
use Log::Writer::Syslog;

# Log channel modules
use Log::Channel::Channel;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION);
    $VERSION = '0.01';
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Storage for global writers, filters and channel configurations
my $WRITERS		 = {};
my $WRITER_TYPES = {};
my $FILTERS		 = {};
my $CHANNELS 	 = {};
my $CH_QUEUE 	 = [];

# Sequential indentifiers
my $SEQ	= 0;

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $loger = Log::Loger->new();
  Purpose   : Create new instance of log loger
  Returns   : Reference to the new instance
  Arguments : None
  Throws    : Croaks if invoked on object
  Comments  : Instance is initialized through filter(), writer() and channel() methods
  See Also  : filter(), writer(), channel() methods

=cut

sub new()
{
    my $class 	= shift;
	confess "Class method not invoked on class" if blessed($class);

	my $self = bless ({}, $class);
	$self->{WRITERS}      = {};
	$self->{WRITER_TYPES} = {};
	$self->{FILTERS}      = {};
	$self->{CHANNELS}     = {};
	$self->{CH_QUEUE}     = [];
	$self->{SEQ}          = 0;
	return $self;
}

=item on() [PUBLIC,HYBRID]

  Usage     : $loger->on();
              Log::Loger->on();
  Purpose   : Quickly turn on logging of all messages to the STDERR
  Arguments : None
  Returns   : Nothing
  Throws    : Nothing
  Comments  : Hybrid method, may be called on class or instance

=cut

sub on()
{
    my $invocant = shift;

	# Default writer writes to STDERR
	$invocant->writer('writer_on','std');
	# Channel without filter accepting all messages
	$invocant->channel(undef,'writer_on','ACCEPT');
}

=item get_debug() [PUBLIC,STATIC]

  Usage     : my ($writers, $filters, $channel) = Log::Loger->get_debug();
  Purpose   : Get references to the writers, filters and channels
  Returns   : References to the the writers, filters and channels
  Arguments : None
  Throws    : Croaks if invoked on object

=cut

sub get_debug()
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	return ($WRITERS, $FILTERS, $CHANNELS, $CH_QUEUE);
}

=item channel($$;$) [PUBLIC,HYBRID]

  Usage     : Log::Loger->channel('filter1,filter2','writer1'); or
			  $instance->channel(['filter1','filter2','filter3'],['writer1']);
  Purpose   : Create new channel from given filters and writers
  Returns   : Nothing
  Arguments : space separated string | array reference $filters - identifiers of the filters
			  space separated string | array reference $writers - identifiers of the writers
			  enum $dfr - default filter chain result (see. Log::Channel::Channel for details)
  Throws    : Dies, if given invalid arguments

=cut

sub channel($$;$)
{
	my $invocant = shift;
	my ($filters, $writers, $dfr) = @_;
	confess "Invalid arguments" unless (defined($writers));

	my $filters_available = (ref $invocant) ? $invocant->{FILTERS} : $FILTERS;
	my $writers_available = (ref $invocant) ? $invocant->{WRITERS} : $WRITERS;

	my $channel_filters = [];
	my $channel_writers = [];

	# Filter identifiers may be given in array
	if (ref($filters) eq 'ARRAY') {
		map { croak ((caller(0))[3] . ": invalid filter id" . ((defined($_))?": '$_'":'')) unless $filters_available->{$_}; } @{$filters};
		map { push(@$channel_filters, $filters_available->{$_} ); } @{$filters};
	}
	# Or as space separated string
	elsif ($filters) {
		map { croak ((caller(0))[3] . ": invalid filter id" . ((defined($_))?": '$_'":'')) unless $filters_available->{$_}; } split(/[,; ]+/, $filters);
		map { push(@$channel_filters, $filters_available->{$_} ); } split(/[,; ]+/, $filters);
	}

	# Writers identifiers may be given in array
	if (ref($writers) eq 'ARRAY') {
		map { croak ((caller(0))[3] . ": invalid writer id" . ((defined($_))?": '$_'":'')) unless $writers_available->{$_}; } @{$writers};
		map { push(@$channel_writers, $writers_available->{$_} ); } @{$writers};
	}
	# Or as space separated string
	elsif ($writers) {
		map { croak ((caller(0))[3] . ": invalid writer id" . ((defined($_))?": '$_'":'')) unless $writers_available->{$_}; } split(/[,; ]+/, $writers);
		map { push(@$channel_writers, $writers_available->{$_} ); } split(/[,; ]+/, $writers);
	}

	my $id;
	my $channel = Log::Channel::Channel->new($channel_filters, $channel_writers, $dfr);

	# If this method was called on behalf of some object, add the module within it`s channels
	if (ref $invocant) {
		$id = ++$invocant->{SEQ};
		$invocant->{CHANNELS}->{$id} = $channel;
		push(@{$invocant->{CH_QUEUE}}, $id);
	}
	# Otherwise put it to the static class channel storage
	else {
		$id = ++$SEQ;
		$CHANNELS->{$id} = $channel;
		push(@$CH_QUEUE, $id);
	}
	return $id;
}

=item filter [PUBLIC, HYBRID]

  Usage     : Log::Loger->filter('filter1','threshold','WARNING','r'); or
			  $instance->filter('filter1','threshold','WARNING','r');
  Purpose   : Create new instance of the specific filter with given ID to be used in the channel() method
  Returns   : Nothing
  Arguments : string $filter_id - identifier of the filter
			  string $filter_type - last name of the filter module
			  (... specific filter arguments - will be passed to the constructor)
  Throws    : Dies, if given invalid arguments
  Comments  :
  See Also  : channel() method

=cut

sub filter {
	my $invocant = shift;
	my ($filter_id, $filter_type, @arguments) = @_;
	confess "Invalid arguments" unless (defined($filter_id) and defined($filter_type));

	# Attempt to create instance of writer module. No checking here on purpose
	$filter_type = ucfirst($filter_type);
	my $filter_module = "Log::Filter::$filter_type";
	my $module = $filter_module->new(@arguments);
	confess "Invalid filter $filter_module" unless ($module->isa('Log::Filter::Module'));
	if (ref $invocant) {
		$invocant->{FILTERS}->{$filter_id} = $module;
	}
	else {
		$FILTERS->{$filter_id} = $module;
	}
}

=item writer [PUBLIC, HYBRID]

  Usage     : Log::Loger->writer('writer1','std'); or
			  $instance->writer('writer1','std');
  Purpose   : Create new instance of the specific writer with given ID to be used in the channel() method
  Returns   : Nothing
  Arguments : string $writer_id - identifier of the writer
			  string $writer_type - last name of the writer module
			  (... specific filter arguments - will be passed to the constructor)
  Throws    : Dies, if given invalid arguments
  Comments  :
  See Also  : channel() method

=cut

sub writer {
	my $invocant = shift;
	my ($writer_id, $writer_type, @arguments) = @_;
	confess "Invalid arguments" unless (defined($writer_id) and defined($writer_type));

	# Attempt to create instance of writer module. No checking here on purpose
	$writer_type = ucfirst($writer_type);
	my $writer_module = "Log::Writer::$writer_type";
	my $module = $writer_module->new(@arguments);
	confess "Invalid writer $writer_module" unless ($module->isa('Log::Writer::Module'));
	if (ref $invocant) {
		$invocant->{WRITERS}->{$writer_id} = $module;
		push(@{$invocant->{WRITER_TYPES}->{lc($writer_type)}}, $module);
	}
	else {
		$WRITERS->{$writer_id} = $module;
		push(@{$WRITER_TYPES->{lc($writer_type)}}, $module);
	}
}

=item channel_count() [PUBLIC, HYBRID]

  Usage     : Log::Loger->channel_count(); or
			  $instance->channel_count();
  Purpose   : Get number of log channels created so far
  Returns   : integer - number of log channels created so far
  Arguments : None
  Throws    : Nothing
  Comments  :
  See Also  :

=cut

sub channel_count()
{
	my $invocant = shift;
	if (ref $invocant) {
		return scalar @{$invocant->{CH_QUEUE}};
	}
	else {
		return scalar @$CH_QUEUE;
	}
}

=item channel_remove($) [PUBLIC, HYBRID]

  Usage     : Log::Loger->channel_remove($id); or
			  $instance->channel_remove($id);
  Purpose   : Remove channel with given ID
  Returns   : Nothing
  Arguments : integer $id - identifier of the channel to remove
  Throws    : Dies, if channel with given ID does not exist
  Comments  :
  See Also  :

=cut

sub channel_remove($)
{
	my $invocant = shift;
	my $id = shift;
	confess "ID of the channel to remove must be given as argument" unless $id;

	my $channel_queue;
	my $channels_current;
	if (ref $invocant) {
		$channel_queue = $invocant->{CH_QUEUE};
		$channels_current = $invocant->{CHANNELS};
	}
	else {
		$channel_queue = $CH_QUEUE;
		$channels_current = $CHANNELS;
	}

	confess "Channel with ID '$id' does not exist" unless exists $channels_current->{$id};
	delete $channels_current->{$id};

	my @a = @$channel_queue;
	my ($index, ) = grep { $a[$_] eq $id } 0..$#a;
	splice(@a, $index, 1) if $index;
	if (ref $invocant) 	{ $invocant->{CH_QUEUE} = \@a; }
	else 				{ $CH_QUEUE = \@a; }
}

=item log [PUBLIC, HYBRID]

  Usage     : Log::Loger->log('source', 'ERROR', 'message');
			  $instance->log('source', 'ERROR', 'message');
  Purpose   : Write new log message to all defined channels
  Returns   : Nothing
  Arguments : string $source - Name of the source of the message
              enum $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              string $message - Message
			  enum $action - action to take after the message is written (exit|die|warn|croak)
  Throws    : Dies, if given invalid arguments
  Comments  :
  See Also  :

=cut

sub log {
	my $invocant = shift;
	my ($source, $severity, $message, $action, $backtrace) = @_;
	confess "Invalid arguments" unless (defined($source) and defined($severity) and defined($message));

	# Choose the output channels based on the invocant (class or object)
	my $channel_queue;
	my $channels_current;
	if (ref $invocant) {
		$channel_queue = $invocant->{CH_QUEUE};
		$channels_current = $invocant->{CHANNELS};
	}
	else {
		$channel_queue = $CH_QUEUE;
		$channels_current = $CHANNELS;
	}

	# Translate severity to string for printing purposes
	$severity = Log::Core::Essentials->severity_as_str($severity);

	# We must validate, that at least severity was correct
	confess "Invalid severity value" unless (defined($severity));

	# Add backtrace to the message, if requested
	$message .= Carp::longmess() if $backtrace;

	# Walk through all output channels
	foreach my $chnl (@$channel_queue) {
		$channels_current->{$chnl}->write($source, $severity, $message);
	}
	if ($action) {
		if    ($action eq 'exit')    { exit 1; }
		elsif ($action eq 'die')     { die "$source $severity $message"; }
		elsif ($action eq 'warn')    { warn "$source $severity $message"; }
		elsif ($action eq 'carp')    { carp "$source $severity $message"; }
		elsif ($action eq 'croak')   { croak "$source $severity $message"; }
		elsif ($action eq 'cluck')   { cluck "$source $severity $message"; }
		elsif ($action eq 'confess') { confess "$source $severity $message"; }
	}
}

=item log_warn [PUBLIC, HYBRID]

=item log_die [PUBLIC, HYBRID]

=item log_exit [PUBLIC, HYBRID]

=item log_carp [PUBLIC, HYBRID]

=item log_croak [PUBLIC, HYBRID]

=item log_cluck [PUBLIC, HYBRID]

=item log_confess [PUBLIC, HYBRID]

  Usage     : Log::Loger->log_die('source', 'ERROR', 'message');
			  $instance->log_warn('source', 'ERROR', 'message');
  Purpose   : Shortcut to the log() method with specific action after writing the message
  Returns   : See log() method for details
  Arguments : See log() method for details
  Throws    : See log() method for details
  Comments  : See log() method for details
  See Also  : See log() method for details

=cut

sub log_warn {
	my $invocant = shift;
	my ($source, $severity, $message, $backtrace) = @_;
	$invocant->log($source, $severity, $message, 'warn', $backtrace);
}

sub log_die {
	my $invocant = shift;
	my ($source, $severity, $message, $backtrace) = @_;
	$invocant->log($source, $severity, $message, 'die', $backtrace);
}

sub log_exit {
	my $invocant = shift;
	my ($source, $severity, $message, $backtrace) = @_;
	$invocant->log($source, $severity, $message, 'exit', $backtrace);
}

sub log_carp {
	my $invocant = shift;
	my ($source, $severity, $message, $backtrace) = @_;
	$invocant->log($source, $severity, $message, 'carp', $backtrace);
}

sub log_croak {
	my $invocant = shift;
	my ($source, $severity, $message, $backtrace) = @_;
	$invocant->log($source, $severity, $message, 'croak', $backtrace);
}

sub log_cluck {
	my $invocant = shift;
	my ($source, $severity, $message, $backtrace) = @_;
	$invocant->log($source, $severity, $message, 'cluck', $backtrace);
}

sub log_confess {
	my $invocant = shift;
	my ($source, $severity, $message, $backtrace) = @_;
	$invocant->log($source, $severity, $message, 'confess', $backtrace);
}

=item emerg [PUBLIC,HYBRID]

=item alert [PUBLIC,HYBRID]

=item crit [PUBLIC,HYBRID]

=item error [PUBLIC,HYBRID]

=item warn [PUBLIC,HYBRID]

=item notice [PUBLIC,HYBRID]

=item info [PUBLIC,HYBRID]

=item debug [PUBLIC,HYBRID]

  Usage     : Log::Loger->emerg('source', 'message');
			  $instance->alert('source', 'message');
  Purpose   : Shortcut to the log() method with specific severity of the message
  Returns   : See log() method for details
  Arguments : See log() method for details
  Throws    : See log() method for details
  Comments  : See log() method for details
  See Also  : See log() method for details

=cut

# Shortcut functions for specified severities
sub emerg {
	my $invocant = shift;
	my ($source, $message, $backtrace) = @_;
	return $invocant->log($source, 'EMERG', $message, undef, $backtrace);
}
sub alert {
	my $invocant = shift;
	my ($source, $message, $backtrace) = @_;
	return $invocant->log($source, 'ALERT', $message, undef, $backtrace);
}
sub crit {
	my $invocant = shift;
	my ($source, $message, $backtrace) = @_;
	return $invocant->log($source, 'CRIT', $message, undef, $backtrace);
}
sub error {
	my $invocant = shift;
	my ($source, $message, $backtrace) = @_;
	return $invocant->log($source, 'ERROR', $message, undef, $backtrace);
}
sub warn {
	my $invocant = shift;
	my ($source, $message, $backtrace) = @_;
	return $invocant->log($source, 'WARNING', $message, undef, $backtrace);
}
sub notice {
	my $invocant = shift;
	my ($source, $message, $backtrace) = @_;
	return $invocant->log($source, 'NOTICE', $message, undef, $backtrace);
}
sub info {
	my $invocant = shift;
	my ($source, $message, $backtrace) = @_;
	return $invocant->log($source, 'INFO', $message, undef, $backtrace);
}
sub debug {
	my $invocant = shift;
	my ($source, $message, $backtrace) = @_;
	return $invocant->log($source, 'DEBUG', $message, undef, $backtrace);
}

=item email [PUBLIC,HYBRID]

  Usage     : Log::Loger->email('source', 'message');
			  $instance->email('source', 'message');
  Purpose   : Send emails via all configured email writers
  Returns   : Nothing
  Arguments : See log() method for details
  Throws    : Nothing

=cut

sub email {
	my $invocant = shift;
	my ($source, $message, $subject) = @_;

	my $writer_queue;
	if (ref $invocant) { $writer_queue = $invocant->{WRITER_TYPES}->{email}; }
	else               { $writer_queue = $WRITER_TYPES->{email}; }

	# Walk through all email writers
	if ($writer_queue) {
		foreach my $writer (@$writer_queue) {
			$writer->write($source, 'EMAIL', $message, $subject);
		}
	}
}

=pod

=back

=cut

1;
