package Log::Writer::Email;
use strict;
use warnings;

#*******************************************************************************
#
# DOCUMENTATION SECTION
#
#*******************************************************************************

=head1 NAME

Log::Writer::Email - Log writer capable of sending messages by email

=head1 SYNOPSIS

  use Log::Writer::Email;

=head1 DESCRIPTION

This writer writes given messages to the given IO::Handle.

=head1 USAGE

=head1 BUGS

=head1 FLAGS

 rep r REPLACE_SOURCE
 ser s SERIALIZE
 tra t TRACEBACK

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
#
# INITIALIZATION AND CLEANUP SECTION
#
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use File::Basename;
use Sys::Hostname;
use Mail::Sendmail;
use Storable 'dclone';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Writer::Module;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.01';
    $DEVEL   = 0;
    @ISA     = ('Log::Writer::Module');
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant FLAG_REPLACE   => 'REPLACE';
use constant FLAG_TRACEBACK => 'TRACEBACK';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
#
# IMPLEMENTATION SECTION
#
#*******************************************************************************

=item write($$$;$) [PUBLIC]

  Usage     : $writer->write($source, $severity, $message, $subject);
  Purpose   : Write given message to the destination
  Arguments : STRING $source - Name of the source of the message
              ENUM $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              STRING $message - Log message
  Returns   : $self
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : Log::Core::Essentials module for permited severity values

=cut

sub write($$$;$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($source, $severity, $message, $subject) = @_;
	confess"Invalid arguments" unless (defined($source) and defined($severity) and defined($message));

	my $email = dclone($self->{EMAIL_PARAMS});

	$email->{'Subject'} = "[".uc($severity)."] ".$email->{'Subject'};
	$email->{'Subject'} = $subject if $subject;
	$email->{'Message'} = $self->_format_message($source, $severity, $message);

	my $success = sendmail(%$email);
	die ("Unable to mail log: " . $Mail::Sendmail::error) unless ($success);

	return $self;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Log::Writer::String instance
#  Arguments : SCALAR $buffer - string buffer variable
#			   (for other _init arguments see Log::Writer::Handle class)
#  Returns   : Log::Writer::Module reference
#  Throws    : Croaks, if invoked on class, or if given invalid arguments
#  See also  : Log::Writer::Handle module

sub _init
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($email_params, $flags) = @_;
	confess "Invalid arguments" unless (defined($email_params));

	$self->_setup_email_params($email_params);
	$self->_setup_flags($flags);

	return $self;
}

# _get_possible_flags() [PROTECTED]
#
#  Usage     : my $flags = $self->_get_possible_flags();
#  Purpose   : Get hash structure describing all possible flags
#  Arguments : None
#  Returns   : HASH REFERENCE $flags
#  Throws    : Croaks, if invoked on class

sub _get_possible_flags()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	return {
		'r'         => FLAG_REPLACE(),
		'rep'       => FLAG_REPLACE(),
		'replace'   => FLAG_REPLACE(),
		't'         => FLAG_TRACEBACK(),
		'tra'       => FLAG_TRACEBACK(),
		'traceback' => FLAG_TRACEBACK(),
	};
}

# _setup_email_params($) [PROTECTED]

#  Usage     : $self->_setup_email_params($email);
#  Purpose   :
#  Arguments :
#  Returns   : $self
#  Throws    : Croaks, if invoked on class, or if given invalid arguments

sub _setup_email_params($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($email_params,) = @_;
	confess "Email parameters must be given as argument" unless $email_params;
	confess "Email parameters must be given as HASH reference" unless ref $email_params eq 'HASH';
	confess "Invalid email parameters: missing source email address" unless $email_params->{'From'};
	confess "Invalid email parameters: missing destination email address" unless $email_params->{'To'};
	confess "Invalid email parameters: missing email subject" unless $email_params->{'Subject'};

	$self->{EMAIL_PARAMS} = $email_params;
	$self->{EMAIL_PARAMS}->{'Content-type'} = 'text/plain; charset="utf-8"' unless $self->{EMAIL_PARAMS}->{'Content-type'};
	$self->{EMAIL_PARAMS}->{'Reply-To'} = $self->{EMAIL_PARAMS}->{'From'} unless $self->{EMAIL_PARAMS}->{'Reply-To'};
	return $self;
}

# _format_message($$$) [PROTECTED]
#
#  Usage     : my $message = $self->_format_message($source, $severity, $message)
#  Purpose   : Format log message from the given source, severity and message body
#  Arguments : STRING $source - Name of the source of the message
#              ENUM $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
#              STRING $message - Message
#  Returns   : STRING formatted message
#  Throws    : Croaks, if invoked on class

sub _format_message($$$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($source, $severity, $message) = @_;

	my $date = localtime;
	my $hostname = hostname;
	$source = basename($0) if ($self->_flag(FLAG_REPLACE()));
	#$date = strftime($self->{DATE_FORMAT}, localtime) if ($self->{DATE_FORMAT});

	my $result = "$date $hostname ${source}[$$] $severity: $message\n";
	$result .= "\nBacktrace:\n * " . join("\n * ", $self->_traceback()) . "\n" if $self->_flag(FLAG_TRACEBACK());
	$result .= "\nNOTE: This email was automatically generated, please do not reply to it!";
	return $result;
}

=pod

=back

=cut

1;
