package Log::Writer::File;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Log::Writer::File - Ordinary text file log writer

=head1 SYNOPSIS

 use Log::Writer::File;

 my $writer = Log::Writer::File->new('./spool/writer-file.log');
 my $writer = Log::Writer::File->new('./spool/writer-file.log',...);
 # For more information about the constructor arguments and some more examples
 # please refer to the Log::Writer::Handle(3) manual page

 $instance->write('module', 'severity', 'Message from Log::Writer::File');

=head1 DESCRIPTION

This writer writes given messages to the simple text log file. Implementation
is based on the Log::Writer::Handle class, see. it for more details.

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1), Log::Writer::Handle(3), Log::Writer::Module(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use IO::File;
use File::stat;

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Writer::Handle;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.01';
    $DEVEL   = 0;
    @ISA     = ('Log::Writer::Handle');
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant DFLT_STEP => 3;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item DESTROY [PUBLIC]

  Usage     : Instance destructor
  Purpose   : Close opened log file handle
  Returns   : Nothing
  Arguments : None
  Throws    : Nothing

=cut

sub DESTROY {
	my $self = shift;
	close($self->{IO_HANDLE}) if $self->{IO_HANDLE};
}

=item write($$$) [PUBLIC]

  Usage     : $writer->write($source, $severity, $message);
  Purpose   : Write given message to the destination
  Arguments : STRING $source - Name of the source of the message
              ENUM $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              STRING $message - Log message
  Returns   : $self
  Throws    : Dies, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : Log::Core::Essentials module for permited severity values

=cut

sub write($$$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	# Check for log file rotation and reopen the log file, if needed
	if ($self->_check_for_log_rotation()) {
		print STDERR "Log file rotation detected\n" if $DEVEL;
		$self->_setup_handle($self->_open_log());
	}

	return $self->SUPER::write(@_);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Log::Writer::File instance
#  Arguments : STRING $log_file_name - name of the lof file
#			   (for other _init arguments see Log::Writer::Handle class)
#  Returns   : Log::Writer::Module reference
#  Throws    : Croaks, if invoked on class, or if given invalid arguments
#  See also  : Log::Writer::Handle module

sub _init
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($log_file_name, @arguments) = @_;
	confess "Expecting log file name to be given as argument" unless ($log_file_name);

	$self->{LOG_FILE_NAME} = $log_file_name;
	$self->{LAST_CHECK}    = time;
	$self->{CHECK_STEP}    = DFLT_STEP();

	my $io_handle = $self->_open_log();
	return $self->SUPER::_init($io_handle, @arguments);
}

# _open_log() [PROTECTED]
#
#  Usage     : $self->_open_log();
#  Purpose   : Lock the lock file for serialization before writing
#  Arguments : None
#  Returns   : $self
#  Throws    : Dies, if invoked on class

sub _open_log() {
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	# Open the log file
	my $io_handle = IO::File->new($self->{LOG_FILE_NAME}, O_WRONLY|O_APPEND|O_CREAT) or die "Can't open log file '$self->{LOG_FILE_NAME}': $!\n";
	$io_handle->autoflush(1);

	# Mark the unique file ID
	my $stat = stat($self->{LOG_FILE_NAME});
	$self->{LOG_FILE_ID} = $stat->dev.':'.$stat->ino;

	return $io_handle;
}

# _check_for_log_rotation() [PROTECTED]
#
#  Usage     : $self->_open_log();
#  Purpose   : Lock the lock file for serialization before writing
#  Arguments : None
#  Returns   : $self
#  Throws    : Dies, if invoked on class

sub _check_for_log_rotation() {
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	# Check for logrotation with certain backoff
	my $t = time;
	return 0 if (($self->{LAST_CHECK} + $self->{CHECK_STEP}) > $t);
	$self->{LAST_CHECK} = $t;

	# File might have been renamed/moved
	my $stat = stat($self->{LOG_FILE_NAME});
	return 1 unless $stat;

	# Compare unique file identifiers "dev:ino" ()
	my $id = $stat->dev.':'.$stat->ino;
	print STDERR "Checking for logrotation $self->{LAST_CHECK} : $t : $id ne $self->{LOG_FILE_ID}\n" if $DEVEL;
	return ($id ne $self->{LOG_FILE_ID})?1:0;
}

=pod

=back

=cut

1;
