package Log::Writer::Handle;
use strict;
use warnings;

#*******************************************************************************
#
# DOCUMENTATION SECTION
#
#*******************************************************************************

=head1 NAME

Log::Writer::Handle - Log writer capable of writing to any filehandle

=head1 SYNOPSIS

 use Log::Writer::Handle;

 # Prepare some IO capable objects (handles)
 my $buffer = '';
 my $handle = IO::Handle->new_from_fd(fileno(STDERR),"w") or die "Can't open 'STDERR': $!\n";
 my $handle = IO::File->new('./spool/writer-handle.log','>') or die "Can't open file './spool/test.log': $!\n";
 my $handle = IO::String->new($buffer) or die "Can't open string buffer: $!\n";

 # Various methods of writer instantination
 my $writer = Log::Writer::Handle->new($handle);
 my $writer = Log::Writer::Handle->new($handle, 'iso8601');
 my $writer = Log::Writer::Handle->new($handle, '%FT%T%z');
 my $writer = Log::Writer::Handle->new($handle, undef, 'r');
 my $writer = Log::Writer::Handle->new($handle, undef, 'rep');
 my $writer = Log::Writer::Handle->new($handle, undef, 'replace');
 my $writer = Log::Writer::Handle->new($handle, undef, 's', './spool/writer-handle.lock');
 my $writer = Log::Writer::Handle->new($handle, undef, 'ser', './spool/writer-handle.lock');
 my $writer = Log::Writer::Handle->new($handle, undef, 'serialize', './spool/writer-handle.lock');
 my $writer = Log::Writer::Handle->new($handle, undef, 'r,s', './spool/writer-handle.lock');
 my $writer = Log::Writer::Handle->new($handle, undef, 'rep,ser', './spool/writer-handle.lock');
 my $writer = Log::Writer::Handle->new($handle, undef, 'replace,serialize', './spool/writer-handle.lock');

 # Write the message into the writer
 $writer->write('module', 'severity', 'Message from Log::Writer::Handle');

 # Process safe writing - if the fork occurs after log writer initialization, lock must be reopened
 my $writer = new Log::Writer::Handle($hnd, undef, 'rep ser', './spool/test.lock');
 die "fork: $!" unless defined (my $pid = fork);
 $writer->handle_fork();
 my $identity = '[CHILD] ';
 if ($pid) { $identity = '[PARENT]'; }
 $writer->write('module', 'severity', '$identity: Message from Log::Writer::Handle');
 if ($pid) { wait; }

=head1 DESCRIPTION

This writer writes given messages to the given IO::Handle.

=head1 FLAGS

=over

=item B<c|com|comment>

 COMMENT out every line with '#' character

=item B<r|rep|replace>

 REPLACE source name given to the write() method with program name

=item B<s|ser|serialize>

 SERIALIZE the write operations using mutual exclusion lock

=item B<t|tra|traceback>

 TRACEBACK

=back

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1), Log::Core::Essentials(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
#
# INITIALIZATION AND CLEANUP SECTION
#
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use POSIX qw(strftime);
use File::Basename;
use Sys::Hostname;

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Writer::Module;
use Mutex::Flock;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.01';
    $DEVEL   = 0;
    @ISA     = ('Log::Writer::Module');
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant FLAG_COMMENT   => 'COMMENT';
use constant FLAG_REPLACE   => 'REPLACE';
use constant FLAG_SERIALIZE => 'SERIALIZE';
use constant FLAG_TRACEBACK => 'TRACEBACK';

use constant ISO8601 => 'ISO8601';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
#
# IMPLEMENTATION SECTION
#
#*******************************************************************************

=item write($$$) [PUBLIC]

  Usage     : $writer->write($source, $severity, $message);
  Purpose   : Write given message to the destination
  Arguments : STRING $source - Name of the source of the message
              ENUM $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              STRING $message - Log message
  Returns   : $self
  Throws    : Dies, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : Log::Core::Essentials module for permited severity values

=cut

sub write($$$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($source, $severity, $message) = @_;
	croak ((caller(0))[3] . ": received invalid arguments") unless (defined($source) and defined($severity) and defined($message));

	$self->_lock();
	print { $self->{IO_HANDLE} } $self->_format_message($source, $severity, $message) or die ("Error writing message: $!");
	$self->_unlock();

	return $self;
}

=item handle_fork() [PUBLIC]

  Usage     : $writer->handle_fork();
  Purpose   : Allow writer to react to the fork() (sometimes it is necessary to reopen
              file descriptors etc.)
  Arguments : None
  Returns   : $self
  Throws    : Dies, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes

=cut

sub handle_fork()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	$self->{LOCK}->reopen();
	return $self;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Log::Writer::Handle instance
#  Arguments : IO::Handle $handle - handle to write to
#			   STRING $date_format - date format (see. man strftime(3) for details on syntax) [MANDATORY]
#			   STRING $flags - various processing flags [OPTIONAL]
#					s | ser | serialize - serialize the writes (for multiprocess application)
#					r | rep | replace - replace source of the message with the program name
#			   STRING $lock_file - name of the lock file [MANDATORY if 'serialize' flag is set]
#  Returns   : Log::Writer::Module reference
#  Throws    : Dies, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses

sub _init
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($handle, $date_format, $flags, $lock_file) = @_;
	confess "IO handle must be given as first argument" unless (defined($handle));

	$self->_setup_handle($handle);
	$self->_setup_date_format($date_format);
	$self->_setup_flags($flags);

	if ($self->_flag(FLAG_SERIALIZE())) {
		croak ((caller(0))[3] . ": missing arguments - lock file") unless (defined($lock_file));
		$self->{LOCK} = Mutex::Flock->new($lock_file);
	}
	return $self;
}

# _get_possible_flags() [PROTECTED]
#
#  Usage     : my $flags = $self->_get_possible_flags();
#  Purpose   : Get hash structure describing all possible flags
#  Arguments : None
#  Returns   : HASH REFERENCE $flags
#  Throws    : Dies, if invoked on class

sub _get_possible_flags()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	return {
		'c'         => FLAG_COMMENT(),
		'com'       => FLAG_COMMENT(),
		'comment'   => FLAG_COMMENT(),
		'r'         => FLAG_REPLACE(),
		'rep'       => FLAG_REPLACE(),
		'replace'   => FLAG_REPLACE(),
		's'         => FLAG_SERIALIZE(),
		'ser'       => FLAG_SERIALIZE(),
		'serialize' => FLAG_SERIALIZE(),
		't'         => FLAG_TRACEBACK(),
		'tra'       => FLAG_TRACEBACK(),
		'traceback' => FLAG_TRACEBACK(),
	};
}

# _setup_handle($) [PROTECTED]
#
#  Usage     : $self->_setup_handle($handle);
#  Purpose   : Set IO::Handle to write messages to
#  Arguments : IO::Handle $handle - handle to write to
#  Returns   : $self
#  Throws    : Dies, if invoked on class, or if given invalid arguments

sub _setup_handle($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($handle,) = @_;
	confess "Given handle is not capable of 'print'" unless ($handle->can("print"));

	$self->{IO_HANDLE} = $handle;
	return $self;
}

# _setup_date_format($) [PROTECTED]
#
#  Usage     : $self->_setup_date_format($date_format);
#  Purpose   : Set format according to which to display the date
#  Arguments : STRING $date_format - date format (see. strftime() for details on syntax)
#  Returns   : $self
#  Throws    : Dies, if invoked on class

sub _setup_date_format($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($date_format,) = @_;

	$self->{DATE_FORMAT} = $date_format;
	$self->{DATE_FORMAT} = '%FT%T%z' if $date_format and uc($date_format) eq ISO8601;
	return $self;
}

# _format_message($$$) [PROTECTED]
#
#  Usage     : my $message = $self->_format_message($source, $severity, $message)
#  Purpose   : Format log message from the given source, severity and message body
#  Arguments : STRING $source - Name of the source of the message
#              ENUM $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
#              STRING $message - Message
#  Returns   : STRING formatted message
#  Throws    : Dies, if invoked on class

sub _format_message($$$)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($source, $severity, $message) = @_;

	my $date = localtime;
	my $hostname = hostname;
	my $cmnt = ($self->_flag(FLAG_COMMENT()))?'# ':'';
	$source = basename($0) if ($self->_flag(FLAG_REPLACE()));
	$date = strftime($self->{DATE_FORMAT}, localtime) if ($self->{DATE_FORMAT});
	$message =~ s/\n(.+)/\n#  $1/g if ($self->_flag(FLAG_COMMENT()));

	return "$cmnt$date $hostname ${source}[$$]: $severity: $message\n" unless $self->_flag(FLAG_TRACEBACK());
	return "$cmnt$date $hostname ${source}[$$]: $severity: $message\n$cmnt Backtrace:\n$cmnt   * " . join("\n$cmnt   * ", $self->_traceback()) . "\n";
	#return "$cmnt$date $hostname ${source}[$$]: $severity: $message\n$cmnt Backtrace:\n$cmnt   * " . join("\n$cmnt   * ", $self->_backtrace()) . "\n";
}

# _lock() [PROTECTED]
#
#  Usage     : $self->_lock();
#  Purpose   : Lock the lock file for serialization before writing
#  Arguments : None
#  Returns   : $self
#  Throws    : Dies, if invoked on class

sub _lock() {
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	$self->{LOCK}->lock() if $self->_flag(FLAG_SERIALIZE());
	return $self;
}

# _unlock() [PROTECTED]
#
#  Usage     : $self->_unlock();
#  Purpose   : Unlock the lock file for serialization after writing
#  Arguments : None
#  Returns   : $self
#  Throws    : Dies, if invoked on class

sub _unlock() {
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	$self->{LOCK}->unlock() if $self->_flag(FLAG_SERIALIZE());
	return $self;
}

=pod

=back

=cut

1;
