package Log::Writer::Module;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Log::Writer::Module - Base module defining mandatory interface for all log writing modules

=head1 SYNOPSIS

 use Log::Writer::Module;

 BEGIN {
    @ISA = ('Log::Writer::Module');
 }

=head1 DESCRIPTION

Classes extending this base class may be used as WRITERS in loging framework. Writers
are used in log message channels for writing messages to various destinations.

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.01';
    $DEVEL   = 0;
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $filter = Log::Writer::(module)->new(... arguments);
  Purpose   : Create new instance of log writer
  Arguments : All arguments are passed to the _init() method
  Returns   : Reference to the new instance
  Throws    : Croaks, if not invoked on class
  Comments  : This method should not be overloaded, descendants should implement their version of _init() method instead
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	my $self = bless ({}, $class);
	return $self->_init(@_);
}

=item write($$$) [ABSTRACT,PUBLIC]

  Usage     : $writer->write($source, $severity, $message);
  Purpose   : Write given message to the destination
  Arguments : STRING $source - Name of the source of the message
              ENUM $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              STRING $message - Log message
  Returns   : $self
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : Log::Core::Essentials module for permited severity values

=cut

sub write($$$)
{
	my $self = shift;
	confess "Method needs implementation";
	# my ($source, $severity, $message) = @_;
}

=item handle_fork() [PUBLIC]

  Usage     : $writer->handle_fork();
  Purpose   : Allow writer to react to the fork() (sometimes it is necessary to reopen
              file descriptors etc.)
  Arguments : None
  Returns   : $self
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes

=cut

sub handle_fork()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	return $self;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [ABSTRACT,PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Log::Writer::Module instance
#  Arguments : Unknown
#  Returns   : Log::Writer::Module reference
#  Throws    : Croaks, if not invoked on instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
	my $self = shift;
	confess "Method needs implementation";
}

#-------------------------------------------------------------------------------

# _get_possible_flags() [PROTECTED]
#
#  Usage     : my $flags = $self->_get_possible_flags();
#  Purpose   : Get hash structure describing all possible flags
#  Arguments : None
#  Returns   : HASH REFERENCE $flags
#  Throws    : Croaks, if not invoked on instance

sub _get_possible_flags()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	return {};
}

# _setup_flags($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if not invoked on instance

sub _setup_flags($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($flags,) = @_;

	if ($flags)	{
		my $possible_flags = $self->_get_possible_flags();
		foreach my $flag (split(/[,;]/, $flags)) {
			if ($possible_flags->{$flag}) {
				$self->{FLAGS}->{uc($possible_flags->{$flag})}++;
			}
			else {
				confess "Unknown flag '$flag'";
			}
		}
	}
	return $self;
}

# _flag($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if not invoked on instance

sub _flag($)
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	my ($flag,) = @_;

	return $self->{FLAGS}->{uc($flag)};
}

# _traceback() [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if not invoked on instance

sub _traceback()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	my @traceback = ();
	my (@result, $tbline);
	my ($package, $filename, $line, $subroutine, $hasargs, $wantarray, $evaltext, $is_require, $hints, $bitmask, $hinthash);
	my $i = 0;
	while (@result = caller($i++))
	{
		($package, $filename, $line, $subroutine, $hasargs, $wantarray, $evaltext, $is_require, $hints, $bitmask, $hinthash) = @result;
		next if $package =~ /^Log::/;
		$tbline = "Subroutine $subroutine called at [$package, $filename, L#$line]";
		push(@traceback, $tbline);
	}
	return @traceback;
}

# _backtrace() [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if not invoked on instance

sub _backtrace()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);

	my $backtrace = Carp::longmess();
	my @lines = split(/\n/, $backtrace);
	my @result = ();
	foreach my $line (@lines) {
		#next if $line =~ /^Log::/;
		push(@result, $line);
	}
	return @result;
}

=pod

=back

=cut

1;
