package Log::Writer::Syslog;
use strict;
use warnings;

#*******************************************************************************
#
# DOCUMENTATION SECTION
#
#*******************************************************************************

=head1 NAME

Log::Writer::Syslog - Log writer capable of logging messages to syslog

=head1 SYNOPSIS

  use Log::Writer::Syslog;


=head1 DESCRIPTION

This writer writes given messages to the given IO::Handle.

=head1 USAGE

=head1 BUGS

=head1 FLAGS

 rep r REPLACE_SOURCE
 ser s SERIALIZE
 tra t TRACEBACK

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
#
# INITIALIZATION AND CLEANUP SECTION
#
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use File::Basename;
use Sys::Syslog qw(:standard :macros setlogsock);

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Core::Essentials;
use Log::Writer::Module;

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA %SEV2PRIO);
    $VERSION = '0.01';
    $DEVEL   = 0;
    @ISA     = ('Log::Writer::Module');

	%SEV2PRIO = (
		Log::Core::Essentials::EMAIL_STR()   => LOG_EMERG,		# special internal severity
		Log::Core::Essentials::EMERG_STR()   => LOG_EMERG,		# system is unusable
		Log::Core::Essentials::ALERT_STR()   => LOG_ALERT,		# action must be taken immediately
		Log::Core::Essentials::CRIT_STR()    => LOG_CRIT,		# critical conditions
		Log::Core::Essentials::ERROR_STR()   => LOG_ERR,		# error conditions
		Log::Core::Essentials::WARNING_STR() => LOG_WARNING,	# warning conditions
		Log::Core::Essentials::NOTICE_STR()  => LOG_NOTICE,		# normal, but significant, condition
		Log::Core::Essentials::INFO_STR()    => LOG_INFO,		# informational message
		Log::Core::Essentials::DEBUG_STR()   => LOG_DEBUG,		# debug-level message
	);
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant FLAG_REPLACE   => 'REPLACE';
use constant FLAG_TRACEBACK => 'TRACEBACK';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
#
# IMPLEMENTATION SECTION
#
#*******************************************************************************

=item write($$$) [PUBLIC]

  Usage     : $writer->write($source, $severity, $message);
  Purpose   : Write given message to the destination
  Arguments : STRING $source - Name of the source of the message
              ENUM $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
              STRING $message - Log message
  Returns   : $self
  Throws    : Croaks, if not implemented in descendant classes
  Comments  : ABSTRACT method, must be implemented in descendant classes
  See Also  : Log::Core::Essentials module for permited severity values

=cut

sub write($$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($source, $severity, $message) = @_;
	croak ((caller(0))[3] . ": invalid arguments") unless (defined($source) and defined($severity) and defined($message));
	croak ((caller(0))[3] . ": invalid severity value '$severity'") unless exists $SEV2PRIO{uc($severity)};

	syslog($SEV2PRIO{uc($severity)}, $self->_format_message($source, $severity, $message));

	return $self;
}

=item DESTROY [PUBLIC]

  Usage     : Instance destructor
  Purpose   : Close opened log file handle
  Returns   : Nothing
  Arguments : None
  Throws    : Nothing

=cut

sub DESTROY {
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

    closelog();
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Log::Writer::String instance
#  Arguments : SCALAR $buffer - string buffer variable
#			   (for other _init arguments see Log::Writer::Handle class)
#  Returns   : Log::Writer::Module reference
#  Throws    : Croaks, if invoked on class, or if given invalid arguments
#  See also  : Log::Writer::Handle module

sub _init
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($facility, $flags) = @_;
	$facility = LOG_LOCAL7 unless $facility;

	setlogsock('unix');
	openlog(basename($0),'pid',$facility);

	$self->_setup_flags($flags);

	return $self;
}

# _get_possible_flags() [PROTECTED]
#
#  Usage     : my $flags = $self->_get_possible_flags();
#  Purpose   : Get hash structure describing all possible flags
#  Arguments : None
#  Returns   : HASH REFERENCE $flags
#  Throws    : Croaks, if invoked on class

sub _get_possible_flags()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);

	return {
		'r'         => FLAG_REPLACE(),
		'rep'       => FLAG_REPLACE(),
		'replace'   => FLAG_REPLACE(),
		't'         => FLAG_TRACEBACK(),
		'tra'       => FLAG_TRACEBACK(),
		'traceback' => FLAG_TRACEBACK(),
	};
}

# _format_message($$$) [PROTECTED]
#
#  Usage     : my $message = $self->_format_message($source, $severity, $message)
#  Purpose   : Format log message from the given source, severity and message body
#  Arguments : STRING $source - Name of the source of the message
#              ENUM $severity - Severity in integer or string format (see Log::Core::Essentials for permited values)
#              STRING $message - Message
#  Returns   : STRING formatted message
#  Throws    : Croaks, if invoked on class

sub _format_message($$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless blessed($self);
	my ($source, $severity, $message) = @_;

	$source = basename($0) if ($self->_flag(FLAG_REPLACE()));

	return "$source: $message\n" unless $self->_flag(FLAG_TRACEBACK());
	return "$source: $message\n Backtrace:\n  * " . join("\n  * ", $self->_traceback()) . "\n";
}

=pod

=back

=cut

1;
