package Module::Dynaloader;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Module::Dynaloader - Dynamical loading of application modules

=head1 SYNOPSIS

  use Module::Dynaloader;

  sub callback {
	my $module = shift;

	diag("Callback called for module: '$module'\n");

	return 0;
  }

  Module::Dynaloader->find_modules('Test::Modules');

  Module::Dynaloader->find_and_load_modules('Test::Modules','Toskip',\&callback);

=head1 DESCRIPTION

This module contains static methods, that can be used for dynamical loading
of application modules.

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Cwd qw(abs_path);
use Data::Dumper;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION);
    $VERSION 	= '0.01';
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item find_modules($;$) [PUBLIC,STATIC]

  Usage     : Module::Dynaloader->find_modules('Test::Modules');
  Purpose   : Find all submodules of the given module
  Arguments : string $master_module - master module name
			  string $skip_pattern - skip all modules that comply to this pattern (PCRE)
  Returns   : array reference - names of the submodules found
  Throws    : Croaks, if invoked on object or invalid arguments given
			  Dies if the master_module is not a directory
  Comments  :
  See Also  :

=cut

sub find_modules($;$)
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my ($master_module, $skip_pattern) = @_;

	my @modules = ();

	# At least master module directory must be specified
	unless(defined($master_module))
	{
		croak "You have to specify name of the master module directory.\n";
	}

	# Create subpath by replacing all double double dots (::) with slash (/)
	my $master_subpath = $master_module;
	$master_subpath =~ s/::/\//g;

	# Walk through all directories in include path
	foreach my $inc (@INC)
	{
		# Strip trailing slashes from the inc path, if present
		$inc =~ s/(\/)+$//;

		# Create module path
		my $master_path = $inc . '/' . $master_subpath;

		# If the directory does not exist, continue
		unless (-d $master_path)
		{
			next;
		}

		# Find all files in a given directory path
		opendir(MD, $master_path) || die("Cannot open directory '$master_path'");
		my @files = readdir(MD);
		closedir(MD);

		# Walk through all files
		foreach my $f (@files)
		{
			my $filename = $master_path . '/' . $f;

			# Module must be ordinary file
			unless (-f $filename)
						{ next; }
			# Filter out weird named files
			unless ($f =~ /^([a-zA-Z0-9_]+).pm$/)	{ next; }

			# Save the module name
			my $module = $1;

			# Maybe we should skip this module from listing
			if (defined($skip_pattern) and $module =~ /$skip_pattern/)
			{
				next;
			}

			# Everything OK, push the module to the stack
			push(@modules,$master_module . "::" . $module);
		}
		return \@modules;
	}
	return \@modules;
}

=item load_modules($;$) [PUBLIC,STATIC]

  Usage     : Module::Dynaloader->load_modules(['Test::Modules::Module1','Test::Modules::Module2'],\&callback);
  Purpose   : Attempts to load all given modules and call optional callback with
			  the module name as argument afterwards. Callback should use following
			  return codes:
					0 = SUCCESS, CONTINUE
					1 = FAILURE, BUT CONTINUE
					2 = FAILURE, DIE
  Arguments : array reference $modules - array containing module names to load
			  subroutine reference $callback - callback to call after module is loaded
  Returns   : array (integer total modules, integer succefully loaded modules)
  Throws    : Croaks, if invoked on object
			  Dies if the callback return code is 2
  Comments  :
  See Also  :

=cut

sub load_modules($;$)
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my ($modules, $callback) = @_;

	my $total   = 0;
	my $success = 0;
	my $rc;

	foreach my $module (@$modules)
	{
		# Increment the total counter
		$total++;

		# Dynamically include the module definition
		eval "require $module;";
		die $@ if( $@ );

		# And call optional callback
		if (defined($callback))
		{
			$rc = &$callback($module);

			# Return code 1 means FAILURE, BUT CONTINUE
			if 		($rc == 1) { next; }
			# Return code 1 means FAILURE, DIE
			elsif	($rc == 2) { die "Fatal error in $module callback"; }
		}
		$success++;
	}
	return ($total, $success);
}

=item find_and_load_modules($;$$) [PUBLIC,STATIC]

  Usage     : Module::Dynaloader->find_and_load_modules('Test::Modules','Toskip',\&callback)
  Purpose   : Attempts to find and load submodules of given master module
  Arguments : string $master_module - see find_modules() method
			  string $skip_pattern - see find_modules() method
			  subroutine reference $callback - see load_modules() method
  Returns   : see load_modules() method for details on returned values
  Throws    : Croaks, if invoked on object
              Dies if the callback return code is 2
  Comments  :
  See Also  : find_modules() and load_modules() methods

=cut

sub find_and_load_modules($;$$)
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my ($master_module, $skip_pattern, $callback) = @_;

	return $class->load_modules($class->find_modules($master_module, $skip_pattern),$callback);

}

=pod

=back

=cut

1;
