package Mutex::Flock;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Mutex::Flock - OO implementation of the exclusive lock

=head1 SYNOPSIS

  use Mutex::Flock;

  my $lock = new Mutex::Flock('./spool/test.lock');

  $lock->lock();
  $lock->unlock();

  $lock->reopen();

=head1 DESCRIPTION

Object oriented implementation of exclusive locks based on the flock()
primitive.

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use IO::File;
use Fcntl qw(:flock);

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION);
    $VERSION 	= '0.01';
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item new() [PUBLIC, STATIC]

  Usage     : my $lock = Mutex::Flock->new('./spool/test.lock');
  Purpose   : Create new exclusive lock
  Returns   : Create new instance of exclusive lock
  Arguments : string $lock_file_name - name of the lock file
  Throws    : Croaks, if invoked on object or invalid arguments given
  Comments  : Internally uses reopen() method
  See Also  : reopen() method

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my ($lock_file_name,) = @_;
	croak ((caller(0))[3] . ": invalid arguments") unless $lock_file_name;

	my $self = bless ({}, $class);
	$self->{LOCK_FILE_NAME} = $lock_file_name;
	$self->reopen();
	return $self;
}

=item DESTROY [PUBLIC]

  Usage     :
  Purpose   : Close opened lock file handle
  Returns   :
  Arguments :
  Throws    :
  Comments  :
  See Also  :

=cut

sub DESTROY {
	my $self = shift;
	close($self->{LOCK_HANDLE}) if $self->{LOCK_HANDLE};
}

=item reopen() [PUBLIC]

  Usage     : $lock->reopen();
  Purpose   : Reopen the lock file
  Returns   : Nothing
  Arguments : None
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub reopen() {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	$self->{LOCK_HANDLE} = IO::File->new($self->{LOCK_FILE_NAME},'w') or die "Can't open lock file '$self->{LOCK_FILE_NAME}': $!\n";
}

=item lock() [PUBLIC]

  Usage     : $lock->lock();
  Purpose   : Lock the exclusive lock
  Returns   : Nothing
  Arguments : None
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub lock() {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	flock($self->{LOCK_HANDLE}, LOCK_EX) or die "Can't lock lock file '$self->{LOCK_FILE_NAME}': $!\n";
}

=item unlock() [PROTECTED]

  Usage     : $lock->unlock();
  Purpose   : Unlock the exclusive lock
  Returns   : Nothing
  Arguments : None
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub unlock() {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	flock($self->{LOCK_HANDLE}, LOCK_UN) or die "Can't unlock lock file '$self->{LOCK_FILE_NAME}': $!\n";
}

=pod

=back

=cut

1;
