package Object::Factory;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Object::Factory - Base class for object factories

=head1 SYNOPSIS

 use Object::Factory;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Storable(); # Do not import anything to local namespace

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Object::Factory->new();
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Object::Factory reference on success, croaks on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	### Creating Object.Factory instance...
	my $self = bless ({}, $class);
	return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	### Destroying Object.Factory instance...
}

#-------------------------------------------------------------------------------

=item name() [PUBLIC,HYBRID]

  Usage     : my $name = Object::Factory->name(); OR
              my $name = $instance->name();
  Purpose   : Get name of the forged objects
  Arguments : None
  Returns   : STRING $name on success

=cut

sub name()
{
	my $invocant = shift;

	if (blessed($invocant)) { $invocant = blessed($invocant); }
	unless ($invocant::ONAME) {
		my $name = (blessed($invocant))?blessed($invocant):$invocant; $name =~ s/::Factory$//;
		my @tmp = split(/:+/, $name);
		$invocant::ONAME = lc($tmp[$#tmp]);
	}
}

=item validate_type($) [PUBLIC,HYBRID]

  Usage     : my $rc = Object::Factory->validate_type($type);
              my $rc = $instance->validate_type($type);
  Purpose   : Validate object type
  Arguments : STRING $type - object type [MANDATORY]
  Returns   : BOOL TRUE on SUCCESS, FALSE on FAILURE

=cut

sub validate_type($)
{
	my $invocant = shift;
	my ($type,) = @_;
	return exists($invocant->_get_type_map->{$type});
}

=item valid_types() [PUBLIC,HYBRID]

  Usage     : my $rc = Object::Factory->valid_types($type);
              my $rc = $instance->valid_types($type);
  Purpose   : Get the list of valid object types
  Arguments : None
  Returns   : ARRAY @list of valid object types

=cut

sub valid_types()
{
	my $invocant = shift;
	return keys(%{$invocant->_get_type_map});
}

#-------------------------------------------------------------------------------

=item forge($;@) [PUBLIC,HYBRID]

  Usage     : my $object = Object::Factory->forge($type, @args);
              my $object = $instance->forge($type, @args);
  Purpose   : Create factory product and return reference to the new instance
  Arguments : STRING $type - type of the object [MANDATORY]
  Returns   : BLESSED REFERENCE on SUCCESS
  Throws    : Dies, if given invalid arguments or in case of failure

=cut

sub forge($;@)
{
	my $invocant = shift;
	my ($type, @args) = @_;
	confess "Type of the '".$invocant->name."' must be given as first argument" unless $type;

	# Determine the class name based on the object type
	my $map = $invocant->_get_type_map();
	my $object_class = $map->{$type};
	die "Invalid '".$invocant->name."' type '$type', valid values are '".join(',', $invocant->valid_types)."'\n" unless $object_class;

	return $object_class->new(@args);
}

=item forge_q($;@) [PUBLIC,HYBRID]

  Usage     : Same as forge() method
  Purpose   : Quietly create factory product and return reference to the new instance
  Arguments : Same as forge() method
  Returns   : BLESSED REFERENCE on SUCCESS, ARRAY (undef, STRING $error) on FAILURE
  Throws    : Nothing, all errors suppressed

=cut

sub forge_q($;@)
{
	my $invocant = shift;
	my $object;
	eval {
		$object = $invocant->forge(@_);
	};
	return (undef, $@) if ($@);
	return $object;
}

=item freeze($) [PUBLIC,HYBRID]
=item thaw($) [PUBLIC,HYBRID]

  Usage     : my $serialization = Object::Factory->freeze($object);
              my $serialization = $instance->freeze($object);
              my $object = Object::Factory->thaw($serialization);
              my $object = $instance->thaw($serialization);
  Purpose   : Serialize/deserialize given object
  Arguments : Same as Storable::freeze() and Storable::thaw() methods
  Returns   : Same as Storable::freeze() and Storable::thaw() methods

=cut

sub freeze(@)
{
	my $invocant = shift;
	return Storable::freeze(@_);
}
sub thaw(@)
{
	my $invocant = shift;
	return Storable::thaw(@_);
}

=item store($$) [PUBLIC,HYBRID]
=item retrieve($) [PUBLIC,HYBRID]

  Usage     : my $result = Object::Factory->store($object, $file);
              my $result = $instance->freeze($object, $file);
              my $object = Object::Factory->retrieve($file);
              my $object = $instance->retrieve($file);
  Purpose   : Serialize/deserialize given object into/from given file
  Arguments : Same as Storable::store() and Storable::retrieve() methods
  Returns   : Same as Storable::store() and Storable::retrieve() methods

=cut

sub store(@)
{
	my $invocant = shift;
	return Storable::store(@_);
}
sub retrieve(@)
{
	my $invocant = shift;
	return Storable::retrieve(@_);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Object::Factory instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Object::Factory $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	### Initializing Object.Factory instance...
	return $self;
}

# _get_type_map() [PROTECTED]
#
#  Usage     : my $map = Object::Factory->_get_type_map(); OR
#              my $map = $instance->_get_type_map();
#  Purpose   : Get the TYPE => CLASS map
#  Arguments : None
#  Returns   : HASH REFERENCE $map
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _get_type_map()
{
	my $invocant = shift;
	confess "Method needs to be implemented in subclass";
}

=pod

=back

=cut

1;
