package Package::Factory;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Package::Factory - Short module description

=head1 SYNOPSIS

 use Package::Factory;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# LIBRARY LOADING SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use File::Find;

use Data::Dumper;                                    #--- DEVELOPMENT ONLY ---#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Package::Deb;
use Package::Tar;

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#
use constant DEV_BIN_SUBDIR			=> "bin";
use constant DEV_CERTS_SUBDIR		=> "certs";
use constant DEV_CONF_SUBDIR		=> "conf";
use constant DEV_DATA_SUBDIR		=> "data";
use constant DEV_DEPLOY_SUBDIR		=> "deploy";
use constant DEV_EXAMPLES_SUBDIR	=> "examples";
use constant DEV_DOC_SUBDIR			=> "doc";
use constant DEV_LIB_SUBDIR			=> "lib";
use constant DEV_PROTOTYPES_SUBDIR	=> "prototypes";
use constant DEV_RESOURCES_SUBDIR	=> "resources";
use constant DEV_RULES_SUBDIR		=> "rules";
use constant DEV_SCRIPTS_SUBDIR		=> "scripts";
use constant DEV_TEMPLATES_SUBDIR	=> "templates";
use constant DEV_TESTS_SUBDIR		=> "tests";
use constant DEV_WWW_SUBDIR			=> "www";

use constant DEV_DEPLOY_CONTROL_SUBDIR	=> "ctrl";

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Translate package type to apropriate class
my $TYPE2CLASS = {
		'deb' => 'Package::Deb',
		'tar' => 'Package::Tar',
	};

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA);
    $VERSION 	= '0.1';
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Package::Factory->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Package::Factory reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my ($dev_root, $wanted_folders, $packages) = @_;

	my $self = bless ({}, $class);

	$self->{DEV_ROOT}		= $dev_root;
	$self->{WANTED_FILES}	= {};
	$self->{TEMPLATES}		= {};
	$self->{PACKAGES}		= $packages;

	$self->_find_wanted_files(@$wanted_folders);
	$self->_find_template_files();

	$self->_check_presence_in_package();

	return $self;
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub dev_root_dir()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	return $self->{DEV_ROOT};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub deploy_dir()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	return $self->dev_root_dir . '/' . DEV_DEPLOY_SUBDIR;
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub deploy_control_dir()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	return $self->deploy_root_dir . '/' . DEV_DEPLOY_CONTROL_SUBDIR;
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub templates_dir()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	return $self->dev_root_dir . '/' . DEV_TEMPLATES_SUBDIR;
}

#-------------------------------------------------------------------------------

# Upload all given packages via scp
sub upload_packages_via_scp($$)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($destination, $packages) = @_;
	croak ((caller(0))[3] . ": upload destination must be specified") unless $destination;

	foreach my $package (@$packages)
	{
		$package->upload_via_scp($destination);
	}
}

# Upload all given packages via ftp
sub upload_packages_via_ftp($$)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($destination, $packages) = @_;
	croak ((caller(0))[3] . ": upload destination must be specified") unless $destination;

	foreach my $package (@$packages)
	{
		$package->upload_via_ftp($destination);
	}
}

#-------------------------------------------------------------------------------

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub build_package($$$)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($type, $suite, $package_id) = @_;
	croak ((caller(0))[3] . ": invalid package type '$type'") unless $TYPE2CLASS->{$type};
	croak ((caller(0))[3] . ": unknown package '$package_id'") unless $self->{PACKAGES}->{$package_id};

	my $package_class = $TYPE2CLASS->{$type};

	my $package = $package_class->new($self->deploy_dir, $package_id, $suite, $self->{PACKAGES}->{$package_id}, $self->{TEMPLATES});
	$package->build();

	return $package;
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub build_packages($$$)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($type, $suite, $package_ids) = @_;
	croak ((caller(0))[3] . ": invalid package type '$type'") unless $TYPE2CLASS->{$type};

	my $packages_built = [];
	foreach my $package_id (@$package_ids)
	{
		print "-----------------------------------------------------------------------\n";
		print " Package $package_id\n";
		print "-----------------------------------------------------------------------\n";

		my $package = $self->build_package($type, $suite, $package_id);
		push(@$packages_built, $package);
	}
	return $packages_built;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _check() [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _check()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	die "Invalid deploy root dir '" . $self->deploy_root_dir() . "'\n" 			unless -d $self->deploy_root_dir();
}

sub _add_to_list
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	# Skip directories
	return unless -f $File::Find::name;

	# Mark the file for adding to the package
	$self->{WANTED_FILES}->{$File::Find::name} = $_;
}

sub _find_wanted_files(@)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my @wanted_dirs = @_;

	foreach my $dir (@wanted_dirs)
	{
		find({'wanted' => sub { $self->_add_to_list }, 'untaint' => 1}, $dir);
	}
}

sub _add_to_t_list
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	# Skip directories
	return unless -f $File::Find::name;

	# Mark the file as a template
	$self->{TEMPLATES}->{$_} = $File::Find::name;
}

sub _find_template_files()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	find({'wanted' => sub { $self->_add_to_t_list }, 'untaint' => 1}, $self->templates_dir);
}

sub _check_presence_in_package()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	# Walk through all defined packages
	foreach my $package (keys(%{$self->{PACKAGES}}))
	{
		# Walk through all files contained in designated package
		foreach my $file (keys(%{$self->{PACKAGES}->{$package}->{'contents'}}))
		{
			delete $self->{WANTED_FILES}->{$file};
			die "Requested file $file does not exist on filesystem\n" unless -f $file;
		}
	}
	# Check, what files are missing
	if (scalar keys(%{$self->{WANTED_FILES}}))
	{
		my $count = scalar keys(%{$self->{WANTED_FILES}});
		die "Following $count file(s) are not contained in any package:\n" . join("\n", sort(keys(%{$self->{WANTED_FILES}}))) . "\n";
	}
}

=pod

=back

=cut

1;
