package Package::General;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Package::General - Short module description

=head1 SYNOPSIS

 use Package::General;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# LIBRARY LOADING SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use File::Find;
use File::Basename;
use File::stat;

#use Data::Dumper;                                    #--- DEVELOPMENT ONLY ---#

#-- Perl CPAN modules ---------------------------------------------------------#
use Text::Template;

#-- Custom application modules ------------------------------------------------#

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION @ISA);
    $VERSION 	= '0.1';
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Package::General->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Package::General reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my ($deploy_root, $id, $suite, $config, $templates) = @_;

	my $self = bless ({}, $class);

	$self->{DEPLOY_ROOT}	= $deploy_root;
	$self->{ID} 			= $id;
	$self->{SUITE} 			= $suite;
	$self->{DIR_STRUCTURE}	= $config->{'dir_structure'};
	$self->{CONTENTS} 		= $config->{'contents'};
	$self->{TEMPLATE_VALS}	= $config->{'template_values'};
	$self->{TEMPLATES} 		= $templates;
	$self->{META}			= {};
	$self->{PACKAGE_FILE}	= undef;

	$self->_load_metadata();

	return $self->_init(@_);
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub DESTROY
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	$self->_store_metadata();
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub deploy_dir()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{DEPLOY_ROOT};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub deploy_control_dir()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{DEPLOY_ROOT} . "/ctrl/$self->{ID}";
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub deploy_package_dir()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{DEPLOY_ROOT} . "/package";
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub id()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{ID};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub suite()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{SUITE};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub name()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{META}->{NAME};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub filename()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub arch()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{META}->{ARCH};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub service()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{META}->{SERVICE};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub version()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return join('.', ($self->version_major, $self->version_minor));
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub version_major()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{META}->{VERSION_MAJOR};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub version_minor()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	return $self->{META}->{VERSION_MINOR};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub inc_version_major()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	my @digits = split('.', $self->{META}->{VERSION_MAJOR});
	$digits[$#digits]++;
	$self->{META}->{VERSION_MAJOR} = join('.', @digits);

	return $self->{META}->{VERSION_MAJOR};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub inc_version_minor()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	$self->{META}->{VERSION_MINOR}++;

	return $self->{META}->{VERSION_MINOR};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub structure($)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($key,) = @_;

	return $self->{DIR_STRUCTURE} unless defined $key;

	return undef unless $self->{DIR_STRUCTURE}->{$key};

	return $self->deploy_package_dir . $self->{DIR_STRUCTURE}->{$key};
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub contents()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	return $self->{CONTENTS};
}

#-------------------------------------------------------------------------------
# BUILD HELPERS
#-------------------------------------------------------------------------------

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub generate_file_from_template($$$)
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($template_file, $dest_file, $values) = @_;

	my $template = Text::Template->new(SOURCE => $template_file, UNTAINT => 1) or die "Couldn't construct template from file '$template_file': $Text::Template::ERROR\n";
    my $result = $template->fill_in(HASH => $values);

	# If the template was filled properly, store the result to the configuration file
    if (defined $result)
    {
		# Write the contents to the destination file
		open (DEST_FILE, ">$dest_file") or die "Failed to open file '$dest_file': $!\n";
		print DEST_FILE "$result\n";
		close(DEST_FILE);

		# Adjust the file permissions according to the template file
		my $sb = stat($template_file);
		my $perms = sprintf("%04o", ($sb->mode & 07777));
		chmod(oct($perms), $dest_file);
	}
    else {
		die "Couldn't fill in template file '$template_file': $Text::Template::ERROR\n";
	}
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub upload_via_scp($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($destination,) = @_;
	croak ((caller(0))[3] . ": upload destination must be specified") unless $destination;

	print " - Uploading package " . $self->{PACKAGE_FILE} . " to the remote destination '$destination' via SCP\n";

	die "Package was not yet generated\n" unless $self->{PACKAGE_FILE};
	die "Package '" . $self->{PACKAGE_FILE} . "' is not ordinary file\n" unless -f $self->{PACKAGE_FILE};

	system('/usr/bin/scp ' . $self->{PACKAGE_FILE} . ' ' . $destination) == 0 or die "Error executing system command: $?\n";
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub upload_via_ftp($)
{
	my $self = shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

#-------------------------------------------------------------------------------
# BUILD SEQUENCE STEPS
#-------------------------------------------------------------------------------

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub create_dir_structure()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	print " - Creating package directory structure\n";

	foreach my $key (keys %{$self->structure})
	{
		my $dir = $self->structure($key);
		print "   - Creating directory '$dir'\n";
		system('/bin/mkdir -p ' . $dir) == 0 or die "Error executing system command: $?\n";
	}
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub remove_dir_structure()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	print " - Removing package directory structure\n";

	system('/bin/rm -rf ' . $self->deploy_package_dir) == 0 or die "Error executing system command: $?\n";
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub copy_speciffic_contents()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub copy_contents()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	print " - Copying package contents to prepared directory structure\n";

	foreach my $file (keys(%{$self->contents}))
	{
		my $file_basename 	= basename($file);
		my $destination 	= $self->contents->{$file};

		if ($destination =~ /\{(.+)\}/)
		{
			my $key = $1;
			die "Unknown copy destination '$key' for file $file\n" unless $self->structure($key);
			my $value = $self->structure($key);
			$destination =~ s/\{.+\}/$value/;
		}
		unless (-d $destination)
		{
			print "  - Preparing destination '$destination'\n";
			system('/bin/mkdir -p ' . $destination) == 0 or die "Error executing system command: $?\n";
		}
		if (exists($self->{TEMPLATES}->{$file_basename}))
		{
			print "   - Generating file '$file_basename' from template and storing it into '$destination/$file_basename'\n";
			die "Template file '$file_basename' does not have any values specified within package\n" unless exists($self->{TEMPLATE_VALS}->{$file_basename});
			$self->generate_file_from_template($self->{TEMPLATES}->{$file_basename}, "$destination/$file_basename", $self->{TEMPLATE_VALS}->{$file_basename});
		}
		else {
			print "   - Copying file '$file' to '$destination'\n";
			system('/bin/cp -R ' . $file . ' ' . $destination) == 0 or die "Error executing system command: $?\n";
		}
	}
	print "   - Copying copyright file\n";
	system('/bin/cp -R ' . $self->deploy_control_dir . '/copyright ' . 	$self->structure('PACKAGE_DOC_DIR')) == 0 or die "Error executing system command: $?\n";
	print "   - Storing version file\n";
	open (VFILE, '>>'.$self->structure('PACKAGE_DOC_DIR').'/version');
	print VFILE $self->version."\n";
	close (VFILE);

	$self->copy_speciffic_contents();
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub generate_man_page()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	# Skip directories
	return unless -f $File::Find::name;

	# Libraries and binaries have different man page locations
	my $pattern_bin = $self->structure('PACKAGE_BIN_DIR') || undef;
	my $pattern_lib = $self->structure('PACKAGE_LIB_DIR') || undef;
	if ($pattern_bin and $File::Find::dir =~ /^$pattern_bin/)
	{
		my $target_name = $self->structure('PACKAGE_MAN1_DIR') . '/' . $_ . '.1.gz';
		print "   - Creating binary module man page: $_\n";
		system('/usr/bin/pod2man --section=1 ' . $File::Find::name . ' | /bin/gzip --best > ' . $target_name) == 0 or die "Error executing system command: $?\n";
	}
	elsif ($pattern_lib and $File::Find::dir =~ /^$pattern_lib/)
	{
		my $man_page_name = $File::Find::name;
		$man_page_name =~ s\$pattern_lib/\\;
		$man_page_name =~ s\/\::\g;
		$man_page_name =~ s\.pm$\.3pm.gz\;

		my $target_name = $self->structure('PACKAGE_MAN3_DIR') . '/' . $man_page_name;

		print "   - Creating library module man page: $man_page_name\n";
		system('/usr/bin/pod2man --section=3pm ' . $File::Find::name . ' | /bin/gzip --best > ' . $target_name) == 0 or die "Error executing system command: $?\n";
	}
	else {
		warn "Unknown directory encountered when generating man pages\n";
	}
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub generate_man_pages()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	print " - Generating manual pages\n";

	find({'wanted' => sub { $self->generate_man_page(); }, 'untaint' => 1}, $self->structure('PACKAGE_BIN_DIR')) if $self->structure('PACKAGE_BIN_DIR');
	find({'wanted' => sub { $self->generate_man_page(); }, 'untaint' => 1}, $self->structure('PACKAGE_LIB_DIR')) if $self->structure('PACKAGE_LIB_DIR');
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub calculate_size()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	print " - Calculating package size\n";

	my $package_root_dir = $self->deploy_package_dir();
	my $size = `/usr/bin/du -k -s $package_root_dir | /usr/bin/cut -f 1`;
	$size =~ s/\s//g;
	return $size;
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub generate_package($$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

=item [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub build()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Package::General instance
#  Arguments :
#  Returns   : Package::General reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": method needs implementation");
}

# _check() [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _check()
{
	my $self 	= shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	die "Invalid deploy root dir '" . $self->deploy_dir() . "'\n" 			unless -d $self->deploy_dir();
	die "Invalid deploy control dir '" . $self->deploy_control_dir() . "'\n" 	unless -d $self->deploy_control_dir();

	# Let the subclassess perform aditional checking
	$self->_check_sub();
}

# _load_metadata() [PROTECTED]
#
#  Usage     :
#  Purpose   : Loads package metadata from persistent text file
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _load_metadata()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	my $meta_file_name = $self->deploy_control_dir() . '/meta';

	print " - Loading package metadata from file '$meta_file_name'\n";

	open(my $file, $meta_file_name) or die "Could not open metadata file '$meta_file_name'\n";
	while (my $line = <$file>)
	{
		$line =~ s/^\s*//;
		$line =~ s/\s*$//;

		if ($line =~ /^NAME=(.+)$/)
		{
			$self->{META}->{NAME} = $1;
		}
		elsif ($line =~ /^ARCH=(.+)$/)
		{
			$self->{META}->{ARCH} = $1;
		}
		elsif ($line =~ /^SERVICE=(.+)$/)
		{
			$self->{META}->{SERVICE} = $1;
		}
		elsif ($line =~ /^VERSION_MAJOR=(\d+(?:\.\d+))$/)
		{
			$self->{META}->{VERSION_MAJOR} = $1;
		}
		elsif ($line =~ /^VERSION_MINOR=(\d+)$/)
		{
			$self->{META}->{VERSION_MINOR} = $1;
		}
		else {
			die "Unknown metadata line: '$line'\n"
		}
	}
	close($file);
}

# _store_metadata() [PROTECTED]
#
#  Usage     :
#  Purpose   : Stores package metadata to persistent text file
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _store_metadata()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	my $meta_file_name = $self->deploy_control_dir() . '/meta';

	print " - Storing package metadata to file '$meta_file_name'\n";

	open(my $file, ">", $meta_file_name) or die "Could not open metadata file '$meta_file_name'\n";
	foreach my $meta (keys %{$self->{META}})
	{
		print $file "$meta=" . $self->{META}->{$meta} . "\n";
	}
	close($file);
}

# _generated_true() [PROTECTED]
#
#  Usage     :
#  Purpose   : Stores package metadata to persistent text file
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _generated_true()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	$self->{PACKAGE_FILE} = $self->deploy_dir() . '/' . $self->filename();
}

=pod

=back

=cut

1;
