package Resource::Registry;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Resource::Registry - Storage for application resources

=head1 SYNOPSIS

 use Resource::Registry;

 # Usage through static methods
 Resource::Registry->resource('test1',1);
 Resource::Registry->resource('test2',2);
 Resource::Registry->resource('test3',3);
 Resource::Registry->resources->{'test1'} = 5;

 Resource::Registry->resources()->{'test1'};
 Resource::Registry->resources();
 Resource::Registry->has_resource('test1');

 # Usage through instance methods
 my $instance = new Resource::Registry();

 $instance->resource('testa',10);
 $instance->resource('testb',20);
 $instance->resource('testc',30);
 $instance->resources->{'testa'} = 50;

 $instance->resources()->{'testa'};
 $instance->resources();
 $instance->has_resource('testa');

=head1 DESCRIPTION

Storage for application resources. During application design, there are
usually global objects (like configurations, ...), that must be accessible
at many different points in the program. One way is to provide access to
such objects is to pass them as arguments. This approach is fine, if the
number of such objects is not high.

Other approach is to use either one object, or a static class, which will
act as a resource registry. Static class is particulary useful, because
there is no need for object instantination.

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION);
    $VERSION 	= '0.01';
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Static class variable, hash containing resource identifiers as keys and
# resources themselves as values.
my $RESOURCES = {};

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item new() [PUBLIC, STATIC]

  Usage     : my $registry = new Resource::Registry();
  Purpose   : Create new resource registry
  Returns   : Create new instance of resource registry
  Arguments : None
  Throws    : Croaks, if invoked on object
  Comments  :
  See Also  :

=cut

sub new
{
	my $class = shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;
	my $self = bless ({}, $class);
	$self->{RESOURCES} = {};
	return $self;
}

=item has_resource() [PUBLIC, HYBRID]

  Usage     : my $result = Resource::Registry->has_resource('resource_id'); or
			  my $result = $instance->has_resource('resource_id');
  Purpose   : Test, if the resource with given id is in the registry
  Returns   : 1 if the resource exists, 0 if it does not
  Arguments : string $resource - identifier of the resource
  Throws    : Nothing
  Comments  :
  See Also  :

=cut

sub has_resource {
	my ($invocant, $resource) = @_;

	# If the invocant is the object reference, look inside the object`s storage
	if (ref $invocant)
	{
		return exists($invocant->{RESOURCES}->{$resource}) ? 1 : 0;
	}
	# If the invocant is the class name, look inside the static class storage
	else {
		return exists($RESOURCES->{$resource}) ? 1 : 0;
	}
}

=item resources() [PUBLIC, HYBRID]

  Usage     : my $resources = Resource::Registry->resources(); or
			  my $resources = $instance->resources();
  Purpose   : Get hash containing all resources stored in the registry
  Returns   : hash reference containing all resources, reference identifiers as keys
  Arguments : None
  Throws    : Nothing
  Comments  :
  See Also  :

=cut

sub resources {
	my $invocant = shift;

	# If the invocant is the object reference, look inside the object`s storage
	if (ref $invocant)
	{
		return $invocant->{RESOURCES}
	}
	# If the invocant is the class name, look inside the static class storage
	else {
		return $RESOURCES;
	}
}

=item resource() [PUBLIC, HYBRID]

  Usage     : my $resource = Resource::Registry->resource('test1',1); or
			  my $resource = $instance->resource('testa',10);
  Purpose   : Get and/or set the resource
  Returns   : resource itself
  Arguments : string $resource - resource identifier
			  mixed $value - resource itself
  Throws    : Nothing
  Comments  :
  See Also  :

=cut

sub resource {
	my ($invocant, $resource, $value) = @_;

	# If the invocant is the object reference, look inside the object`s storage
	if (ref $invocant)
	{
		if (defined($value))
		{
			$invocant->{RESOURCES}->{$resource} = $value;
		}
		return $invocant->{RESOURCES}->{$resource}
	}
	# If the invocant is the class name, look inside the static class storage
	else {
		if (defined($value))
		{
			$RESOURCES->{$resource} = $value;
		}
		return $RESOURCES->{$resource};
	}
}

=pod

=back

=cut

1;
