package Tie::Array::Sorted::Stable;

use 5.006;

use strict;
use warnings;

use base 'Tie::Array';

our $VERSION = '1.41';

=head1 NAME

Tie::Array::Sorted::Stable - An array which is kept sorted with stable sort

=head1 SYNOPSIS

	use Tie::Array::Sorted::Stable;

	tie @a, "Tie::Array::Sorted::Stable", sub { $_[0] <=> $_[1] };

	push @a, 10, 4, 7, 3, 4;
	print "@a"; # "3 4 4 7 10"

=head1 DESCRIPTION

This presents an ordinary array, but is kept sorted. All pushes and
unshifts cause the elements in question to be inserted in the
appropriate location to maintain order. This module is inspired by
C<Tie::Array::Sorted> module, but apart from that this implementation
uses stable sorting algorithm. This stability costs roughly 10-15% in
performance based on the size of the array (the bigger the array the
smaller the cost).

Direct stores (C<$a[10] = "wibble">) effectively splice out the original
value and insert the new element. It's not clear why you'd want to use
direct stores like that, but this module does the right thing if you do.

If you don't like the ordinary lexical comparator, you can provide your
own; it should compare the two elements it is given. For instance, a
numeric comparator would look like this:

	tie @a, "Tie::Array::Sorted::Stable", sub { $_[0] <=> $_[1] }

Whereas to compare a list of files by their sizes, you'd so something
like:

	tie @a, "Tie::Array::Sorted::Stable", sub { -s $_[0] <=> -s $_[1] }

=head1 LAZY SORTING

If you do more stores than fetches, you may find
L<Tie::Array::Sorted::Lazy> more efficient.

=head1 SORTING ALGORITHM

Used sorting algorithm was inspired by "Binary Search Algorithm" with
deferred detection of equality (DDE), some initial information can be
found here:

	http://en.wikipedia.org/wiki/Binary_search_algorithm#Deferred_detection_of_equality

The deferred detection of equality must have been used instead of regular
binary search, because the latter does not provide stability. Algorithm below
is using one modification against the original version, which identified the
left most element from set of elements with same value. Version below identifies
the right most element, which is the prerequisite for stable sort.

=cut

sub TIEARRAY {
	my ($class, $comparator) = @_;
	bless {
		array => [],
		comp  => (defined $comparator ? $comparator : sub { $_[0] cmp $_[1] })
	}, $class;
}

sub STORE {
	my ($self, $index, $elem) = @_;
	splice @{ $self->{array} }, $index, 0;
	$self->PUSH($elem);
}

# Following sorting algorithm was inspired by "Binary Search Algorithm" with
# deferred detection of equality (DDE), some initial information can be found here:
#
#	http://en.wikipedia.org/wiki/Binary_search_algorithm#Deferred_detection_of_equality
#
# The deferred detection of equality must have been used instead of regular
# binary search, because the latter does not provide stability. Algorithm below
# is using one modification against the original version, which identified the
# left most element from set of elements with same value. Version belowidentifies
# the right most element, which is the prerequisite for stable sort.
sub PUSH {
	my ($self, @elems) = @_;

	my ($lo, $hi, $mid, $val, $cmp);
	for my $elem (@elems) {
		($lo, $hi) = (0, $#{$self->{array}});
		# Continually narrow search until just one element remains
		while ($lo < $hi) {
			$mid = int(($lo + $hi) / 2);
			# Code must guarantee the interval is reduced at each iteration
			last unless $mid < $hi;
			# note: 0 <= $lo < $hi implies imid will always be less than imax
			$val = $self->{array}[$mid];
			$cmp = $self->{comp}($val, $elem);

			# Reduce the search:
			#   middle value is greater than the value being inserted, make the
			#   'ceiling' lower
			if ($cmp > 0) {
				$hi = $mid;
			}
			#   middle value is equal or lower than the value being inserted,
			#   make the 'floor' higher
			else {
				$lo = $mid + 1;
			}
		}
		# At exit of while:
		# 	if the array was empty, then $hi < $lo
		#   otherwise $hi == $lo

		# Now the deferred test for equality
		if ($hi == $lo) {
			$val = $self->{array}[$lo];
			$cmp = $self->{comp}($val, $elem);

			# Compare the element being inserted with element selected by previous
			# binary search and insert it AFTER that in case of '<=' relation
			if ($cmp <= 0) {
				splice(@{ $self->{array} }, $lo+1, 0, $elem);
				next;
			}
		}
		# Otherwise insert the element instead of selected element (or at first
		# place in empty array)
		splice(@{ $self->{array} }, $lo, 0, $elem);
	}
}

sub UNSHIFT { goto &PUSH }

sub FETCHSIZE { scalar @{ $_[0]->{array} } }
sub STORESIZE { $#{ $_[0]->{array} } = $_[1] - 1 }
sub FETCH     { $_[0]->{array}->[ $_[1] ] }
sub CLEAR     { @{ $_[0]->{array} } = () }
sub POP       { pop(@{ $_[0]->{array} }) }
sub SHIFT     { shift(@{ $_[0]->{array} }) }

sub EXISTS { exists $_[0]->{array}->[ $_[1] ] }
sub DELETE { delete $_[0]->{array}->[ $_[1] ] }

1;

=head1 AUTHOR

Author: Jan Mach

Author of original module: Simon Cozens

Maintainer of original module: Tony Bowden

=head1 BUGS and QUERIES

Please direct all correspondence regarding this module to:
	bug-Tie-Array-Sorted@rt.cpan.org

This module was originall written as part of the L<Plucene> project.
However, as Plucene no longer uses this, it is effectively unmaintained.

=head1 COPYRIGHT AND LICENSE

  Copyright (C) 2003-2006 Simon Cozens and Tony Bowden.

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License; either version 2 of the License,
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.


=cut
