package Time::Timer;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Time::Timer - Short module description

=head1 SYNOPSIS

 use Time::Timer;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Parent::Module(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Time::HiRes 'time';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#
my $TIMER;

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Time::Timer->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Time::Timer reference on success, croaks on failure
  Throws    : Confesses, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
	my $class = shift;
	confess "Class method not invoked on class" if blessed($class);
	### Creating Time.Timer instance...
	my $self = bless ({}, $class);
	return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Confesses, if invoked on class

=cut

sub DESTROY()
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	### Destroying Time.Timer instance...
}

=item start() [PUBLIC]

  Usage     : Time::Timer->start();
              $timer->start();
  Purpose   : Start the timer (and return as quickly as possible)
  Arguments : None
  Returns   : Nothing
  Throws    : Nothing

=cut

sub start()
{
	my $invocant = shift;
	unless (blessed($invocant)) { $TIMER = time; }
	else                        { $invocant->{TIMER} = time; }
	return 0;
}

=item mark() [PUBLIC]

  Usage     : my $duration = Time::Timer->mark();
              my $duration = $timer->mark();
  Purpose   : Check the timer and return the duration from previous start/stop call
  Arguments : None
  Returns   : FLOAT $duration
  Throws    : Nothing

=cut

sub mark()
{
	my $invocant = shift;
	my $t = time; my $p;
	unless (blessed($invocant)) { $p = $TIMER; }
	else                        { $p = $invocant->{TIMER}; }

	return 0 unless $p;
	return ($t - $p);
}

=item stop() [PUBLIC]

  Usage     : my $duration = Time::Timer->stop();
              my $duration = $timer->stop();
  Purpose   : Reset the timer and return the duration from previous start/stop call
  Arguments : None
  Returns   : FLOAT $duration
  Throws    : Nothing

=cut

sub stop()
{
	my $invocant = shift;
	my $t = time; my $p;
	unless (blessed($invocant)) { $p = $TIMER; $TIMER = $t; }
	else                        { $p = $invocant->{TIMER}; $invocant->{TIMER} = $t; }

	return 0 unless $p;
	return ($t - $p);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Time::Timer instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Time::Timer $self reference
#  Throws    : Confesses, if not invoked on instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
	my $self = shift;
	confess "Instance method not invoked on object instance" unless blessed($self);
	### Initializing Time.Timer instance...
	return $self;
}

=pod

=back

=cut

1;
