package Unix::Accounts;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Unix::Accounts - Unix account manager

=head1 SYNOPSIS

 use Unix::Accounts;

=head1 DESCRIPTION

=head1 USAGE

=head1 BUGS

=head1 SUPPORT

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION);
    $VERSION 	= '0.01';
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item new() [PUBLIC,STATIC]

  Usage     : my $accounts = Unix::Accounts->new($user_account_file, $group_account_file);
  Purpose   : Create new instance of log channel
  Returns   : Reference to the new instance
  Arguments : All arguments are passed to the _init() function
  Throws    : Croaks if invoked on object
  Comments  : This method should not be overloaded, descendants should implement their version of _init() method instead
  See Also  : _init() method

=cut

sub new
{
	my $class 	= shift;
	croak ((caller(0))[3] . ": class method invoked on object") if ref $class;

	my $self 	= {};
	bless($self, $class);
	return $self->_init(@_);
}

=item get_user($) [PUBLIC]

  Usage     : my $account = $accounts->get_user('root');
  Purpose   : Get information about user account
  Returns   : hash reference containing information about user account
				 - login name
				 - optional encrypted password
				 - numerical user ID
				 - numerical group ID
				 - user name or comment field
				 - user home directory
				 - optional user command interpreter
  Arguments : string | integer $user_id - user login or UID
  Throws    : Croaks, if invoked on class or invalid arguments given
  Comments  :
  See Also  :

=cut

sub get_user($) {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($user_id,) = @_;
	croak ((caller(0))[3] . ": you have to provide user login name or UID") if not defined $user_id;

	# Lazy load the user accounts from file
	if (not defined $self->{USERS_LOGIN}) { $self->_load_users(); }

	# If the user_id is all numbers, look for account by UID
	if ($user_id =~ /^\d+$/)
	{
		if(exists($self->{USERS_UID}->{$user_id}))
		{
			return $self->{USERS_UID}->{$user_id};
		}
	}
	else {
		if(exists($self->{USERS_LOGIN}->{$user_id}))
		{
			return $self->{USERS_LOGIN}->{$user_id};
		}
	}
	return {};
}

=item get_logins() [PUBLIC]

  Usage     : my $logins = $accounts->get_logins('login', '^r.+$');
  Purpose   : Get list of all user logins, optional arguments allow to filter them based on the PCRE applied to the
  Returns   : array reference containing all user logins mathing the criteria
  Arguments : enum $field - field to which to apply the PCRE pattern
					(password|uid|shell|home|geccos|gid|login)
			  string $pattern - PCRE pattern against which to validate fields value
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_logins {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($field, $pattern) = @_;
	if (not defined $self->{USERS_LOGIN}) { $self->_load_users(); }

	my $retval = [];

	# If the field and pattern were both not specified, return all users
	unless (defined($field) and defined($pattern))
	{
		foreach my $user (keys(%{$self->{USERS_UID}}))
		{
			push(@$retval, $self->{USERS_UID}->{$user}->{'login'})
		}
	}
	# Otherwise return only those whose field value mathes the given pattern
	else {
		foreach my $user (keys(%{$self->{USERS_UID}}))
		{
			if ($self->{USERS_UID}->{$user}->{$field} =~ /$pattern/)
			{
				push(@$retval, $self->{USERS_UID}->{$user}->{'login'})
			}
		}
	}
	return $retval;
}

=item get_users() [PUBLIC]

  Usage     : my $logins = $accounts->get_users('login', '^r.+$');
  Purpose   : Get list of all user information, optional arguments allow to filter them based on the PCRE applied to the
  Returns   : array reference containing hash references containing user account data matching criteria
  Arguments : enum $field - field to which to apply the PCRE pattern
					(password|uid|shell|home|geccos|gid|login)
			  string $pattern - PCRE pattern against which to validate fields value
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_users {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($field, $pattern) = @_;
	if (not defined $self->{USERS_LOGIN}) { $self->_load_users(); }

	my $retval = [];

	# If the field and pattern were both not specified, return all users
	unless (defined($field) and defined($pattern))
	{
		foreach my $user (keys(%{$self->{USERS_UID}}))
		{
			push(@$retval, $self->{USERS_UID}->{$user})
		}
	}
	# Otherwise return only those whose field value mathes the given pattern
	else {
		foreach my $user (keys(%{$self->{USERS_UID}}))
		{
			if ($self->{USERS_UID}->{$user}->{$field} =~ /$pattern/)
			{
				push(@$retval, $self->{USERS_UID}->{$user})
			}
		}
	}
	return $retval;
}

=item get_group($) [PUBLIC]

  Usage     : my $account = $accounts->get_group('root');
  Purpose   : Get information about group account
  Returns   : hash reference containing information about user account
				 - group_name
				 - password
				 - GID
				 - user_list(comma separated)
  Arguments : string | integer $group_id - group name or GID
  Throws    : Croaks, if invoked on class or invalid arguments given
  Comments  :
  See Also  :

=cut

sub get_group {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($group_id, $type) = @_;
	croak ((caller(0))[3] . ": you have to suply group name") if not defined $group_id;

	# Lazy load the group accounts from file
	if (not defined $self->{GROUPS_NAME}) { $self->_load_groups(); }

	# If the user_id is all numbers, look for account by UID
	if ($group_id =~ /^\d+$/)
	{
		if(exists($self->{GROUPS_GID}->{$group_id}))
		{
			return $self->{GROUPS_GID}->{$group_id};
		}
	}
	else {
		if(exists($self->{GROUPS_NAME}->{$group_id}))
		{
			return $self->{GROUPS_NAME}->{$group_id};
		}
	}
	return {};
}

=item get_user_real() [PUBLIC]

  Usage     : my $account = $accounts->get_user_real();
  Purpose   : Get information about current real user account
  Arguments : None
  Returns   : HASH REFERENCE containing information about user account
				 - login name
				 - optional encrypted password
				 - numerical user ID
				 - numerical group ID
				 - user name or comment field
				 - user home directory
				 - optional user command interpreter
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_user_real() {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	return $self->get_user($<);
}

=item get_user_effective() [PUBLIC]

  Usage     : my $account = $accounts->get_user_effective();
  Purpose   : Get information about current effective user account
  Arguments : None
  Returns   : HASH REFERENCE containing information about user account
				 - login name
				 - optional encrypted password
				 - numerical user ID
				 - numerical group ID
				 - user name or comment field
				 - user home directory
				 - optional user command interpreter
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_user_effective() {
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	return $self->get_user($>);
}

# _init(;$$) [PROTECTED]
#
#  Usage     : Used from constructor as follows: return $self->_init(@_);
#  Purpose   : Initialize newly created channel instance
#  Returns   : Must return $self
#  Arguments : STRING $userfile - name of the user account file
#			   STRING $groupfile - name of the group account file
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  : new() method

sub _init(;$$)
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
	my ($userfile, $groupfile) = @_;

	unless (defined($userfile))  { $userfile  = "/etc/passwd"; }
	unless (defined($groupfile)) { $groupfile = "/etc/group"; }
	$self->{UF} = $userfile;
	$self->{GF} = $groupfile;

	return $self;
}

# _load_users() [PROTECTED]
#
#  Usage     : $self->_load_users();
#  Purpose   : Load user accounts from file given during initializations
#  Returns   : Nothing
#  Arguments : None
#  Throws    : Croaks, if invoked on class
#			  Dies, if account file could not be opened
#  Comments  : Used on demand from get_(...) methods
#  See Also  :

sub _load_users()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	$self->{USERS_LOGIN} 	= {};
	$self->{USERS_UID} 		= {};
	open(USERFILE, $self->{UF}) or die("Could not open user accounts file $self->{UF}\n");
	while (my $line = <USERFILE>)
	{
		chomp($line);
		if ($line =~ /#.*/)
		{
			next;
		}
		my @keys 	= ('login','password','uid','gid','geccos','home','shell');
		my @values 	= split(/:/, $line);
		my %uinfo;
		@uinfo{@keys} = @values;
		$self->{USERS_LOGIN}->{$uinfo{'login'}} = \%uinfo;
		$self->{USERS_UID}->{$uinfo{'uid'}} = \%uinfo;
	}
	close(USERFILE);
}

# _load_groups() [PROTECTED]
#
#  Usage     : $self->_load_groups();
#  Purpose   : Load group accounts from file given during initializations
#  Returns   : Nothing
#  Arguments : None
#  Throws    : Croaks, if invoked on class
#			  Dies, if account file could not be opened
#  Comments  : Used on demand from get_(...) methods
#  See Also  :

sub _load_groups()
{
	my $self = shift;
	croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

	$self->{GROUPS_NAME} 	= {};
	$self->{GROUPS_GID} 	= {};
	open(GROUPFILE, $self->{GF}) or die("Could not open group accounts file $self->{GF}\n");
	while (my $line = <GROUPFILE>)
	{
		chomp($line);
		if ($line =~ /#.*/)
		{
			next;
		}
		my @keys 	= ('name','password','gid','users');
		my @values 	= split(/:/, $line);
		$values[3]	= defined($values[3]) ? [split(/,/, $values[3])] : [];
		my %ginfo;
		@ginfo{@keys} = @values;
		$self->{GROUPS_NAME}->{$ginfo{'name'}} = \%ginfo;
		$self->{GROUPS_GID}->{$ginfo{'gid'}} = \%ginfo;
	}
	close(GROUPFILE);
}

=pod

=back

=cut

1;
