package Utils::SSH;
use strict;
use warnings;

################################################################################
#
# DOCUMENTATION SECTION
#
################################################################################

=head1 NAME

Utils::String - String related utilities

=head1 SYNOPSIS

 use Utils::SSH;

 # Check, if the .ssh folder exists and has correct permissions
 my $error = ssh_folder_check('root');

 # Create the .ssh folder for given user
 my $error = ssh_folder_create('root');



=head1 DESCRIPTION

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

################################################################################
#
# INITIALIZATION AND CLEANUP SECTION
#
################################################################################

#-- Perl core modules ---------------------------------------------------------#
use File::stat;
use Cwd qw(abs_path);
use Carp;

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Unix::Accounts;
use Utils::String;

#-- Module initializations ----------------------------------------------------#
BEGIN {
	use Exporter ();

	our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

	$VERSION = 1.00;
	@ISA = qw(Exporter);
	@EXPORT_OK = qw(&ssh_folder_check &ssh_folder_create &ssh_key_add &ssh_keys_get &ssh_key_delete &ssh_key_update &ssh_key_parse &ssh_key_validate);
	%EXPORT_TAGS = (
		folder => [ qw(ssh_folder_check ssh_folder_create) ],
		key => [ qw(ssh_key_add ssh_keys_get ssh_key_delete ssh_key_update ssh_key_parse ssh_key_validate) ],
	);
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

################################################################################
#
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#
################################################################################

#-- Constants -----------------------------------------------------------------#
use constant RE_SSH_KEY_CHAR 	=> '[-+a-zA-Z0-9/=]';
use constant RE_SSH_KEY 		=> '(?:(.*) )?(ssh-dss|ssh-rsa) ('.RE_SSH_KEY_CHAR.'+) ([-_.@\[\]a-zA-Z0-9]+)';
use constant RE_SSH_KEY_SHORT	=> '^(ssh-dss|ssh-rsa) ('.RE_SSH_KEY_CHAR.'+) ([-_.@\[\]a-zA-Z0-9]+)';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#
my $ACCOUNTS = Unix::Accounts->new();

################################################################################
#
# IMPLEMENTATION SECTION
#
################################################################################

=item ssh_folder_check($) [PUBLIC]

  Usage     : my $error = ssh_folder_check('root');
  Purpose   : Check, if the .ssh folder of the user with given name exists
  Arguments : STRING $user - user login or UID
  Returns   : empty STRING on SUCCESS, STRING describing error on FAILURE
  Throws    : Croaks on bigger failure
  Comments  :
  See Also  :

=cut

sub ssh_folder_check($)
{
	my $user = shift;
	croak ((caller(0))[3] . ": user name must be given as argument") unless defined($user);

	# Load user information from /etc/passwd file
	my $user_entry = $ACCOUNTS->get_user($user);
	croak ((caller(0))[3] . ": user '$user' does not exist") unless(defined($user_entry->{'uid'}));

	# Determine the .ssh path using the user's home directory as base path
	my $folder_path = $user_entry->{'home'} . "/.ssh";
	croak ((caller(0))[3] . ": directory tree '$folder_path' is not accessible at some point") unless (defined($folder_path));
	return "Directory '$folder_path' does not exist" unless (-d $folder_path);

	# Retrieve the directory attributes
	my $dir_info = stat($folder_path);
	croak ((caller(0))[3] . ": unable to stat $dir_info: $!") unless (defined($dir_info));

	# Check the directory ownership and permissions
	return "Directory $folder_path is not owned by user $user_entry->{'login'}" if ($dir_info->uid != $user_entry->{'uid'});
	return "Directory $folder_path is not in mode rwx------ : " . $dir_info->mode unless ($dir_info->mode == 16832);

	# Report success by returning no error
    return '';
}

=item ssh_folder_create($) [PUBLIC]

  Usage     : my $error = ssh_folder_create('root');
  Purpose   : Create the .ssh folder for the user with given name
  Arguments : STRING $user - user login or UID
  Returns   : empty STRING on SUCCESS, STRING describing error on FAILURE
  Throws    : Croaks on bigger failure
  Comments  :
  See Also  :

=cut

sub ssh_folder_create($)
{
	my $user = shift;
	croak ((caller(0))[3] . ": user name must be given as argument") unless defined($user);

	# Load user information from /etc/passwd file
	my $user_entry = $ACCOUNTS->get_user($user);
	croak ((caller(0))[3] . ": user '$user' does not exist") unless(defined($user_entry->{'uid'}));

	# Determine and untaint the .ssh path using the user's home directory as base path
	my $folder_path = $user_entry->{'home'} . "/.ssh";
	if ($folder_path =~ /(.*)/)	{ $folder_path = $1; }

	# Check accessibility and existence
	croak ((caller(0))[3] . ": directory tree '". $user_entry->{'home'} ."' is not accessible at some point") unless (defined($folder_path));
	return "Directory '$folder_path' already exists" if (-d $folder_path);

	# If previous checks succeeded, perform the creation
	mkdir($folder_path);
	my ($uid,$gid);
	if ($user_entry->{'uid'} =~ /(.*)/)	{ $uid = $1; }
	if ($user_entry->{'gid'} =~ /(.*)/)	{ $gid = $1; }
	chown($uid, $gid, $folder_path)	or croak ((caller(0))[3] . ": could not chown '$folder_path' to '$uid':'$gid'");
	chmod(0700, $folder_path) or croak ((caller(0))[3] . ": could not chmod '$folder_path' to 0700");

	# Report success by returning no error
	return '';
}

=item ssh_keys_get($;$) [PUBLIC]

  Usage     : my $keys = ssh_keys_get('root');
  Purpose   : Get all (or specific) keys contained in authorized_keys file of given user
  Arguments : STRING $user - user login or UID
              STRING $spec - optional search filter for key comments
  Returns   : ARRAY REFERENCE containing the retrieved keys
  Throws    : Croaks on failure
  Comments  :
  See Also  :

=cut

sub ssh_keys_get($;$)
{
	my ($user, $spec) = @_;
	croak ((caller(0))[3] . ": user name must be given as argument") unless defined($user);

	# Load user information from /etc/passwd file
	my $user_entry = $ACCOUNTS->get_user($user);
	croak ((caller(0))[3] . ": user '$user' does not exist") unless($user_entry->{'uid'});

	# Determine the path to the authorized keys file of given user
	#my $authorized_keys = abs_path($user_entry->{'home'} . "/.ssh/authorized_keys");
	my $authorized_keys = $user_entry->{'home'} . "/.ssh/authorized_keys";
	croak ((caller(0))[3] . ": directory tree '". $user_entry->{'home'} ."' is not accessible at some point") unless (defined($authorized_keys));

	# Create the authorized keys file on demand
	unless (-f $authorized_keys)
	{
		_create_ak_file($user_entry->{'uid'}, $user_entry->{'gid'}, $authorized_keys);
	}

	# Load all keys from given authorized_keys file
	my %keys = _ssh_keys_read($authorized_keys);

	# Return only specific one
	if (defined($spec))
	{
		return [$keys{$spec}];
	}
	# Or all of them
	else {
		return [values(%keys)];
	}
}

=item ssh_key_add($$) [PUBLIC]

  Usage     : my $result = ssh_key_add('root', $key);
  Purpose   : Add new key to the user's authorized_keys file
  Arguments : STRING $user - user login or UID
              STRING $key - SSH key to add
  Returns   : empty STRING on SUCCESS, STRING describing error on FAILURE
  Throws    : Croaks on bigger failure
  Comments  : Internally uses ssh_key_parse() function for parsing given SSH key to chunks
  See Also  : ssh_key_parse() function

=cut

sub ssh_key_add($$)
{
	my ($user, $key) = @_;
	croak ((caller(0))[3] . ": user name must be given as argument") unless defined($user);

	# Load user information from /etc/passwd file
	my $user_entry = $ACCOUNTS->get_user($user);
	croak ((caller(0))[3] . ": user '$user' does not exist") unless($user_entry->{'uid'});

	# Determine the path to the authorized keys file of given user
	#my $authorized_keys = abs_path($user_entry->{'home'} . "/.ssh/authorized_keys");
	my $authorized_keys = $user_entry->{'home'} . "/.ssh/authorized_keys";
	croak ((caller(0))[3] . ": directory tree '". $user_entry->{'home'} ."' is not accessible at some point") unless (defined($authorized_keys));

	# Create the authorized keys file on demand
	unless (-f $authorized_keys)
	{
		_create_ak_file($user_entry->{'uid'}, $user_entry->{'gid'}, $authorized_keys);
	}

	# Parse the given SSH key into chunks
	my @ssh_key = ssh_key_parse($key);
	return 'Invalid SSH key given' unless ($ssh_key[3]);

	# Load all keys current user has and check, if the given one already exists
	my %keys = _ssh_keys_read($authorized_keys);
	return "SSH key $ssh_key[3] already exists" if ($keys{$ssh_key[3]});

	# Add the key to the chain and write all keys back to the file
	$keys{$ssh_key[3]} = \@ssh_key;
	return _ssh_keys_write($authorized_keys, %keys);

}

=item ssh_key_update($$) [PUBLIC]

  Usage     : my $error = ssh_key_update('root', $spec);
  Purpose   : Update existing SSH key
  Arguments : STRING $user - user login or UID
              STRING $key - SSH key to edit
  Returns   : empty STRING on SUCCESS, STRING describing error on FAILURE
  Throws    : Croaks on bigger failure
  Comments  : Internally uses ssh_key_parse() function for parsing given SSH key to chunks
  See Also  : ssh_key_parse() function

=cut

sub ssh_key_update($$)
{
	my ($user, $key) = @_;
	croak ((caller(0))[3] . ": user name must be given as argument") unless defined($user);

	# Load user information from /etc/passwd file
	my $user_entry = $ACCOUNTS->get_user($user);
	croak ((caller(0))[3] . ": user '$user' does not exist") unless($user_entry->{'uid'});

	# Determine the path to the authorized keys file of given user
	#my $authorized_keys = abs_path($user_entry->{'home'} . "/.ssh/authorized_keys");
	my $authorized_keys = $user_entry->{'home'} . "/.ssh/authorized_keys";
	croak ((caller(0))[3] . ": directory tree '". $user_entry->{'home'} ."' is not accessible at some point") unless (defined($authorized_keys));
	return "SSH key $key does not exist" unless (-f $authorized_keys);

	# Parse the given SSH key into chunks
	my @ssh_key = ssh_key_parse($key);
	return 'Invalid SSH key given' unless ($ssh_key[3]);

	# Load all keys current user has and check, if the given one already exists
	my %keys = _ssh_keys_read($authorized_keys);
	return "SSH key $ssh_key[3] does not exist" unless ($keys{$ssh_key[3]});

	# Replace the key in the chain and write all keys back to the file
	$keys{$ssh_key[3]} = \@ssh_key;
	return _ssh_keys_write($authorized_keys, %keys);
}

=item ssh_key_delete($;$) [PUBLIC]

  Usage     : my $error = ssh_key_delete('root', $spec);
  Purpose   : Remove existing SSH key
  Arguments : STRING $user - user login or UID
              STRING $spec - optional search filter for key comments
  Returns   : empty STRING on SUCCESS, STRING describing error on FAILURE
  Throws    : Croaks on bigger failure
  Comments  :
  See Also  :

=cut

sub ssh_key_delete($;$)
{
	my ($user, $spec) = @_;
	croak ((caller(0))[3] . ": user name must be given as argument") unless defined($user);

	# Load user information from /etc/passwd file
	my $user_entry = $ACCOUNTS->get_user($user);
	croak ((caller(0))[3] . ": user '$user' does not exist") unless($user_entry->{'uid'});

	# Determine the path to the authorized keys file of given user
	#my $authorized_keys = abs_path($user_entry->{'home'} . "/.ssh/authorized_keys");
	my $authorized_keys = $user_entry->{'home'} . "/.ssh/authorized_keys";
	croak ((caller(0))[3] . ": directory tree '". $user_entry->{'home'} ."' is not accessible at some point") unless (defined($authorized_keys));
	return "SSH key $spec does not exist" unless (-f $authorized_keys);

	# Read all keys from file
	my %keys = _ssh_keys_read($authorized_keys);

	# Remove the speciffic one
	if ($spec)
	{
		delete $keys{$spec};
	}
	# Or all of them
	else {
		%keys = ();
	}

	# Write the rest of them back
	return _ssh_keys_write($authorized_keys, %keys);
}

=item ssh_key_parse($) [PUBLIC]

  Usage     : my ($options, $keytype, $key, $coment) = ssh_key_parse($pub_key_line);
  Purpose   : Parse given public SSH key using regular expression RE_SSH_KEY
  Arguments : STRING $key - SSH key line as appears in authorized keys file
  Returns   : LIST containing following chunks on SUCCESS:
                STRING $options - SSH options for this key
                STRING $keytype - type of the key ('ssh-dss'|'ssh-rsa')
                STRING $key - base64 encoded key
                STRING $comment - comment
              undef on FAILURE
  Throws    : Croaks, if not given SSH key to parse
  Comments  :
  See Also  :

=cut

sub ssh_key_parse($)
{
	my ($key, )	= @_;
	croak ((caller(0))[3] . ": SSH key must be given as argument") unless defined($key);

	my $pattern = RE_SSH_KEY;
	if ($key =~ /^$pattern$/)
	{
		return ($1,$2,$3,$4);
	}
	return undef;
}

=item ssh_key_parse_short($) [PUBLIC]

  Usage     : my ($keytype, $key, $coment) = ssh_key_parse_short($pub_key_line);
  Purpose   : Parse given public SSH key using regular expression RE_SSH_KEY_SHORT
  Arguments : STRING $key - SSH key line as appears in authorized keys file
  Returns   : LIST containing following chunks on SUCCESS:
                STRING $keytype - type of the key ('ssh-dss'|'ssh-rsa')
                STRING $key - base64 encoded key
                STRING $comment - comment
              undef on FAILURE
  Throws    : Croaks, if not given SSH key to parse
  Comments  :
  See Also  :

=cut

sub ssh_key_parse_short($)
{
	my ($key, )	= @_;
	croak ((caller(0))[3] . ": SSH key must be given as argument") unless defined($key);

	my $pattern = RE_SSH_KEY_SHORT;
	if ($key =~ /^$pattern$/)
	{
		return ($1,$2,$3);
	}
	return undef;
}

=item ssh_key_validate($) [PUBLIC]

  Usage     : my $result = ssh_key_validate($pub_key_line);
  Purpose   : Validate given public SSH key against regular expression RE_SSH_KEY
  Arguments : STRING $key - SSH key line as appears in authorized keys file
  Returns   : STRING $key on SUCCESS, undef on FAILURE
  Throws    : Croaks, if not given SSH key to validate
  Comments  :
  See Also  :

=cut

sub ssh_key_validate($)
{
	my ($key, )	= @_;
	croak ((caller(0))[3] . ": SSH key must be given as argument") unless defined($key);

	my $pattern = RE_SSH_KEY;
	if ($key =~ /^$pattern$/)
	{
		return $key;
	}
	return undef;
}

=item ssh_key_validate_short($) [PUBLIC]

  Usage     : my $result = ssh_key_validate_short($pub_key_line);
  Purpose   : Validate given public SSH key against regular expression RE_SSH_KEY_SHORT
  Arguments : STRING $key - SSH key line as appears in authorized keys file
  Returns   : STRING $key on SUCCESS, undef on FAILURE
  Throws    : Croaks, if not given SSH key to validate
  Comments  :
  See Also  :

=cut

sub ssh_key_validate_short($)
{
	my ($key, )	= @_;
	croak ((caller(0))[3] . ": SSH key must be given as argument") unless defined($key);

	my $pattern = RE_SSH_KEY_SHORT;
	if ($key =~ /^$pattern$/)
	{
		return $key;
	}
	return undef;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _ssh_keys_read($) [MODULE PRIVATE]
#
#  Usage     : my %keys = _ssh_keys_read($authorized_keys);
#  Purpose   : Read all keys from given authorized_keys file
#  Arguments : STRING $authorized_keys - path to the authorized_keys file
#  Returns   : HASH %keys containing SSH key comments as keys and ARRAY REFERENCES
#              as values, containing following items:
#                STRING $options - SSH options for this key
#                STRING $keytype - type of the key ('ssh-dss'|'ssh-rsa')
#                STRING $key - base64 encoded key
#                STRING $comment - comment
#  Throws    : Croaks on failure
#  Comments  :
#  See Also  :

sub _ssh_keys_read($)
{
	my ($authorized_keys, ) = @_;
	if ($authorized_keys =~ /(.*)/)	{ $authorized_keys = $1; }
	my %keys = ();

	open (AUTHORIZED_KEYS, $authorized_keys) or croak ((caller(0))[3] . ": could not open authorized keys file '$authorized_keys'");
	while (my $line = <AUTHORIZED_KEYS>)
	{
		trim($line);

		# Skip empty lines and comments
		if ($line =~ /^[#;].*/)	{ next; }
		if ($line =~ /^\s*$/)	{ next; }

		my @key_data = ssh_key_parse($line);
		croak ((caller(0))[3] . ": invalid SSH key '$line' found in file '$authorized_keys'") unless ($key_data[3]);

		$keys{$key_data[3]} = [@key_data];
	}
	close(AUTHORIZED_KEYS);
	return %keys;
}

# _ssh_keys_write($%) [MODULE PRIVATE]
#
#  Usage     : return _ssh_keys_write($authorized_keys, %keys);
#  Purpose   : Write given keys to given authorized_keys file
#  Arguments : STRING $authorized_keys - path to the authorized_keys file
#              HASH %keys - containing SSH key comments as keys and ARRAY REFERENCES
#              as values, containing following items:
#                STRING $options - SSH options for this key
#                STRING $keytype - type of the key ('ssh-dss'|'ssh-rsa')
#                STRING $key - base64 encoded key
#                STRING $comment - comment
#  Returns   : empty STRING on SUCCESS
#  Throws    : Croaks on failure
#  Comments  :
#  See Also  :

sub _ssh_keys_write($%)
{
	my ($authorized_keys, %keys) = @_;
	if ($authorized_keys =~ /(.*)/)	{ $authorized_keys = $1; }

	open (AUTHORIZED_KEYS, "> $authorized_keys") or croak ((caller(0))[3] . ": could not open authorized keys file '$authorized_keys' for writing");
	foreach my $index (keys(%keys))
	{
		unless ($keys{$index}->[0])
		{
			shift(@{$keys{$index}});
		}
		print AUTHORIZED_KEYS join(" ", @{$keys{$index}}) . "\n";
	}
	close(AUTHORIZED_KEYS);

	return '';
}

# _create_ak_file($$$) [MODULE PRIVATE]
#
#  Usage     : _create_ak_file($authorized_keys);
#  Purpose   : Create authorized_keys file
#  Arguments : INT $uid - user ID
#              INT $gid - group ID
#              STRING $authorized_keys - path to the authorized_keys file
#  Returns   : empty STRING on SUCCESS
#  Throws    : Croaks on failure
#  Comments  :
#  See Also  :

sub _create_ak_file($$$)
{
	my ($uid, $gid, $authorized_keys) = @_;
	if ($authorized_keys =~ /(.*)/)	{ $authorized_keys = $1; }
	if ($uid =~ /(.*)/)				{ $uid = $1; }
	if ($gid =~ /(.*)/)				{ $gid = $1; }

	open (AUTHORIZED_KEYS, "> $authorized_keys") or croak ((caller(0))[3] . ": could not open authorized keys file '$authorized_keys' for writing");
	print AUTHORIZED_KEYS "";
	close(AUTHORIZED_KEYS);
	chown($uid, $gid, $authorized_keys)	or croak ((caller(0))[3] . ": could not chown '$authorized_keys' to '$uid':'$gid'");

	return '';
}

=pod

=back

=cut

1;
